#pragma once

// AFAP includes.
#include "apigw/src/transaction.hpp"
#include "components/filesystem/src/filesystem.hpp"
#include "components/types/src/PromisesMap.hpp"
#include "pm/provider/src/pm_node.hpp"

// STL includes.
#include <future>
#include <list>
#include <shared_mutex>

namespace apigw
{

    constexpr auto verifyCallVerdictCallbackId = "VerifyCallVerdictCallback";
    constexpr auto callInformationRequestCallbackId = "CallInformationRequestCallback";
    constexpr auto storeRecordCallbackId = "StoreRecordCallback";
    constexpr auto updateRecordsCallbackId = "UpdateRecordsCallback";
    constexpr auto setNodeStatusCallbackId = "SetNodeStatusCallback";

    using SetNodeStatusCallback = std::function<void(const Uvr::NodeStatus &status)>;

    class node : public pm::node
    {
        class user_callbacks_map
        {
            using UserCallbacks = std::unordered_map<std::string, std::any>;

        public:
            user_callbacks_map() = default;
            ~user_callbacks_map();

            /// @brief Registers callback function.
            /// @note Thread-safe with unique lock.
            /// @param[in] handlerName The callback ID
            /// @param[in] func_ptr The pointer to callback function
            /// @param[in] userId The user ID (if the user ID is not defined, it will be assigned)
            /// @return On success, assigned or specified user ID, for which the function was registered.
            ///         On fail, undefined user ID value.
            Uvr::UserId_t doRegister(const std::string &handlerName,
                                     std::any func_ptr,
                                     Uvr::UserId_t userId);

            /// @brief Checks that for all the known users the callbacks in @a handlersNames list are registered.
            /// @note Thread-safe with shared lock.
            /// @param[in] handlersNames list of callbacks id to be checked.
            /// @returns @c true if the callbacks are registered for all known users,
            ///          @c false if the callbacks are not registered for at least one user.
            bool isRegistered(const std::list<std::string> &handlersNames) const;

            /// @brief Gets registered callback function by callback ID and user ID.
            /// @note Thread-safe with shared lock.
            /// @param[in] handlerName The callback ID
            /// @param[in] userId The user ID
            /// @return The registered callback function
            std::any get(const std::string &handlerName,
                         Uvr::UserId_t userId) const;

            /// @brief Gets the list of functions registered for callback ID among known users.
            /// @note Thread-safe with shared lock.
            /// @param[in] handlerName The callback ID
            /// @param[out] funcs The list of callbacks pointers
            void getForAllUsers(const std::string &handlerName,
                                std::vector<std::any> &funcs) const;

            /// @note Thread-safe with unique lock.
            void reset(void);

        private:
            mutable std::shared_mutex m_funcsGuard;
            std::unordered_map<Uvr::UserId_t, UserCallbacks> m_funcs;
        };

    protected:
        user_callbacks_map m_callbacks;
        PromisesMap<Uvr::SessionId_t, Uvr::VerificationResult> m_verificationPromises;
        PromisesMap<Uvr::SessionId_t, Uvr::UvrIds> m_uvrIdsPromises;
        PromisesMap<Uvr::SessionId_t, Uvr::UvrInfo> m_uvrInfoPromises;
        PromisesMap<Uvr::SessionId_t, Uvr::HubInfo> m_hubInfoPromises;
        PromisesMap<Uvr::SessionId_t, Uvr::OperatorInfo> m_operatorInfoPromises;

        ::transaction::manager<transaction::data> m_verifyCallsTransactionManager;
        ::transaction::manager<::transaction::base_data> m_digestsTransactionManager;

        fs::local m_fsLocal;
        Uvr::UserId_t m_internalUserId{Uvr::undefinedUserId};

        void handleSpecificPrimitive(const primitive::GenericPtr &prim) override;

        void handleVerifyCallVerdict(const primitive::GenericPtr &prim);
        void handleNotifyCallInfoReq(const primitive::GenericPtr &prim);

        void handleQueryUvrIdsRsp(const primitive::GenericPtr &prim);
        void handleQueryUvrInfoRsp(const primitive::GenericPtr &prim);
        void handleQueryHubInfoRsp(const primitive::GenericPtr &prim);
        void handleQueryOperatorInfoRsp(const primitive::GenericPtr &prim);

        void handleExecCmdReq(const primitive::GenericPtr &prim);
        void execCmdUpdateRecords(const std::shared_ptr<primitive::ExecCmdReq> &req);

        void sendExecCmdRsp(const TypeInst &to, Uvr::Strings &&responses);
        void handleStatusChangeInd(const primitive::GenericPtr &prim);

        void prepareCounters() override;

    public:
        explicit node(const TypeInst &ownInst,
                      boost::asio::io_context &ioc,
                      const configuration::manager &config);
        void start() override;

        Uvr::UserId_t doRegister(const std::string &handlerName,
                                 std::any func_ptr,
                                 Uvr::UserId_t userId);

        void doRegisterSetNodeStatusCallback(std::any callback);

        /// @brief Checks that for all the known users the callbacks in @a handlersNames list are registered.
        /// @param handlersNames list of callbacks id to be checked.
        /// @returns @c true if the callback is registered for all known users,
        ///          @c false if the callback is not registered for at least one user.
        bool isRegistered(const std::list<std::string> &handlersNames) const;

        /// @brief Verifies the call asynchronously. The verification result will
        ///        be passed to the registered callback.
        /// @returns The session ID that will be passed to the registered callback.
        ///          It can be used to understand for which call the response was
        ///          passed to the callback.
        /// @see doRegister()
        Uvr::SessionId_t doVerifyCall(Uvr::UserId_t userId,
                                      Uvr::OperatorId_t id_src,
                                      Uvr::OctetStringPtr callingNumber,
                                      Uvr::OctetStringPtr calledNumber,
                                      Uvr::OctetStringPtr originalCalledNumber,
                                      Uvr::OctetStringPtr genericNumber,
                                      const Uvr::CallId_t &callId,
                                      bool isTrusted);

        /// @brief Verifies the call synchronously.
        /// @returns Verification result, see @ref VerificationResult for more details.
        ///          If @a isTrusted was set to @c true then return @ref VerificationResult value
        ///          with verdict equal to @ref Uvr::VerificationVerdict::success.
        Uvr::VerificationResult
        doVerifyCallSynchronously(const Uvr::OperatorId_t idSrc,
                                  const Uvr::OctetStringPtr &callingNumber,
                                  const Uvr::OctetStringPtr &calledNumber,
                                  const Uvr::OctetStringPtr &originalCalledNumber,
                                  const Uvr::OctetStringPtr &genericNumber,
                                  const Uvr::CallId_t &callId,
                                  const bool isTrusted);

        /// @brief Creates a temporary entry in the calls registry for an outgoing call.
        /// @param[in] callingNumber Calling number in ITU-T Recommendation Q.763 format.
        /// @param[in] calledNumber Called number in ITU-T Recommendation Q.763 format.
        /// @param[in] callId Call ID. Empty by default.
        /// @param[in] idDst Destination Operator ID. std::nullopt by default.
        /// @see TRecQ763Number
        void doNotifyAboutOutgoingCall(Uvr::OctetStringPtr callingNumber,
                                       Uvr::OctetStringPtr calledNumber,
                                       const Uvr::CallId_t &callId = "",
                                       const Uvr::OptionalOperatorId &idDst = std::nullopt);

        void processCallInformation(const Uvr::CallInformationResponses &response);

        /// @brief Gets IDs of the UVRs to which the given @a number belongs.
        /// @param number The number for which IDs are requested.
        /// @returns UVR IDs for primary and secondary UVRs.
        Uvr::UvrIds getUvrIds(const Uvr::OctetStringPtr &number);

        /// @brief The function returns own UVR ID.
        /// @returns Own UVR ID.
        Uvr::UvrId_t getOwnUvrId();

        /// @brief Gets information about UVR for passed UVR ID.
        /// @param[in] uvrId UVR ID.
        /// @returns Information about UVR.
        Uvr::UvrInfo getUvrInfo(const Uvr::UvrId_t &uvrId);

        /// @brief Gets information about HUB for passed HUB ID.
        /// @param[in] hubId HUB ID.
        /// @returns Information about HUB.
        Uvr::HubInfo getHubInfo(const Uvr::Gt_t &hubId);

        /// @brief Gets the information about Operator for the passed Operator ID.
        /// @param[in] operatorId The Operator ID for which information is being requested.
        /// @returns Information about Operator.
        Uvr::OperatorInfo getOperatorInfo(const Uvr::OperatorId_t &operatorId);

    private:
        Uvr::SessionId_t openVerifyCallTransaction(const Uvr::UserId_t userId,
                                                   const Uvr::OperatorId_t idSrc,
                                                   const Uvr::OctetStringPtr &callingNumber,
                                                   const Uvr::OctetStringPtr &calledNumber,
                                                   const Uvr::OctetStringPtr &originalCalledNumber,
                                                   const Uvr::OctetStringPtr &genericNumber,
                                                   const Uvr::CallId_t callId);

        Uvr::SessionId_t openDigestsTransaction();

        void sendVerifyCallQuery(const Uvr::OperatorId_t idSrc,
                                 const Uvr::OctetStringPtr &callingNumber,
                                 const Uvr::OctetStringPtr &calledNumber,
                                 const Uvr::OctetStringPtr &originalCalledNumber,
                                 const Uvr::OctetStringPtr &genericNumber,
                                 const Uvr::CallId_t callId,
                                 const Uvr::SessionId_t transactionId);

        void registerCliCommands();
    };

    using base_api = workers::api<node, workers::type::APIGW, workers::max_num_of_insts>;
    class api : public base_api
    {
    public:
        static int init(boost::asio::io_context &ioc,
                        configuration::manager &config,
                        const bool asyncMode = true);

        static bool start(std::size_t instId = 0);
    };

} // namespace apigw
