#include "ccngw/src/node.hpp"

// AFAP includes.
#include "components/sequencer/src/sequencer.hpp"
#include "components/worker_node/src/config_parameters.hpp"

// STL includes.
#include <regex>
#include <filesystem>

#define GET_DIGESTS_FILES_PATH(_path, _node, _digestName, ...)                \
    {                                                                         \
        std::string digestPath =                                              \
            m_config.getValue<std::string>(DIGESTS_PATH(_node, _digestName)); \
        if (digestPath.empty())                                               \
        {                                                                     \
            LOGERROR(m_ownTypeInst, " ",                                      \
                     __VA_ARGS__,                                             \
                     ": path to " _node " digest is not configured");         \
            return;                                                           \
        }                                                                     \
        _path = digestPath;                                                   \
    }

#define GET_CONNECTIONS_FILES_PATH(_path, _node, _connName, ...)                \
    {                                                                           \
        std::string connPath =                                                  \
            m_config.getValue<std::string>(CONNECTIONS_PATH(_node, _connName)); \
        if (connPath.empty())                                                   \
        {                                                                       \
            LOGERROR(m_ownTypeInst, " ",                                        \
                     __VA_ARGS__,                                               \
                     ": path to " _node " connection is not configured");       \
            return;                                                             \
        }                                                                       \
        _path = connPath;                                                       \
    }

namespace
{
    const ccngw::ReportType s_allReportTypes[] = {
        ccngw::ReportType::Incident,
        ccngw::ReportType::IncidentA,
        ccngw::ReportType::Statistics,
        ccngw::ReportType::CallInfoRsp};
} // namespace

namespace ccngw
{

    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config) : workers::node(ownInst, ioc, config),
                                                       m_incidentsTransMngr(ownInst, GET_INCIDENTS_MNGR_ID, m_uvrId),
                                                       m_pmMngr(ownInst, PM_JOB_MNGR_ID, ioc, config),
                                                       m_sftpClient(ownInst,
                                                                    config.getValue<std::string>(CCN_SFTP_USER, ""),
                                                                    config.getValue<std::string>(CCN_IP_ADDRESS, ""),
                                                                    config.getValue<std::string>(CCN_SFTP_PORT, ""),
                                                                    config.getValue<std::string>(CCN_SFTP_PRIVATE_KEY, ""),
                                                                    config.getValue<std::string>(CCN_EXTRA_OPTIONS, "")),
                                                       m_fsLocal(ownInst)
    {
        m_pmMngr.workers::interfaces::sender::client::bind(this);
        if (!m_sftpClient.isInitialized())
        {
            ::workers::killAll();
        }
    }

    void node::start()
    {
        // Get reference books from CCN
        loadNumbersFiles();
        loadNodesFiles();
        loadOperatorsFiles();
        // Get public keys of other nodes
        loadPublicKeys();
        // Start node specific timers
        startNodeTimers();
        startPmJobs();
        // Run node
        workers::node::start();
    }

    void node::stop()
    {
        m_pmMngr.stop_and_del_all_jobs();
        workers::node::stop();
    }

    void node::startNodeTimers()
    {
        // INCIDENTS LIST timer
        m_getIncidentsTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "GET_INCIDENTS",
                m_config.getValue(TIMER_GET_INCIDENT_REQ,
                                  timers::Duration(std::chrono::hours(1))));

        // REFRESH NUMBERS TIMER
        m_refreshNumbersTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "REFRESH_NUMBERS",
                m_config.getValue(TIMER_REFRESH_NUMBERS_INFO,
                                  timers::Duration(std::chrono::hours(4))));

        // REFRESH NODES TIMER
        m_refreshNodesTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "REFRESH_NODES",
                m_config.getValue(TIMER_REFRESH_NODES_INFO,
                                  timers::Duration(std::chrono::hours(24))));

        // REFRESH OPERATORS TIMER
        m_refreshOperatorsTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "REFRESH_OPERATORS",
                m_config.getValue(TIMER_REFRESH_OPERATORS_INFO,
                                  timers::Duration(std::chrono::hours(24))));

        // REFRESH CALL INFO REQUESTS TIMER
        m_refreshCallInfoRequestsId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "REFRESH_CALL_INFO_REQUESTS",
                m_config.getValue(TIMER_REFRESH_CALL_INFO_REQUESTS,
                                  timers::Duration(std::chrono::minutes(15))));

        // REFRESH PUBLIC KEYS TIMER
        m_refreshPubKeysTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "REFRESH_PUBLIC_KEYS",
                m_config.getValue(TIMER_REFRESH_PUB_KEYS,
                                  timers::Duration(std::chrono::hours(24))));

        // RESEND REPORTS TIMER
        m_resendReportsTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "RESEND_REPORTS",
                m_config.getValue(TIMER_RESEND_REPORTS,
                                  timers::Duration(std::chrono::hours(1))));
    }

    void node::startPmJobs()
    {
        std::vector<std::string> pmJobs = m_config.getList<std::string>(PMJOBS_LIST("ccngw"));
        if (pmJobs.empty())
        {
            LOGERROR(m_ownTypeInst, " PM jobs are not configured");
            return;
        }
        if (pm::errors::OK != m_pmMngr.generate_jobs(pmJobs))
        {
            return;
        }
        m_pmMngr.start_all_jobs();
    }

    bool node::isFileActual(const timers::time_point &fileTp,
                            const timers::duration &period) const
    {
        auto secPeriod = std::chrono::duration_cast<std::chrono::seconds>(period).count();
        timers::time_point currTime = timers::now();
        currTime.tv_nsec = 0; // we need accuracy only up to seconds
        if (fileTp < currTime && (currTime - fileTp) > secPeriod)
        {
            LOGWARN(m_ownTypeInst,
                    " File is out-of-date: current time=\'", currTime.utcFormat("%Y-%m-%d %T"),
                    "\', file time=\'", fileTp.utcFormat("%Y-%m-%d %T"),
                    "\', update period=", secPeriod, "s");
            return false;
        }
        return true;
    }

    bool node::checkLatestFile(const std::string &fileMask,
                               const std::string &locDigestsPath,
                               const std::string &remDigestsPath,
                               std::shared_ptr<fs::files::list> const &updates_p) const
    {
        std::string locLatest;
        std::string locFileMask(locDigestsPath + "/" + fileMask);
        m_fsLocal.getLatestFile(locFileMask, locLatest);
        std::string remLatest;
        std::string remFileMask(remDigestsPath + "/" + fileMask);
        m_sftpClient.getLatestFile(remFileMask, remLatest);
        if (remLatest.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to get latest file from CCN folder=\"", remDigestsPath,
                    "\" by mask=\"", fileMask, "\".");
            return false;
        }
        timers::time_point locTp = fs::utils::getTimestampFromFilename(locLatest);
        timers::time_point remTp = fs::utils::getTimestampFromFilename(remLatest);
        if (remTp > locTp)
        {
            updates_p->push_back(remLatest);
        }
        return true;
    }

    bool node::checkLatestFileOnStart(const std::string &fileMask,
                                      const timers::duration &period,
                                      const std::string &locDigestsPath,
                                      const std::string &remDigestsPath,
                                      std::shared_ptr<fs::files::list> const &updates_p)
    {
        std::string locLatest;
        std::string locFileMask(locDigestsPath + "/" + fileMask);
        m_fsLocal.getLatestFile(locFileMask, locLatest);
        timers::time_point locTp = fs::utils::getTimestampFromFilename(locLatest);
        std::string remLatest;
        std::string remFileMask(remDigestsPath + "/" + fileMask);
        m_sftpClient.getLatestFile(remFileMask, remLatest);
        if (remLatest.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to get latest file from CCN folder=\"", remDigestsPath,
                    "\" by mask=\"", fileMask, "\". Checking local folder=\"",
                    locDigestsPath, "\".");

            if (locLatest.empty())
            {
                LOGERROR(m_ownTypeInst,
                         " No latest file is found in local folder=\"",
                         locDigestsPath, "\" by mask=\"", fileMask,
                         "\". Stop execution.");
                ::workers::killAll();
                return false;
            }

            if (!isFileActual(locTp, period))
            {
                LOGW(m_ownTypeInst,
                     " Outdated local latest file=\"",
                     locLatest, "\". Continue with local digest anyway.");
            }
            return true; // continue with local files
        }
        timers::time_point remTp = fs::utils::getTimestampFromFilename(remLatest);
        if (remTp > locTp)
        {
            updates_p->push_back(remLatest);
        }
        return true;
    }

    bool node::checkNumbersFiles(const std::string &locDigestsPath,
                                 const std::string &remDigestsPath,
                                 std::shared_ptr<fs::files::list> const &updates_p)
    {
        // Checks numeration files on local and remote firstly
        std::string locNumFile;
        m_fsLocal.getLatestFile(locDigestsPath + "/NUM_*.zip", locNumFile);
        // if no local numeration file, default time point will be created
        timers::time_point locTp = fs::utils::getTimestampFromFilename(locNumFile);
        std::string remNumFile;
        m_sftpClient.getLatestFile(remDigestsPath + "/NUM_*.zip", remNumFile);
        if (remNumFile.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to get the list of numeration files from CCN folder=\"",
                    remDigestsPath, "\". Checking local folder=\"", locDigestsPath, "\".");

            if (locNumFile.empty())
            {
                LOGERROR(m_ownTypeInst,
                         " No numeration files are found in local folder=\"",
                         locDigestsPath, "\". Stop execution.");
                ::workers::killAll();
                return false;
            }

            std::string locDeltaFile;
            m_fsLocal.getLatestFile(locDigestsPath + "/DELTA_*.zip", locDeltaFile);
            timers::time_point locDeltaTp = fs::utils::getTimestampFromFilename(locDeltaFile);

            auto period =
                m_config.getValue(TIMER_REFRESH_NUMBERS_INFO,
                                  timers::Duration(std::chrono::hours(4)));
            if (!isFileActual(locDeltaTp > locTp ? locDeltaTp : locTp, period))
            {
                LOGW(m_ownTypeInst,
                     " Outdated local numbers files: numeration=\"",
                     locNumFile, "\" (delta=\"", locDeltaFile, "\"). Continue with local numeration digests anyway.");
            }
            return true; // continue with local files
        }
        timers::time_point remTp = fs::utils::getTimestampFromFilename(remNumFile);
        // Local numeration file is up-to-date
        // Checks new Deltas only
        if (locTp >= remTp)
        {
            fs::files::list locDeltas;
            fs::files::list locDeltaAfterNumList;
            m_fsLocal.listDir(locDigestsPath + "/DELTA_*.zip", locDeltas);
            fs::utils::filterFilesByTimestamp(locTp, locDeltas, locDeltaAfterNumList);
            fs::files::list remDeltas;
            fs::files::list remDeltaAferNumList;
            m_sftpClient.listDir(remDigestsPath + "/DELTA_*.zip", remDeltas);
            fs::utils::filterFilesByTimestamp(locTp, remDeltas, remDeltaAferNumList);
            fs::utils::findNewFiles(remDeltaAferNumList, locDeltaAfterNumList, *updates_p);
            return true;
        }
        // Local numeration file is old or absent
        // Upload fresh numeration file and all Deltas after it
        updates_p->push_back(remNumFile);
        fs::files::list remDeltas;
        m_sftpClient.listDir(remDigestsPath + "/DELTA_*.zip", remDeltas);
        fs::utils::filterFilesByTimestamp(remTp, remDeltas, *updates_p);
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
        for (auto filename : *updates_p)
        {
            LOGDEBUG(m_ownTypeInst, " New file: ", filename);
        }
#endif
        return true;
    }

    bool node::checkDeltaFilesUpdates(const std::string &locDigestsPath,
                                      const std::string &remDigestsPath,
                                      std::shared_ptr<fs::files::list> const &updates_p) const
    {
        std::string locNumLatest;
        m_fsLocal.getLatestFile(locDigestsPath + "/NUM_*.zip", locNumLatest);
        timers::time_point locNumTp = fs::utils::getTimestampFromFilename(locNumLatest);
        std::string locDeltaLatest;
        m_fsLocal.getLatestFile(locDigestsPath + "/DELTA_*.zip", locDeltaLatest);
        timers::time_point locDeltaTp = fs::utils::getTimestampFromFilename(locDeltaLatest);
        timers::time_point locTp = locNumTp > locDeltaTp ? locNumTp : locDeltaTp;
        fs::files::list remDeltas;
        m_sftpClient.listDir(remDigestsPath + "/DELTA_*.zip", remDeltas);
        updates_p->clear();
        fs::utils::filterFilesByTimestamp(locTp, remDeltas, *updates_p);
        return !updates_p->empty();
    }

    bool node::checkNodesFilesUpdates(const std::string &locDigestsPath,
                                      const std::string &remDigestsPath,
                                      std::shared_ptr<fs::files::list> const &updates_p) const
    {
        checkLatestFile("UVR_*.zip", locDigestsPath, remDigestsPath, updates_p);
        checkLatestFile("HUB_*.zip", locDigestsPath, remDigestsPath, updates_p);
        return (updates_p->size() != 0);
    }

    bool node::checkOperatorsFilesUpdates(const std::string &locDigestsPath,
                                          const std::string &remDigestsPath,
                                          const std::shared_ptr<fs::files::list> &updates) const
    {
        checkLatestFile("OPR_*.zip", locDigestsPath, remDigestsPath, updates);
        return (updates->size() != 0);
    }

    bool node::checkCallInfoRequestsUpdates(const std::string &remDigestsPath,
                                            std::shared_ptr<fs::files::list> const &updates_p) const
    {
        std::string fileMask = "REQ_" + std::to_string(m_uvrId) + "_*.zip";
        fs::files::list remList;
        if (!m_sftpClient.listDir(remDigestsPath + "/" + fileMask, remList))
            return false;

        updates_p->insert(updates_p->end(), remList.begin(), remList.end());

        return true;
    }

    bool node::checkPublicKeysUpdates(const std::string &prefix,
                                      const std::string &newDir,
                                      const std::string &currDir,
                                      std::shared_ptr<fs::files::list> const &toCopyList_p,
                                      std::shared_ptr<fs::files::list> const &toRemoveList_p) const
    {
        // Get the list of files in the new directory
        toCopyList_p->clear();
        if (!m_fsLocal.listDir(newDir + "/" + prefix + "*-key.pub", *toCopyList_p))
            return false;
        // Get the list of files in the current directory
        toRemoveList_p->clear();
        if (!m_fsLocal.listDir(currDir + "/" + prefix + "*-key.pub", *toRemoveList_p))
            return false;

        // Compare files in both lists to generate copy and remove lists
        int newKeyIx = toCopyList_p->size() - 1;
        int currKeyIx = toRemoveList_p->size() - 1;
        while (newKeyIx >= 0 && currKeyIx >= 0)
        {
            // Try to compare two files with the same name
            int result = m_fsLocal.compareFiles(newDir + "/" + (*toCopyList_p)[newKeyIx],
                                                currDir + "/" + (*toCopyList_p)[newKeyIx]);

            // Find file position in the remove list
            if ((*toCopyList_p)[newKeyIx] == (*toRemoveList_p)[currKeyIx])
            {
                toRemoveList_p->erase(toRemoveList_p->begin() + currKeyIx);
            }
            else
            {
                // File is absent in the current directory
                // and should be left in the copy list
                if (result == fs::files::not_exist)
                {
                    // go to next file
                    --newKeyIx;
                    continue;
                }
                // If file exists in both directories,
                // find its position to delete from the remove list
                while (currKeyIx != 0 && (*toCopyList_p)[newKeyIx] != (*toRemoveList_p)[currKeyIx])
                    --currKeyIx;
                // As the lists are sorted in alphabetic order,
                // all files before found file are not exist in the new directory,
                // so, leave them in the remove list and delete only the found one
                if (currKeyIx >= 0)
                {
                    toRemoveList_p->erase(toRemoveList_p->begin() + currKeyIx);
                }
            }
            // remove it from copy list as files are equals
            if (result == fs::files::match)
                toCopyList_p->erase(toCopyList_p->begin() + newKeyIx);

            --newKeyIx;
            --currKeyIx;
        }
        return true;
    }

    bool node::updatePublicKeys(const std::string &newDir,
                                const std::string &currDir,
                                std::shared_ptr<fs::files::list> const &toCopyList_p,
                                std::shared_ptr<fs::files::list> const &toRemoveList_p) const
    {
        // Update files in the local directory
        LOGDEBUG(m_ownTypeInst,
                 " Going to copy ", toCopyList_p->size(), " files");
        for (auto const &filename : *toCopyList_p)
        {
            m_fsLocal.copyFile(newDir + "/" + filename,
                               currDir + "/" + filename);
        }
        // Delete files in the local directory
        LOGDEBUG(m_ownTypeInst,
                 " Going to remove ", toRemoveList_p->size(), " files");
        for (auto const &filename : *toRemoveList_p)
        {
            m_fsLocal.deleteFile(currDir + "/" + filename);
        }
        return true;
    }

    bool node::downloadFiles(const std::string &locDigestsPath,
                             const std::string &remDigestsPath,
                             std::shared_ptr<fs::files::list> const &updates_p) const
    {
        for (auto const &fileIter : *updates_p)
        {
            std::string filename(remDigestsPath + "/" + fileIter);
            m_sftpClient.getFile(filename, locDigestsPath);
        }
        return true;
    }

    bool node::notifyAboutNumbersFiles(const std::string &locDigestsPath)
    {
        // Get latest numeration file
        std::string numFile;
        m_fsLocal.getLatestFile(locDigestsPath + "/NUM_*.zip", numFile);
        if (numFile.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to provide VFC with Number file: NUM file is absent locally");
            return false;
        }
        timers::time_point locTp = fs::utils::getTimestampFromFilename(numFile);
        // Get DELTAs after it
        Uvr::Strings deltasList;
        m_fsLocal.listDir(locDigestsPath + "/DELTA_*.zip", deltasList);
        Uvr::Strings deltasAfterNumList;
        fs::utils::filterFilesByTimestamp(locTp, deltasList, deltasAfterNumList);
        Uvr::Strings finalDeltasList;
        for (auto const &deltaFile : deltasAfterNumList)
        {
            finalDeltasList.emplace_back(locDigestsPath + "/" + deltaFile);
        }
        sendUploadNumbersInd(locDigestsPath + "/" + numFile, finalDeltasList);
        return true;
    }

    bool node::notifyAboutNumbersFilesUpdates(const std::string &locDigestsPath,
                                              std::shared_ptr<fs::files::list> const &updates_p)
    {
        Uvr::Strings deltas;
        for (auto const &deltaFile : *updates_p)
        {
            deltas.emplace_back(locDigestsPath + "/" + deltaFile);
        }
        sendUploadNumbersInd("", deltas);
        return true;
    }

    bool node::notifyAboutNodesFiles(const std::string &locDigestsPath)
    {
        // Get nodes files, UVR file should be put as first one into the list
        fs::files::list files;
        // Get latest file with UVRs
        std::string uvrFile;
        m_fsLocal.getLatestFile(locDigestsPath + "/UVR_*.zip", uvrFile);
        if (uvrFile.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to provide VFGW with Nodes files: UVR file is absent locally");
            return false;
        }
        files.emplace_back(locDigestsPath + "/" + uvrFile);
        // Get latest file with HUBs
        std::string hubFile;
        m_fsLocal.getLatestFile(locDigestsPath + "/HUB_*.zip", hubFile);
        if (hubFile.empty())
        {
            LOGWARN(m_ownTypeInst,
                    " Failed to provide VFGW with Nodes files: HUB file is absent locally");
            return false;
        }
        files.emplace_back(locDigestsPath + "/" + hubFile);
        sendUploadNodesInd(files);
        return true;
    }

    bool node::notifyAboutNodesFilesUpdates(const std::string &locDigestsPath,
                                            std::shared_ptr<fs::files::list> const &updates_p)
    {
        Uvr::Strings files;
        for (auto const &nodeFile : *updates_p)
        {
            files.emplace_back(locDigestsPath + "/" + nodeFile);
        }
        sendUploadNodesInd(files);
        return true;
    }

    bool node::notifyAboutOperatorsFiles(const std::string &locDigestsPath)
    {
        // Get latest file with Operators.
        std::string operatorsFile;
        m_fsLocal.getLatestFile(locDigestsPath + "/OPR_*.zip", operatorsFile);
        if (operatorsFile.empty())
        {
            LOGINFO(m_ownTypeInst,
                    " Failed to provide VFGW with Operators files: Operators file is absent locally");
            return false;
        }

        fs::files::list files;
        files.emplace_back(locDigestsPath + "/" + operatorsFile);
        sendUploadOperatorsInd(files);

        return true;
    }

    bool node::notifyAboutOperatorsFilesUpdates(const std::string &locDigestsPath,
                                                const std::shared_ptr<fs::files::list> &updates)
    {
        Uvr::Strings files;
        for (auto const &nodeFile : *updates)
        {
            files.emplace_back(locDigestsPath + "/" + nodeFile);
        }
        sendUploadOperatorsInd(files);
        return true;
    }

    bool node::notifyAboutCallInfoRequestsUpdates(const Uvr::UvrId_t uvrId,
                                                  const std::string &locDigestsPath,
                                                  std::shared_ptr<fs::files::list> const &updates_p)
    {
        Uvr::Strings files;
        for (auto const &reqFile : *updates_p)
        {
            files.emplace_back(locDigestsPath + "/" + reqFile);
        }
        sendNotifyCallInfoReq(uvrId, files);
        return true;
    }

    bool node::notifyAboutPublicKeys(const std::string &locDigestsPath)
    {
        fs::files::list hubKeysList;
        if (!m_fsLocal.listDir(locDigestsPath + "/hub*-key.pub", hubKeysList))
            return false;
        fs::files::list nodeKeysList;
        if (!m_fsLocal.listDir(locDigestsPath + "/node*-key.pub", nodeKeysList))
            return false;
        Uvr::Strings hubFilePathsToCopy;
        for (auto const &filename : hubKeysList)
        {
            hubFilePathsToCopy.emplace_back(locDigestsPath + "/" + filename);
        }
        Uvr::Strings nodeFilePathsToCopy;
        for (auto const &filename : nodeKeysList)
        {
            nodeFilePathsToCopy.emplace_back(locDigestsPath + "/" + filename);
        }
        sendUploadPublicKeys(hubFilePathsToCopy, nodeFilePathsToCopy, {}, {});
        return true;
    }

    bool node::notifyAboutPublicKeysUpdates(const std::string &locDigestsPath,
                                            std::shared_ptr<fs::files::list> const &hubFilesToCopy_p,
                                            std::shared_ptr<fs::files::list> const &nodeFilesToCopy_p,
                                            std::shared_ptr<fs::files::list> const &hubFilesToRemove_p,
                                            std::shared_ptr<fs::files::list> const &nodeFilesToRemove_p)
    {
        Uvr::Strings hubFilePathsToCopy;
        for (auto const &filename : *hubFilesToCopy_p)
        {
            hubFilePathsToCopy.emplace_back(locDigestsPath + "/" + filename);
        }
        Uvr::Strings nodeFilePathsToCopy;
        for (auto const &filename : *nodeFilesToCopy_p)
        {
            nodeFilePathsToCopy.emplace_back(locDigestsPath + "/" + filename);
        }
        Uvr::Strings hubFilePathsToRemove;
        for (auto const &filename : *hubFilesToRemove_p)
        {
            hubFilePathsToRemove.emplace_back(locDigestsPath + "/" + filename);
        }
        Uvr::Strings nodeFilePathsToRemove;
        for (auto const &filename : *nodeFilesToRemove_p)
        {
            nodeFilePathsToRemove.emplace_back(locDigestsPath + "/" + filename);
        }
        if (!hubFilesToCopy_p->empty() || !nodeFilesToCopy_p->empty() || !hubFilesToRemove_p->empty() || !nodeFilesToRemove_p->empty())
        {
            sendUploadPublicKeys(hubFilePathsToCopy, nodeFilePathsToCopy,
                                 hubFilePathsToRemove, nodeFilePathsToRemove);
        }
        return true;
    }

    void node::sendUploadNumbersInd(const std::string &numFile,
                                    const Uvr::Strings &deltas)
    {
        auto prim = std::make_shared<primitive::UploadNumbersInd>(numFile, deltas);
        prim->m_header->to = TypeInst(workers::type::VFC, 0);
        workers::node::send(prim);
    }

    void node::sendUploadNodesInd(const Uvr::Strings &files)
    {
        auto prim = std::make_shared<primitive::UploadNodesInd>(files);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        workers::node::send(prim);
    }

    void node::sendUploadOperatorsInd(const Uvr::Strings &files)
    {
        auto prim = std::make_shared<primitive::UploadOperatorsInd>(files);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        workers::node::send(prim);
    }

    void node::sendNotifyCallInfoReq(const Uvr::UvrId_t uvrId,
                                     const Uvr::Strings &files)
    {
        auto prim = std::make_shared<primitive::NotifyCallInfoRequest>(uvrId, files);
        prim->m_header->to = TypeInst(workers::type::APIGW, 0);
        workers::node::send(prim);
    }

    void node::sendUploadPublicKeys(const Uvr::Strings &hubFilesToCopy,
                                    const Uvr::Strings &nodeFilesToCopy,
                                    const Uvr::Strings &hubFilesToRemove,
                                    const Uvr::Strings &nodeFilesToRemove)
    {
        auto prim = std::make_shared<::primitive::UploadPublicKeysInd>(
            hubFilesToCopy, nodeFilesToCopy,
            hubFilesToRemove, nodeFilesToRemove);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        workers::node::send(prim);
    }

    void node::sendReportToCcn(const std::string &localFilePath,
                               const ReportType reportType)
    {
        std::string unsentDir;
        std::string remoteDir;
        getReportDirs(reportType, unsentDir, remoteDir);

        if (remoteDir.empty())
        {
            LOGERROR(m_logTag, "Failed to send report=\"", localFilePath, "\".");
            return;
        }

        const auto zipFilePath = std::filesystem::path(localFilePath).replace_extension("zip").string();
        if (!m_fsLocal.zipFile(localFilePath, zipFilePath))
        {
            return;
        }

        if (m_sftpClient.sendFile(zipFilePath, remoteDir))
        {
            m_fsLocal.deleteFile(zipFilePath);
        }
        else
        {
            auto seq = std::make_shared<sequencer>("Move Report To Unsent", m_ownTypeInst, &m_ioContext);
            seq->addFunction(std::bind(&fs::local::createDir, &m_fsLocal, unsentDir));
            seq->addFunction(std::bind(&fs::local::copyFile, &m_fsLocal, zipFilePath, unsentDir));
            seq->addFunction(std::bind(&fs::local::deleteFile, &m_fsLocal, zipFilePath));
            seq->start();
        }
    }

    bool node::removeCallInfoRequestFile(const std::string &callInfoResponseFilePath) const
    {
        if (callInfoResponseFilePath.empty())
        {
            LOGERROR(m_ownTypeInst,
                     " Failed to remove Call Info Request file: Call Info Response file path is empty.");
            return false;
        }

        // Extract ID_UVR and ID_REQ from Call Info Response file name.
        const std::filesystem::path filePath(callInfoResponseFilePath);
        const std::string fileStem = filePath.stem().string();
        const std::regex idsRegex("RSP_(\\d+)_(\\d+)_.*");
        std::smatch idsMatches;
        if (!std::regex_match(fileStem, idsMatches, idsRegex))
        {
            LOGERROR(m_ownTypeInst,
                     " Failed to remove Call Info Request file:"
                     " Unable to extract ID_UVR and ID_REQ from Call Info Response file name.");
            return false;
        }

        // Remove Call Info Request file(s).
        std::string connectionsPath =
            m_config.getValue<std::string>(CONNECTIONS_PATH("ccngw", CALL_INFO_REQUESTS));
        if (connectionsPath.empty())
        {
            LOGERROR(m_ownTypeInst,
                     " Failed to remove Call Info Request file:",
                     " Path to ccngw connections is not configured.");
            return false;
        }

        const std::string callInfoRequestFileMask =
            std::string("REQ_") + idsMatches[1].str() + "_" + idsMatches[2].str() + "_*.zip";

        return m_fsLocal.deleteFile(connectionsPath + "/" + callInfoRequestFileMask);
    }

    void node::refreshNumbersFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_NUMBERS,
                               "Failed to refresh \"", DIGESTS_NUMBERS, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_NUMBERS,
                               "Failed to refresh \"", DIGESTS_NUMBERS, "\" digests files");

        auto updates_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Refresh Numbers Files", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&node::checkDeltaFilesUpdates, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::notifyAboutNumbersFilesUpdates, this, locDigestsPath, updates_p));
        seq_p->start();
    }

    void node::refreshNodesFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_NODES,
                               "Failed to refresh \"", DIGESTS_NODES, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_NODES,
                               "Failed to refresh \"", DIGESTS_NODES, "\" digests files");

        auto updates_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Refresh Nodes Files", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&node::checkNodesFilesUpdates, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::notifyAboutNodesFilesUpdates, this, locDigestsPath, updates_p));
        seq_p->start();
    }

    void node::refreshOperatorsFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_OPERATORS,
                               "Failed to refresh \"", DIGESTS_OPERATORS, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_OPERATORS,
                               "Failed to refresh \"", DIGESTS_OPERATORS, "\" digests files");

        auto updates = std::make_shared<fs::files::list>();

        auto seq = std::make_shared<sequencer>("Refresh Operators Files", m_ownTypeInst, &m_ioContext);
        seq->addFunction(std::bind(&node::checkOperatorsFilesUpdates, this, locDigestsPath, remDigestsPath, updates));
        seq->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates));
        seq->addFunction(std::bind(&node::notifyAboutOperatorsFilesUpdates, this, locDigestsPath, updates));
        seq->start();
    }

    void node::refreshCallInfoRequestsFiles()
    {
        // TODO[multi-uvr]: Iterate through all serviced UVR IDs.

        std::string locConnectionsPath;
        GET_CONNECTIONS_FILES_PATH(locConnectionsPath, "ccngw", CALL_INFO_REQUESTS,
                                   "Failed to refresh \"", CALL_INFO_REQUESTS, "\" digests files");
        std::string remConnectionsPath;
        GET_CONNECTIONS_FILES_PATH(remConnectionsPath, "ccn", CALL_INFO_REQUESTS,
                                   "Failed to refresh \"", CALL_INFO_REQUESTS, "\" digests files");

        auto updates_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Refresh Call Info Requests Files", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&node::checkCallInfoRequestsUpdates, this, remConnectionsPath, updates_p));
        seq_p->addFunction(std::bind(&node::downloadFiles, this, locConnectionsPath, remConnectionsPath, updates_p));
        // TODO[multi-uvr]: Use proper UVR ID instead of m_uvrId.
        seq_p->addFunction(std::bind(&node::notifyAboutCallInfoRequestsUpdates, this, m_uvrId, locConnectionsPath, updates_p));
        seq_p->start();
    }

    void node::refreshPublicKeys()
    {
        std::string locKeysPath;
        GET_DIGESTS_FILES_PATH(locKeysPath, "ccngw", DIGESTS_PUBKEYS,
                               "Failed to refresh public keys");
        std::string newLocKeysPath(locKeysPath + "/new");
        std::string remKeysPath;
        GET_DIGESTS_FILES_PATH(remKeysPath, "ccn", DIGESTS_PUBKEYS,
                               "Failed to refresh public keys");

        auto toCopyHubFiles_p = std::make_shared<fs::files::list>();
        auto toRemoveHubFiles_p = std::make_shared<fs::files::list>();
        auto toCopyNodeFiles_p = std::make_shared<fs::files::list>();
        auto toRemoveNodeFiles_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Refresh Public Keys", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&fs::local::createDir, &m_fsLocal, newLocKeysPath));
        seq_p->addFunction(std::bind(&sftp::client::getFile, &m_sftpClient, remKeysPath + "/*-key.pub", newLocKeysPath));
        seq_p->addFunction(std::bind(&node::checkPublicKeysUpdates, this, "hub", newLocKeysPath, locKeysPath, toCopyHubFiles_p, toRemoveHubFiles_p));
        seq_p->addFunction(std::bind(&node::checkPublicKeysUpdates, this, "node", newLocKeysPath, locKeysPath, toCopyNodeFiles_p, toRemoveNodeFiles_p));
        seq_p->addFunction(std::bind(&node::updatePublicKeys, this, newLocKeysPath, locKeysPath, toCopyHubFiles_p, toRemoveHubFiles_p));
        seq_p->addFunction(std::bind(&node::updatePublicKeys, this, newLocKeysPath, locKeysPath, toCopyNodeFiles_p, toRemoveNodeFiles_p));
        seq_p->addFunction(std::bind(&fs::local::deleteDir, &m_fsLocal, newLocKeysPath));
        seq_p->addFunction(std::bind(&node::notifyAboutPublicKeysUpdates, this, locKeysPath, toCopyHubFiles_p, toCopyNodeFiles_p, toRemoveHubFiles_p, toRemoveNodeFiles_p));
        seq_p->start();
    }

    void node::loadNumbersFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_NUMBERS,
                               "Failed to load \"", DIGESTS_NUMBERS, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_NUMBERS,
                               "Failed to load \"", DIGESTS_NUMBERS, "\" digests files");

        auto updates_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Load Numbers Files", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&node::checkNumbersFiles, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::notifyAboutNumbersFiles, this, locDigestsPath));
        seq_p->start();
    }

    void node::loadNodesFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_NODES,
                               "Failed to load \"", DIGESTS_NODES, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_NODES,
                               "Failed to load \"", DIGESTS_NODES, "\" digests files");
        timers::duration period =
            m_config.getValue(TIMER_REFRESH_NODES_INFO,
                              timers::Duration(std::chrono::hours(24)));

        auto updates_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Load Nodes Files", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&node::checkLatestFileOnStart, this, "UVR_*.zip", period, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::checkLatestFileOnStart, this, "HUB_*.zip", period, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates_p));
        seq_p->addFunction(std::bind(&node::notifyAboutNodesFiles, this, locDigestsPath));
        seq_p->start();
    }

    void node::loadOperatorsFiles()
    {
        std::string locDigestsPath;
        GET_DIGESTS_FILES_PATH(locDigestsPath, "ccngw", DIGESTS_OPERATORS,
                               "Failed to load \"", DIGESTS_OPERATORS, "\" digests files");
        std::string remDigestsPath;
        GET_DIGESTS_FILES_PATH(remDigestsPath, "ccn", DIGESTS_OPERATORS,
                               "Failed to load \"", DIGESTS_OPERATORS, "\" digests files");
        timers::duration period =
            m_config.getValue(TIMER_REFRESH_OPERATORS_INFO,
                              timers::Duration(std::chrono::hours(24)));

        auto updates = std::make_shared<fs::files::list>();

        auto seq = std::make_shared<sequencer>("Load Operators Files", m_ownTypeInst, &m_ioContext);
        seq->addFunction(std::bind(&node::checkLatestFileOnStart, this, "OPR_*.zip", period, locDigestsPath, remDigestsPath, updates));
        seq->addFunction(std::bind(&node::downloadFiles, this, locDigestsPath, remDigestsPath, updates));
        seq->addFunction(std::bind(&node::notifyAboutOperatorsFiles, this, locDigestsPath));
        seq->start();
    }

    void node::loadPublicKeys()
    {
        std::string locKeysPath;
        GET_DIGESTS_FILES_PATH(locKeysPath, "ccngw", DIGESTS_PUBKEYS,
                               "Failed to load public keys");
        std::string newLocKeysPath(locKeysPath + "/new");
        std::string remKeysPath;
        GET_DIGESTS_FILES_PATH(remKeysPath, "ccn", DIGESTS_PUBKEYS,
                               "Failed to load public keys");

        auto toCopyFiles_p = std::make_shared<fs::files::list>();
        auto toRemoveFiles_p = std::make_shared<fs::files::list>();

        auto seq_p = std::make_shared<sequencer>("Load Public Keys", m_ownTypeInst, &m_ioContext);
        seq_p->addFunction(std::bind(&fs::local::createDir, &m_fsLocal, newLocKeysPath));
        seq_p->addFunction(std::bind(&sftp::client::getFile, &m_sftpClient, remKeysPath + "/*-key.pub", newLocKeysPath));
        seq_p->addFunction(std::bind(&node::checkPublicKeysUpdates, this, "", newLocKeysPath, locKeysPath, toCopyFiles_p, toRemoveFiles_p));
        seq_p->addFunction(std::bind(&node::updatePublicKeys, this, newLocKeysPath, locKeysPath, toCopyFiles_p, toRemoveFiles_p));
        seq_p->addFunction(std::bind(&fs::local::deleteDir, &m_fsLocal, newLocKeysPath));
        seq_p->addFunction(std::bind(&node::notifyAboutPublicKeys, this, locKeysPath));
        seq_p->start();
    }

    void node::resendReports()
    {
        for (const auto reportType : s_allReportTypes)
        {
            std::string unsentDir;
            std::string remoteDir;
            getReportDirs(reportType, unsentDir, remoteDir);

            if (remoteDir.empty())
            {
                LOGERROR(m_logTag, "Failed to resend reports=\"", unsentDir, "\".");
                continue;
            }

            fs::files::list unsentReports;
            m_fsLocal.listDir(unsentDir, unsentReports);
            for (const auto &unsentReport : unsentReports)
            {
                const auto unsentReportPath = unsentDir + unsentReport;
                if (m_sftpClient.sendFile(unsentReportPath, remoteDir))
                {
                    m_fsLocal.deleteFile(unsentReportPath);
                }
            }
        }
    }

    void node::getReportDirs(const ReportType reportType,
                             std::string &unsentDir,
                             std::string &remoteDir) const
    {
        std::string localDirDefaultValue;
        std::string localDirConfigKey = CCNGW_REPORT_PATH;
        std::string remoteDirConfigKey = CCN_REPORT_PATH;
        switch (reportType)
        {
            case ReportType::Incident:
            {
                localDirDefaultValue = INCIDENTS_REPORT_DEFAULT_DIR;
                localDirConfigKey += INCIDENTS_REPORT;
                remoteDirConfigKey += INCIDENTS_REPORT;
                break;
            }
            case ReportType::IncidentA:
            {
                localDirDefaultValue = INCIDENTSA_REPORT_DEFAULT_DIR;
                localDirConfigKey += INCIDENTSA_REPORT;
                remoteDirConfigKey += INCIDENTSA_REPORT;
                break;
            }
            case ReportType::Statistics:
            {
                localDirDefaultValue = STATS_REPORT_DEFAULT_DIR;
                localDirConfigKey += STATS_REPORT;
                remoteDirConfigKey += STATS_REPORT;
                break;
            }
            case ReportType::CallInfoRsp:
            {
                localDirDefaultValue = CALL_INFO_RSP_DEFAULT_DIR;
                localDirConfigKey = CONNECTIONS_PATH("ccngw", CALL_INFO_RESPONSES);
                remoteDirConfigKey = CONNECTIONS_PATH("ccn", CALL_INFO_RESPONSES);
                break;
            }
        }

        unsentDir = m_config.getValue<std::string>(localDirConfigKey, localDirDefaultValue)
                    + UNSENT_REPORTS_SUBDIR;

        remoteDir = m_config.getValue<std::string>(remoteDirConfigKey);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case primitive::GET_INCIDENTS_RSP:
            {
                transaction::errors res;
                m_incidentsTransMngr.closeTransaction(prim->m_header->ulpKey, res);
                if (transaction::errors::OK != res)
                {
                    LOGERROR(m_ownTypeInst,
                             " Received ", primitive::id2str(prim->m_header->id), " for unknown transactionId=", prim->m_header->ulpKey);
                    return;
                }

                auto rsp = std::dynamic_pointer_cast<::primitive::GetIncidentsRsp>(prim);
                if (rsp)
                {
                    sendReportToCcn(rsp->filename, ReportType::Incident);
                }
                break;
            }
            case primitive::SEND_INCIDENTSA_IND:
            {
                auto rsp = std::dynamic_pointer_cast<::primitive::GetIncidentsARsp>(prim);
                if (rsp)
                {
                    sendReportToCcn(rsp->filename, ReportType::IncidentA);
                }
                break;
            }
            case primitive::GET_STATISTICS_RSP:
            {
                auto rsp = std::dynamic_pointer_cast<::pm::primitive::GetStatisticsRsp>(prim);
                if (rsp)
                {
                    std::string reportName;
                    if (m_pmMngr.generate_report(rsp, reportName))
                    {
                        sendReportToCcn(reportName, ReportType::Statistics);
                    }
                }
                break;
            }
            case primitive::NOTIFY_CALL_INFO_RSP:
            {
                auto rsp = std::dynamic_pointer_cast<primitive::NotifyCallInfoResponse>(prim);
                if (rsp)
                {
                    const std::string callInfoResponseFilePath = rsp->filename;
                    sendReportToCcn(callInfoResponseFilePath, ReportType::CallInfoRsp);
                    removeCallInfoRequestFile(callInfoResponseFilePath);
                }
                break;
            }
            default:
                workers::node::handleSpecificPrimitive(prim);
        }
    }

    void node::handle_timeout(const ::timers::id &id,
                              const ::timers::time_point &startTime)
    {
        if (id == m_getIncidentsTimerId)
        {
            transaction::errors result;
            auto prim = std::make_shared<primitive::GetIncidentsReq>();
            prim->m_header->to = TypeInst(workers::type::VFC, 0);
            prim->m_header->ulpKey = m_incidentsTransMngr.openTransaction(result);
            if (result == transaction::errors::OK)
            {
                workers::node::send(prim);
            }
            else
            {
                LOGERROR(m_ownTypeInst,
                         " The sending of ", primitive::id2str(prim->m_header->id), " is rejected due transaction issue");
            }
        }
        else if (id == m_refreshNumbersTimerId)
        {
            refreshNumbersFiles();
        }
        else if (id == m_refreshNodesTimerId)
        {
            refreshNodesFiles();
        }
        else if (id == m_refreshOperatorsTimerId)
        {
            refreshOperatorsFiles();
        }
        else if (id == m_refreshCallInfoRequestsId)
        {
            refreshCallInfoRequestsFiles();
        }
        else if (id == m_refreshPubKeysTimerId)
        {
            refreshPublicKeys();
        }
        else if (id == m_resendReportsTimerId)
        {
            resendReports();
        }
        else
        {
            LOGERROR(m_ownTypeInst, " Unknown timer: id=", id);
        }
    }

} // namespace ccngw
