#pragma once

// AFAP includes.
#include "components/filesystem/src/filesystem.hpp"
#include "components/sftp/src/sftp_client.hpp"
#include "components/worker_node/src/transaction_manager.hpp"
#include "components/worker_node/src/node.hpp"
#include "pm/manager/src/pm_manager.hpp"

// Boost includes.
#include <boost/function.hpp>

namespace ccngw
{

    constexpr int GET_INCIDENTS_MNGR_ID = 1;
    constexpr int PM_JOB_MNGR_ID = 2;

    enum class ReportType
    {
        Incident,
        IncidentA,
        Statistics,
        CallInfoRsp
    };

    class node : public workers::node
    {
        void startNodeTimers();
        void startPmJobs();

        bool isFileActual(const timers::time_point &fileTp,
                          const timers::duration &period) const;

        bool checkLatestFile(const std::string &fileMask,
                             const std::string &locDigestsPath,
                             const std::string &remDigestsPath,
                             std::shared_ptr<fs::files::list> const &updates_p) const;

        bool checkLatestFileOnStart(const std::string &fileMask,
                                    const timers::duration &period,
                                    const std::string &locDigestsPath,
                                    const std::string &remDigestsPath,
                                    std::shared_ptr<fs::files::list> const &updates_p);

        bool checkNumbersFiles(const std::string &locDigestsPath,
                               const std::string &remDigestsPath,
                               std::shared_ptr<fs::files::list> const &updates_p);

        bool checkDeltaFilesUpdates(const std::string &locDigestsPath,
                                    const std::string &remDigestsPath,
                                    std::shared_ptr<fs::files::list> const &updates_p) const;
        bool checkNodesFilesUpdates(const std::string &locDigestsPath,
                                    const std::string &remDigestsPath,
                                    std::shared_ptr<fs::files::list> const &updates_p) const;

        bool checkOperatorsFilesUpdates(const std::string &locDigestsPath,
                                        const std::string &remDigestsPath,
                                        const std::shared_ptr<fs::files::list> &updates) const;

        bool checkCallInfoRequestsUpdates(const std::string &remDigestsPath,
                                          std::shared_ptr<fs::files::list> const &updates_p) const;
        bool checkPublicKeysUpdates(const std::string &prefix,
                                    const std::string &newDir,
                                    const std::string &currDir,
                                    std::shared_ptr<fs::files::list> const &toCopyList_p,
                                    std::shared_ptr<fs::files::list> const &toRemoveList_p) const;

        bool updatePublicKeys(const std::string &newDir,
                              const std::string &currDir,
                              std::shared_ptr<fs::files::list> const &toCopyList_p,
                              std::shared_ptr<fs::files::list> const &toRemoveList_p) const;

        bool downloadFiles(const std::string &locDigestsPath,
                           const std::string &remDigestsPath,
                           std::shared_ptr<fs::files::list> const &updates_p) const;

        bool notifyAboutNumbersFiles(const std::string &locDigestsPath);
        bool notifyAboutNumbersFilesUpdates(const std::string &locDigestsPath,
                                            std::shared_ptr<fs::files::list> const &updates_p);
        bool notifyAboutNodesFiles(const std::string &locDigestsPath);
        bool notifyAboutNodesFilesUpdates(const std::string &locDigestsPath,
                                          std::shared_ptr<fs::files::list> const &updates_p);

        bool notifyAboutOperatorsFiles(const std::string &locDigestsPath);
        bool notifyAboutOperatorsFilesUpdates(const std::string &locDigestsPath,
                                              const std::shared_ptr<fs::files::list> &updates);

        bool notifyAboutCallInfoRequestsUpdates(const Uvr::UvrId_t uvrId,
                                                const std::string &locDigestsPath,
                                                std::shared_ptr<fs::files::list> const &updates_p);

        bool notifyAboutPublicKeys(const std::string &locDigestsPath);
        bool notifyAboutPublicKeysUpdates(const std::string &locDigestsPath,
                                          std::shared_ptr<fs::files::list> const &hubFilesToCopy_p,
                                          std::shared_ptr<fs::files::list> const &nodeFilesToCopy_p,
                                          std::shared_ptr<fs::files::list> const &hubFilesToRemove_p,
                                          std::shared_ptr<fs::files::list> const &nodeFilesToRemove_p);

        void sendUploadNumbersInd(const std::string &numFile,
                                  const Uvr::Strings &deltas);
        void sendUploadNodesInd(const Uvr::Strings &files);

        void sendUploadOperatorsInd(const Uvr::Strings &files);

        void sendNotifyCallInfoReq(const Uvr::UvrId_t uvrId,
                                   const Uvr::Strings &files);

        void sendUploadPublicKeys(const Uvr::Strings &hubFilesToCopy,
                                  const Uvr::Strings &nodeFilesToCopy,
                                  const Uvr::Strings &hubFilesToRemove,
                                  const Uvr::Strings &nodeFilesToRemove);

        /// @brief Archives the given report local file, sends it via SFTP to a remote directory
        ///        on CCN, and then deletes the local file.
        /// @param[in] localFilePath Path to a local file.
        /// @param[in] ReportType    The type of report to send.
        void sendReportToCcn(const std::string &localFilePath,
                             const ReportType reportType);

        /// @brief Removes Call Info Request file based on the Call Info Response file name.
        /// @param[in] callInfoResponseFilePath path to a Call Info Response file.
        /// @returns @c true in case of success, @c false otherwise.
        bool removeCallInfoRequestFile(const std::string &callInfoResponseFilePath) const;

        void refreshNumbersFiles(void);
        void refreshNodesFiles(void);
        void refreshOperatorsFiles();
        void refreshCallInfoRequestsFiles(void);
        void refreshPublicKeys(void);
        void loadNumbersFiles(void);
        void loadNodesFiles(void);
        void loadOperatorsFiles();
        void loadPublicKeys(void);
        void resendReports();

        void getReportDirs(const ReportType reportType,
                           std::string &unsentDir,
                           std::string &remoteDir) const;

    protected:
        transaction::manager_novalue m_incidentsTransMngr;
        pm::manager m_pmMngr;
        sftp::client m_sftpClient;
        fs::local m_fsLocal;
        timers::id m_getIncidentsTimerId;
        timers::id m_refreshNumbersTimerId;
        timers::id m_refreshNodesTimerId;
        timers::id m_refreshOperatorsTimerId;
        timers::id m_refreshCallInfoRequestsId;
        timers::id m_refreshPubKeysTimerId;
        timers::id m_resendReportsTimerId;

        void handleSpecificPrimitive(const primitive::GenericPtr &prim) override;
        void handle_timeout(const ::timers::id &id,
                            const ::timers::time_point &startTime) override;

    public:
        explicit node(const TypeInst &ownInst,
                      boost::asio::io_context &ioc,
                      const configuration::manager &config);
        void start(void) override;
        void stop(void) override;
    };

    using api = workers::api<node, workers::type::CCNGW, workers::max_num_of_insts>;

} // namespace ccngw
