#include "connection.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// Boost includes.
#include <boost/bind/bind.hpp>
#include <boost/endian/conversion.hpp>

namespace connections
{
    connection::connection(boost::asio::io_context &ioc,
                           connections::type type,
                           const TypeInst &owner,
                           int connId) : m_type(type),
                                         m_owner(owner),
                                         m_peerInst(workers::type::ANY, 0),
                                         m_state(CONNECTION_INIT),
                                         m_connId(connId),
                                         m_strand(boost::asio::make_strand(ioc.get_executor()))
    {
    }
    connection::connection(boost::asio::io_context &ioc,
                           connections::type type,
                           const TypeInst &owner,
                           const TypeInst &inst,
                           std::string address, std::string port,
                           int connId) : m_type(type),
                                         m_owner(owner),
                                         m_peerInst(inst),
                                         m_state(CONNECTION_INIT),
                                         m_connId(connId),
                                         m_peerAddress(address),
                                         m_peerPort(port),
                                         m_strand(boost::asio::make_strand(ioc.get_executor()))
    {
    }

    connection_ptr connection::createIncomingConnection(boost::asio::io_context &ioc,
                                                        const TypeInst &owner,
                                                        int id)
    {
        connection_ptr ptr(new connection(ioc, INCOMING, owner, id));
        return ptr;
    }

    connection_ptr connection::createOutgoingConnection(boost::asio::io_context &ioc,
                                                        const TypeInst &owner,
                                                        const TypeInst &inst,
                                                        connections::type type,
                                                        const std::string &address,
                                                        const std::string &port,
                                                        int id)
    {

        connection_ptr ptr(new connection(ioc, type, owner, inst, address, port, id));
        return ptr;
    }

    tcp_socket_ptr connection::socket()
    {
        return m_socket;
    }

    connections::buffer &connection::buffer()
    {
        return m_buffer;
    }

    type connection::type() const
    {
        return m_type;
    }

    state connection::state() const
    {
        return m_state;
    }

    const TypeInst &connection::peerInstance() const
    {
        return m_peerInst;
    }

    const std::string &connection::peerAddress() const
    {
        return m_peerAddress;
    }

    void connection::peerAddress(const std::string &address)
    {
        m_peerAddress = address;
    }

    const std::string &connection::peerPort() const
    {
        return m_peerPort;
    }

    void connection::peerPort(const std::string &port)
    {
        m_peerPort = port;
    }

    int connection::id() const
    {
        return m_connId;
    }

    void connection::set_peerInstance(TypeInst inst)
    {
        m_peerInst = inst;
    }

    void connection::change_state(connections::state state)
    {
        LOGDEBUG(m_owner, " Change state ", connState2str(m_state), "->", connState2str(state), " for ", *this);
        m_state = state;
    }

    void connection::start_wait()
    {
        boost::asio::async_read(*m_socket, boost::asio::buffer(m_buffer),
                                boost::asio::transfer_exactly(sizeof(uint16_t)),
                                boost::bind(&connection::readMessageSize, shared_from_this(),
                                            boost::asio::placeholders::error,
                                            boost::asio::placeholders::bytes_transferred));
    }

    void connection::send(const primitive::raw_data_t &msg)
    {
        if (m_state < CONNECTION_SEND_WHOAMI)
        {
            LOGERROR(m_owner, " Cannot send message via ", *this, ": wrong state=", connState2str(m_state));
            return;
        }

        // TODO: need to check do we really needed async_write here
        boost::system::error_code err;
        boost::asio::write(*m_socket, boost::asio::buffer(msg), err);
        if (err && err != boost::asio::error::operation_aborted)
        {
            LOGINFO(m_owner, " Lost ", *this, ": ", err.message());
            reset();
            interfaces::controller::client1::handle_brokenConnection(m_peerInst);
        }
    }

    void connection::disconnect()
    {
        boost::asio::post(m_strand,
                          std::bind(&connection::disconnect_int, shared_from_this()));
    }

    void connection::disconnect_int()
    {
        LOGDEBUG(m_owner, " Disconnect ", *this);

        boost::system::error_code err;
        if (m_state > CONNECTION_DOWN)
        {
            change_state(CONNECTION_DOWN);
            m_socket->shutdown(boost::asio::ip::tcp::socket::shutdown_both, err);
            if (err)
            {
                LOGERROR(m_owner, " Cannot close ", *this, ": ", err.message());
            }
        }
    }

    void connection::reset()
    {
        change_state(CONNECTION_INIT);
        m_socket.reset(new tcp_socket(m_strand));
    }

    void connection::readMessageSize(const boost::system::error_code &errorCode,
                                  std::size_t bytesTransferred)
    {
        if (!errorCode)
        {
            const uint16_t *messageSize = reinterpret_cast<const uint16_t *>(m_buffer.data());

            // Read the primitive body.
            boost::asio::async_read(*m_socket, boost::asio::buffer(m_buffer),
                                    boost::asio::transfer_exactly(*messageSize),
                                    boost::bind(&connection::readMessage, shared_from_this(),
                                                boost::asio::placeholders::error,
                                                boost::asio::placeholders::bytes_transferred,
                                                *messageSize));
        }
        else if (errorCode != boost::asio::error::operation_aborted)
        {
            LOGINFO(m_owner, " Lost ", *this, ": ", errorCode.message());
            reset();
            interfaces::controller::client1::handle_brokenConnection(m_peerInst);
        }
        else
        {
            LOGINFO(m_owner, " Aborted ", *this, ": ", errorCode.message());
        }
    }

    void connection::readMessage(const boost::system::error_code &errorCode,
                                 std::size_t bytesTransferred,
                                 uint16_t messageSize)
    {
        if (!errorCode)
        {
            primitive::raw_data_t message(m_buffer.data(), messageSize);

            // Notify about new primitive upper layer.
            interfaces::controller::client1::handleMessage(shared_from_this(), message);

            // Waiting for the new primitive data from peer.
            start_wait();
        }
        else if (errorCode != boost::asio::error::operation_aborted)
        {
            LOGINFO(m_owner, " Lost ", *this, ": ", errorCode.message());
            reset();
            interfaces::controller::client1::handle_brokenConnection(m_peerInst);
        }
        else
        {
            LOGINFO(m_owner, " Aborted ", *this, ": ", errorCode.message());
        }
    }

    std::string connType2str(connections::type type)
    {
        switch (type)
        {
        case INCOMING:
            return "INCOMING";
        case OUTGOING_DYNAMICIP:
            return "OUTGOING";
        case OUTGOING_STATICIP:
            return "OUTGOING";
        default:
            return "UNKWN(" + std::to_string(type) + ")";
        }
    }

    std::string connState2str(connections::state state)
    {
        switch (state)
        {
        case CONNECTION_INIT:
            return "INIT";
        case CONNECTION_STARTED:
            return "STARTED";
        case CONNECTION_SEND_WHOAMI:
            return "SEND_WHOAMI";
        case CONNECTION_WAIT_WHOAMI:
            return "WAIT_WHOAMI";
        case CONNECTION_DOWN:
            return "DOWN";
        case CONNECTION_ESTBLISHED:
            return "ESTABLISHED";
        default:
            return "UNKWN(" + std::to_string(state) + ")";
        }
    }

    std::ostream &operator<<(std::ostream &os, const connection &c)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);
        logger << "conn[id:" << c.id() << ":" << c.m_owner;
        if (c.m_type != INCOMING)
        {
            logger << "->";
        }
        if (c.m_type == INCOMING)
        {
            logger << "<-";
        }
        logger << c.m_peerInst << ";" << connState2str(c.m_state);
        if (c.m_type == OUTGOING_STATICIP)
        {
            logger << ";ep=" << c.m_peerAddress << ":" << c.m_peerPort;
        }
        logger << "]";
        os << trace.str();
        return os;
    }
}
