#include "config/src/config_manager.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "components/tools/src/Tools.hpp"

namespace
{
    constexpr configuration::ConfigFileVersion_t s_configFileVersion = 1;
}

namespace configuration
{
    std::ostream &operator<<(std::ostream &os, const errors &err)
    {
        switch (err)
        {
        case errors::OK:
            os << "OK";
            break;
        case errors::FAIL:
            os << "FAIL";
            break;
        }
        return os;
    }

    ConfigFileVersion_t getConfigFileVersion()
    {
        return s_configFileVersion;
    }

    errors manager::loadConfigFile(const std::string &filename)
    {
        if (Json::Errors::OK != Json::parseFile(filename, m_json_config))
        {
            LOGERROR("Configuration file was not loaded "
                     "or parsed: file=\"",
                     filename, "\"");
            return errors::FAIL;
        }

        LOGDEBUG("Configuration file was successfully loaded "
                 "and parsed: file=\"",
                 filename, "\"");

        return errors::OK;
    }

    errors manager::validateConfigParams()
    {
        ConfigFileVersion_t configFileVersion = getValue<ConfigFileVersion_t>(CONFIG_FILE_VERSION, 0);
        if (configFileVersion < s_configFileVersion)
        {
            LOGE("Configuration file version ", configFileVersion, " is obsolete, actual version ",
                 s_configFileVersion, " should be used instead.");
            return errors::FAIL;
        }

        errors result = errors::OK;

        if (getValue<Uvr::UvrId_t>(UVR_ID, Uvr::UvrId::undefined) == Uvr::UvrId::undefined)
        {
            result = errors::FAIL;
        }

        if (getValue<std::string>(CCN_IP_ADDRESS, "").empty())
        {
            result = errors::FAIL;
        }

        if (getValue<std::string>(CCN_SFTP_USER, "").empty())
        {
            result = errors::FAIL;
        }

        return result;
    }

    boost::json::value manager::getJsonValue(const std::string &path) const
    {
        boost::json::value val;
        try
        {
            val = m_json_config.at_pointer(path);
        }
        catch (const boost::json::system_error &ex)
        {
            LOGERROR("Parameter is not found: path=\"", path, "\": ", ex.what());
        }
        return val;
    }

    namespace utils
    {
        timers::duration convertToDuration(std::optional<int> duration,
                                           std::optional<std::string> unit)
        {
            if (duration == std::nullopt || unit == std::nullopt)
            {
                LOGERROR("Cannot convert to duration: bad arguments [period=",
                         duration.value_or(0), ", unit=",
                         unit.value_or("none"), "]");
                return timers::duration::zero();
            }

            if (unit.value().compare("msec") == 0)
            {
                return std::chrono::milliseconds(*duration);
            }
            else if (unit.value().compare("sec") == 0)
            {
                return std::chrono::seconds(*duration);
            }
            else if (unit.value().compare("min") == 0)
            {
                return std::chrono::minutes(*duration);
            }
            else if (unit.value().compare("hr") == 0)
            {
                return std::chrono::hours(*duration);
            }
            else
            {
                std::string err = "Cannot convert to duration: unknown unit type=" + *unit;
                throw std::invalid_argument(err.c_str());
            }
        }
    } // namespace utils
} // namespace configuration

namespace timers
{
    Duration tag_invoke(boost::json::value_to_tag<Duration>,
                        boost::json::value const &jv)
    {
        std::optional<std::string> unit = std::nullopt;
        std::optional<int> value = std::nullopt;

        if (jv.is_object())
        {
            const boost::json::object &dataObject = jv.get_object();
            unit = boost::json::value_to<std::string>(dataObject.at("unit"));
            value = boost::json::value_to<int>(dataObject.at("value"));
        }
        else
        {
            throw std::invalid_argument("There is no specified JSON parameter.");
        }
        return Duration(::configuration::utils::convertToDuration(value, unit));
    }
}
