#pragma once

// AFAP includes.
#include "components/timers/src/timer_types.hpp"
#include "components/types/src/UvrTypes.hpp"
#include "components/log/src/Log.hpp"

// Boost includes.
#include <boost/function.hpp>
#include <boost/lexical_cast.hpp>

// STL includes.
#include <fstream>

namespace Csv
{

    const std::string new_line = "\r\n";
    const std::string delimiter = ";";
    const std::string no_value = "";
    const char FileTsFormat[] = "%Y_%m_%d_%H_%M_%S";

    namespace Incident
    {
        const std::vector<std::string> header{"NUM_A", "NUM_B", "NUM_D", "NUM_C", "DATE", "ID_REL", "RLC", "ID_SRC", "ID_UVR_T", "CALL_ID"};

        inline std::string prefix(const Uvr::UvrId_t &uvrId)
        {
            return "INCID_" + std::to_string(uvrId);
        }

    } // namespace Incident

    namespace IncidentA
    {
        const std::vector<std::string> header{"NUM_A", "NUM_B", "NUM_C", "DATE", "ID_REL", "RLC", "ID_UVR_O"};

        inline std::string prefix(const Uvr::UvrId_t &uvrId)
        {
            return "INCID_A_" + std::to_string(uvrId);
        }

    } // namespace IncidentA

    namespace CallInfoResponse
    {
        const std::vector<std::string> header{"NUM_A", "NUM_B", "NUM_D", "NUM_C", "DATE", "DATE_ACT",
                                              "ID_REQ", "ID_SRC", "ID_DST", "RSP_CODE", "VRF_RSP",
                                              "SESSION_ID", "T_ACTION_CODE", "RELEASE_CODE",
                                              "DURATION", "ID_UVR_T", "CALL_ID", "ID_REL"};

        inline std::string prefix(const Uvr::UvrId_t &uvrId, const Uvr::RequestId_t &reqId)
        {
            return "RSP_" + std::to_string(uvrId) + "_" + std::to_string(reqId);
        }
    } // namespace CallInfoResponse

    class File
    {
    public:
        template <typename ValueType>
        struct OptionalValue
        {
            ValueType &value;
            const ValueType &defaultValue;
        };

        /// @brief Generates CSV file
        /// @tparam Record Type of data, which should be written into CSV file
        /// @param[in] prefix The prefix of generated CSV file
        /// @param[in] header The list of CSV header fields
        /// @param[in] records The data, which should be written into CSV file
        /// @param[out] filename The filename of generated CSV file
        /// @return @c true if CSV file was generated,
        ///         @c false if CSV file failed to be generated.
        template <class Record>
        static bool generate(const std::string &prefix,
                             const std::vector<std::string> &header,
                             const std::vector<Record> &records,
                             std::string &filename)
        {
            timers::time_point currTime(timers::now());
            filename = prefix + "_" + currTime.utcFormat(FileTsFormat) + ".csv";
            LOGD("Generates CSV file START: file=\"", filename, "\"");
            std::ofstream out(filename);
            if (!out.is_open())
            {
                LOGERROR("Cannot generate CSV file: cannot create file=\"", filename);
                return false;
            }
            if (header.empty())
            {
                LOGERROR("Cannot generate CSV file: header is empty for file=\"", filename);
                return false;
            }
            // Print header
            auto hIter = header.begin();
            out << *hIter;
            ++hIter;
            for (; hIter != header.end(); hIter++)
            {
                out << delimiter << *hIter;
            }
            out << new_line;

            // Print records
            for (auto rec : records)
            {
                out << rec << new_line;
            }
            out.close();
            LOGD("Generates CSV file DONE: file=\"", filename, "\"");
            return true;
        }

        /// @brief Parses CSV file and calls specified handler per each row.
        /// @param[in] filename     The file to parse.
        /// @param[in] rowHandler   The function to process a row. Should return @c true if the parsing should be
        ///                         interrupted, @c false if it can be continued. In case of parsing errors an
        ///                         exception should be thrown.
        /// @param[in] ignoreErrors If @c true, then the parsing will continue even if the @a rowHandler throws
        ///                         an exception. Doesn't affect the return value of @a rowHandler, so if it returns
        ///                         @c true the parsing will be interrupted anyway. @c true by default.
        /// @returns @c true if parsing was successful, @c false if parsing failed.
        static bool parse(const std::string &filename,
                          std::function<bool(const std::string &)> const &rowHandler,
                          bool ignoreErrors = true)
        {
            LOGDEBUG("START parsing file=", filename);
            std::ifstream inFile(filename);
            if (!inFile.is_open())
            {
                LOGERROR("Can't open the file=\"", filename, "\"");
                return false;
            }

            size_t rowIndex = 1;
            std::string rowStr;
            getline(inFile, rowStr); // Skip first row with the header.
            while (getline(inFile, rowStr))
            {
                ++rowIndex;
                if (rowStr.empty())
                {
                    continue;
                }

                std::size_t pos = rowStr.rfind("\r");
                if (pos == (rowStr.length() - 1))
                {
                    rowStr.erase(pos, 1); // delete "\r" symbol if it presents
                }

                try
                {
                    bool shouldInterrupt = rowHandler(rowStr);
                    if (shouldInterrupt)
                    {
                        break;
                    }
                }
                catch (const std::exception &ex)
                {
                    LOGERROR("Can't parse the file=\"", filename, "\".",
                             " Error on row ", rowIndex, ": ", ex.what());
                    if (!ignoreErrors)
                    {
                        inFile.close();
                        return false;
                    }
                }
                catch (...)
                {
                    LOGERROR("Can't parse the file=\"", filename, "\".",
                             " Error on row ", rowIndex, ".");
                    if (!ignoreErrors)
                    {
                        inFile.close();
                        return false;
                    }
                }
            }

            inFile.close();
            LOGDEBUG("DONE parsing file=", filename);
            return true;
        }

        static bool parseZip(const std::string &filename,
                             std::function<bool(std::string &&)> const &rowHandler,
                             bool ignoreErrors = true)
        {
            LOGDEBUG("START parsing zip file=", filename);

            constexpr int maxOutputLineSize = 8192;
            std::string cmd = "unzip -qq -c " + filename;

            std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(cmd.c_str(), "r"), &pclose);
            std::array<char, maxOutputLineSize> buffer;
            if (!pipe)
            {
                LOGERROR("Can't open the file=\"", filename, "\"");
                return false;
            }

            size_t rowIndex = 1;
            std::string rowStr;
            fgets(buffer.data(), buffer.size(), pipe.get()); // Skip first row with the header.
            while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr)
            {
                std::string rowStr(buffer.data(), std::strlen(buffer.data()) - 1);

                ++rowIndex;
                if (rowStr.empty())
                {
                    continue;
                }

                std::size_t pos = rowStr.rfind("\r");
                if (pos == (rowStr.length() - 1))
                {
                    rowStr.erase(pos, 1); // delete "\r" symbol if it presents
                }

                try
                {
                    bool shouldInterrupt = rowHandler(std::move(rowStr));
                    if (shouldInterrupt)
                    {
                        break;
                    }
                }
                catch (const std::exception &ex)
                {
                    LOGERROR("Can't parse zip file=\"", filename, "\".",
                             " Error on row ", rowIndex, ": ", ex.what());
                    if (!ignoreErrors)
                    {
                        return false;
                    }
                }
                catch (...)
                {
                    LOGERROR("Can't parse zip file=\"", filename, "\".",
                             " Error on row ", rowIndex, ".");
                    if (!ignoreErrors)
                    {
                        return false;
                    }
                }
            }

            LOGDEBUG("DONE parsing zip file=", filename);
            return true;
        }

    private:
        template <typename ColumnType>
        static void parseColumn(const std::string &columnStr,
                                ColumnType &columnValue)
        {
            if (columnStr.empty())
            {
                throw std::invalid_argument("Value can't be optional");
            }
            columnValue = boost::lexical_cast<ColumnType>(columnStr);
        }

        template <typename ColumnType>
        static void parseColumn(const std::string &columnStr,
                                OptionalValue<ColumnType> &columnValue)
        {
            if (columnStr.empty())
            {
                columnValue.value = columnValue.defaultValue;
            }
            else
            {
                columnValue.value = boost::lexical_cast<ColumnType>(columnStr);
            }
        }

        template <typename ColumnType>
        static void parseColumn(const std::string &columnStr,
                                std::optional<ColumnType> &columnValue)
        {
            if (columnStr.empty())
            {
                columnValue.reset();
            }
            else
            {
                columnValue.emplace(boost::lexical_cast<ColumnType>(columnStr));
            }
        }

        static void parseColumn(const std::string &columnStr,
                                std::string &columnValue)
        {
            if (columnStr.empty())
            {
                throw std::invalid_argument("Value can't be optional");
            }
            columnValue = columnStr;
        }

        static void parseRowHelper(const char *& /*rowPtr*/,
                                   const char & /*delimeter*/,
                                   const size_t & /*columnIndex*/)
        {
            // To stop parsing when all arguments for columns run out.
        }

        template <typename ColumnType, typename... ColumnsTypes>
        static void parseRowHelper(const char *&rowPtr,
                                   const char &delimeter,
                                   const size_t &columnIndex,
                                   ColumnType &columnValue,
                                   ColumnsTypes &...columnsValues)
        {
            const char *delimeterPtr = std::strchr(rowPtr, delimeter);
            std::string columnStr = "";
            if (delimeterPtr == nullptr)
            {
                // No more columns can be found.
                // Use what's left for the last value.
                columnStr = std::string(rowPtr);
                rowPtr += columnStr.length();
            }
            else
            {
                columnStr = std::string(rowPtr, (delimeterPtr - rowPtr));
                rowPtr = delimeterPtr + 1;
            }

            try
            {
                parseColumn(columnStr, columnValue);
            }
            catch (const std::exception &ex)
            {
                throw std::invalid_argument(std::string("Error on column ") + std::to_string(columnIndex) + ": " + ex.what());
            }

            parseRowHelper(rowPtr, delimeter, columnIndex + 1, columnsValues...);
        }

    public:
        /// @brief Parses a string into passed variables.
        /// @tparam ...ColumnsTypes Data types in the corresponding columns.
        /// @param[in]  rowPtr           Pointer to the null-terminated byte string to be parsed.
        /// @param[in]  delimeter        Delimeter character between columns.
        /// @param[out] ...columnsValues References to the variables where the parsed values will be stored.
        template <typename... ColumnsTypes>
        static void parseRow(const char *rowPtr,
                             const char &delimeter,
                             ColumnsTypes &...columnsValues)
        {
            parseRowHelper(rowPtr, delimeter, 1, columnsValues...);
        }
    };
} // namespace Csv
