#include "components/db/src/ClickHouse.hpp"

// AFAP includes.
#include "components/db/src/ConfigParameters.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "TRecQ763Number.hpp"
#include "UvrTypes.hpp"

// Boost includes.
#include <boost/lexical_cast.hpp>


namespace
{
    constexpr int s_expectedSchemaVersion = 1;
    const std::string s_versionsTableName = "__versions__";
    const std::string s_schemaVersionColumnName = "schema_version";
    const std::string s_schemaVersionColumn = s_schemaVersionColumnName + " Int32";
    const std::string s_appliedAtColumnName = "applied_at";
    const std::string s_appliedAtColumn = s_appliedAtColumnName + " DateTime DEFAULT now()";

    const bool RestartTimer = true;
    const bool NotRestartTimer = false;

    const std::string s_sqlNullValue = "null";

} // namespace

namespace db
{
    ClickHouse::ClickHouse(const Uvr::UvrId_t uvrId,
                           const configuration::manager &config)
        : m_uvrId(uvrId)
    {
        setSettings(config);
        // Verdicts
        this->db::interfaces::tables::calls::server::tableName = db::Calls::tableName;
        this->db::interfaces::tables::calls::server::tableFields = "NUM_A String not null, "
                                                                   "NUM_B String not null, "
                                                                   "NUM_D String not null, "
                                                                   "NUM_C String not null, "
                                                                   "DATE DateTime64(9) not null, "
                                                                   "DATE_ACT DateTime64(0) not null, "
                                                                   "ID_SRC UInt32 null, "
                                                                   "ID_DST UInt32 null, "
                                                                   "SESSION_ID UInt64 not null, "
                                                                   "ID_REL UInt8 not null, "
                                                                   "VRF_RSP Int16 not null, "
                                                                   "RELEASE_CODE UInt8 null, "
                                                                   "DURATION UInt16 null, "
                                                                   "ID_UVR_T UInt16 null, "
                                                                   "CALL_ID String not null, "
                                                                   "CALL_TYPE UInt8 not null, "
                                                                   "INDEX NUM_A_IX(CALL_ID) TYPE set(0) GRANULARITY 8192";

        this->db::interfaces::tables::calls::server::tableSettings = "ENGINE = ReplacingMergeTree() "
                                                                     "PARTITION BY toUInt64(formatDateTime(DATE, '%Y%m%d%H')) "
                                                                     "ORDER BY (DATE,NUM_A,NUM_B,CALL_TYPE) "
                                                                     "TTL toDateTime(DATE) + toIntervalYear(1) + toIntervalHour(1) DELETE "
                                                                     "SETTINGS merge_with_ttl_timeout=3600";

        this->db::interfaces::tables::calls::server::selectSettings = "";

        // Incidents
        this->db::interfaces::tables::incidents::server::tableName = db::incident::tableName;
        // IncidentsA
        this->db::interfaces::tables::incidentsA::server::tableName = db::incidentA::tableName;
        // Subscribers
        this->db::interfaces::tables::subscribers::server::tableName = db::subscriber::tableName;
        // UVRs
        this->db::interfaces::tables::uvrs::server::tableName = db::uvr::tableName;
        // HUBs
        this->db::interfaces::tables::hubs::server::tableName = db::hub::tableName;
    }

    void ClickHouse::setSettings(const configuration::manager &config)
    {
        // TODO[multi-uvr]: Use the proper logic to obtain the correct settings
        //                  section for a given UVR.

        m_clientPtr.reset();
        m_db = config.getValue<std::string>(Config::DB::ClckH::Database, "");
        if (m_db.empty())
        {
            LOGW("Database for Clickhouse DB is not configured, set default: database=\'",
                 Config::DB::ClckH::DefaultDatabase, "\'");
            m_db = Config::DB::ClckH::DefaultDatabase;
        }

        // Append UVR ID to database name.
        m_db += "_" + std::to_string(m_uvrId);

        m_user = config.getValue<std::string>(Config::DB::ClckH::User, "");
        if (m_user.empty())
        {
            LOGW("User for Clickhouse DB is not configured, set default: user=\'",
                 Config::DB::ClckH::DefaultUser, "\'");
            m_user = Config::DB::ClckH::DefaultUser;
        }
        m_password = config.getValue<std::string>(Config::DB::ClckH::Password, "");
        if (m_password.empty())
        {
            LOGW("Password for Clickhouse DB is not configured, leave it empty");
        }
        m_host = config.getValue<std::string>(Config::DB::ClckH::Host, "");
        if (m_host.empty())
        {
            LOGW("Host for Clickhouse DB is not configured, set default: host=\'",
                 Config::DB::ClckH::DefaultHost, "\'");
            m_host = Config::DB::ClckH::DefaultHost;
        }
        std::string portStr = config.getValue<std::string>(Config::DB::ClckH::Port,
                                                           Config::DB::ClckH::DefaultPort);
        try
        {
            m_port = boost::lexical_cast<int>(portStr);
        }
        catch (const std::exception &ex)
        {
            LOGW("Invalid port string=\"", portStr,
                 "\", set default: port=", Config::DB::ClckH::DefaultPort);
            m_port = boost::lexical_cast<int>(Config::DB::ClckH::DefaultPort);
        }
    }

    ::clickhouse::Client *ClickHouse::getClient()
    {
        if (!m_clientPtr.get())
        {
            ::clickhouse::ClientOptions options;
            options.SetDefaultDatabase(m_db);
            options.SetUser(m_user);
            if (!m_password.empty())
            {
                options.SetPassword(m_password);
            }
            options.SetHost(m_host);
            options.SetPort(m_port);
            options.TcpKeepAlive(true);
            m_clientPtr.reset(new ::clickhouse::Client(options));
        }
        return m_clientPtr.get();
    }

    bool ClickHouse::connect(void)
    {
        try
        {
            getClient();
        }
        catch (const std::exception &ex)
        {
            LOGE("Failed to connect to ClickHouse DB: ", ex.what());
            return false;
        }
        return true;
    }

    void ClickHouse::disconnect(void)
    {
        m_clientPtr.reset();
    }

    int ClickHouse::getActualSchemaVersion()
    {
        if (!is_exist(s_versionsTableName))
        {
            return -1;
        }

        std::string query =
            "SELECT MAX(" + s_schemaVersionColumnName + ") "
            + "FROM " + s_versionsTableName;

        int version = -1;
        auto selectCb = [&version](const ::clickhouse::Block &block)
        {
            if (block.GetRowCount() > 0 && block.GetColumnCount() > 0)
            {
                version = block[0]->As<::clickhouse::ColumnInt32>()->At(0);
            }
        };

        exec<Select>(query, selectCb);

        return version;
    }

    int ClickHouse::getExpectedSchemaVersion() const
    {
        return s_expectedSchemaVersion;
    }

    void ClickHouse::putSchemaVersion(const int version)
    {
        if (!is_exist(s_versionsTableName))
        {
            std::string query =
                "CREATE TABLE IF NOT EXISTS " + s_versionsTableName
                + " ("
                + s_schemaVersionColumn + ", "
                + s_appliedAtColumn
                + ") "
                + "ENGINE = MergeTree "
                + "ORDER BY tuple(" + s_appliedAtColumnName + ");";
            exec<Query>(query);
        }

        std::string query =
            "INSERT INTO " + s_versionsTableName
            + " VALUES (" + std::to_string(getExpectedSchemaVersion()) + ", DEFAULT);";
        exec<Query>(query);
    }

    // INTERFACE TABLES
    void ClickHouse::create(const ::db::table::name &tbName,
                            const ::db::table::fields &tbFields,
                            const ::db::table::settings &tbSettings)
    {
        std::string query = "create table if not exists " + tbName + "(" + tbFields + ") " + tbSettings + ";";
        exec<Query>(query);
    }
    void ClickHouse::drop(const ::db::table::name &tbName)
    {
        std::string query = "drop table if exists " + tbName;
        exec<Query>(query);
    }

    ::db::limit_t ClickHouse::size(const ::db::table::name &tbName)
    {
        std::string query = "select count(*) from " + tbName;
        ::db::limit_t tSize = 0;
        auto selectCb = [&tSize](const ::clickhouse::Block &block)
        {
            if (block.GetRowCount() > 0 && block.GetColumnCount() > 0)
            {
                tSize = block[0]->As<::clickhouse::ColumnUInt64>()->At(0);
            }
        };
        exec<Select>(query, selectCb);
        return tSize;
    }

    bool ClickHouse::is_exist(const ::db::table::name &tbName)
    {
        std::string query = "exists table " + tbName;
        uint8_t result = 0;
        auto selectCb = [&result](const ::clickhouse::Block &block)
        {
            if (block.GetRowCount() > 0 && block.GetColumnCount() > 0)
            {
                result = block[0]->As<::clickhouse::ColumnUInt8>()->At(0);
            }
        };
        exec<Select>(query, selectCb);
        return result;
    }

    // INTERFACE VERDICTS
    void ClickHouse::put(const Uvr::CallRecord &rec)
    {
        if (rec.callType == Uvr::CallType::undefined)
        {
            LOGW("Failed to insert record=\"", rec, "\": undefined Call Type");
            return;
        }

        std::string query = "insert into " + this->db::interfaces::tables::calls::server::tableName + " (*) values ("
                            + "\'" + rec.numA + "\', "
                            + "\'" + rec.numB + "\', "
                            + "\'" + rec.numD + "\', "
                            + "\'" + rec.numC + "\', "
                            + std::to_string(rec.date.tv_sec * timers::giga + rec.date.tv_nsec) + ", "
                            + std::to_string(rec.dateAct.tv_sec) + ", "
                            + Uvr::toString(rec.idSrc, s_sqlNullValue) + ", "
                            + Uvr::toString(rec.idDst, s_sqlNullValue) + ", "
                            + std::to_string(rec.sessionId) + ", "
                            + std::to_string(rec.idRel) + ", "
                            + std::to_string(rec.response) + ", "
                            + Uvr::toString(rec.releaseCode, s_sqlNullValue) + ", "
                            + Uvr::toString(rec.callDuration, s_sqlNullValue) + ", "
                            + Uvr::toString(rec.idUvrT, s_sqlNullValue) + ", "
                            + "\'" + rec.callId + "\', "
                            + std::to_string(rec.callType) + ")";
        exec<Query>(query);
    }

    bool ClickHouse::update(const Uvr::CallRecord &rec)
    {
        if (rec.callId.empty()
            || rec.callType == Uvr::CallType::undefined)
        {
            return false;
        }

        size_t numOfFieldsForUpdate = 0;
        std::string query = std::string("alter table ")
                            + this->db::interfaces::tables::calls::server::tableName
                            + std::string(" update");

        if (rec.idSrc.has_value())
        {
            query.append(" ID_SRC = " + std::to_string(rec.idSrc.value()) + " ,");
            ++numOfFieldsForUpdate;
        }
        if (rec.idDst.has_value())
        {
            query.append(" ID_DST = " + std::to_string(rec.idDst.value()) + " ,");
            ++numOfFieldsForUpdate;
        }
        if (rec.releaseCode.has_value())
        {
            query.append(" RELEASE_CODE = " + std::to_string(static_cast<int>(rec.releaseCode.value())) + " ,");
            ++numOfFieldsForUpdate;
        }
        if (rec.callDuration.has_value())
        {
            query.append(" DURATION = " + std::to_string(rec.callDuration.value()) + " ,");
            ++numOfFieldsForUpdate;
        }

        if (numOfFieldsForUpdate == 0)
        {
            return false;
        }

        query.pop_back(); // Delete last ','
        query.append("where CALL_ID = \'" + rec.callId + "\' AND CALL_TYPE = " + std::to_string(rec.callType) + ";");
        bool res = exec<Query>(query);
        return res;
    }

    void ClickHouse::get(Uvr::CallInformationResponses &rsps,
                         const Uvr::CallInformationRequest &req)
    {
        ::timers::time_point date(req.date);
        Uvr::CallMatchInterval_t interval = req.interval.value_or(Uvr::CallMatchInterval::def);
        std::string query = "select NUM_D,NUM_C,DATE,DATE_ACT,ID_SRC,ID_DST,VRF_RSP,SESSION_ID,RELEASE_CODE,DURATION,ID_UVR_T,CALL_ID,ID_REL from "
                            + this->db::interfaces::tables::calls::server::tableName
                            + this->db::interfaces::tables::calls::server::selectSettings
                            + " where DATE > " + std::to_string(date.tv_sec - interval) // we need accuracy only up to seconds
                            + " and DATE <= " + std::to_string(date.tv_sec + interval)
                            + " and NUM_A = \'" + req.numA + "\'"
                            + " and NUM_B = \'" + req.numB + "\'";
        if (!req.numD.empty())
        {
            query.append(" and NUM_D = \'" + req.numD + "\'");
        }
        if (!req.numC.empty())
        {
            query.append(" and NUM_C = \'" + req.numC + "\'");
        }
        if (req.idSrc.has_value())
        {
            query.append(" and ID_SRC = " + std::to_string(req.idSrc.value()));
        }
        if (req.idDst.has_value())
        {
            query.append(" and ID_DST = " + std::to_string(req.idDst.value()));
        }
        if (!req.callId.empty())
        {
            query.append(" and CALL_ID = \'" + req.callId + "\'");
        }
        if (req.callDuration.has_value()
            && req.callDuration.value() > 0)
        {
            query.append(" and DURATION = \'" + std::to_string(req.callDuration.value()) + "\'");
        }
        query.append(" order by DATE");
        query.append(";");

        rsps.clear();
        auto selectCb = [this, &rsps, &req](const ::clickhouse::Block &block)
        {
            for (std::size_t rowIx = 0; rowIx < block.GetRowCount(); ++rowIx)
            {
                std::size_t clmnIx = 0;
                Uvr::CallInformationResponse rsp(req.numA, req.numB, "", "", "", "",
                                                 req.idReq,
                                                 req.idSrc,
                                                 req.idDst,
                                                 Uvr::ResponseCode::found,
                                                 Uvr::VerificationResponse::notVerified);
                rsp.numD = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                rsp.numC = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);

                uint64_t date = block[clmnIx++]->As<::clickhouse::ColumnDateTime64>()->At(rowIx);
                timers::time_point dateTimePoint;
                dateTimePoint.tv_sec = date / timers::giga;
                dateTimePoint.tv_nsec = 0;
                rsp.date = dateTimePoint.format();

                uint64_t dateAct = block[clmnIx++]->As<::clickhouse::ColumnDateTime64>()->At(rowIx);
                timers::time_point dateActTimePoint;
                dateActTimePoint.tv_sec = dateAct;
                dateActTimePoint.tv_nsec = 0;
                rsp.dateAct = dateActTimePoint.format();

                auto colNullIdSrc = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                if (!colNullIdSrc->IsNull(rowIx))
                {
                    rsp.idSrc = colNullIdSrc->Nested()->As<::clickhouse::ColumnUInt32>()->At(rowIx);
                }
                auto colNullIdDst = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                if (!colNullIdDst->IsNull(rowIx))
                {
                    rsp.idDst = colNullIdDst->Nested()->As<::clickhouse::ColumnUInt32>()->At(rowIx);
                }
                rsp.vrfRsp = block[clmnIx++]->As<::clickhouse::ColumnInt16>()->At(rowIx);
                rsp.sessionId = block[clmnIx++]->As<::clickhouse::ColumnUInt64>()->At(rowIx);
                auto colNullReleaseCode = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                if (!colNullReleaseCode->IsNull(rowIx))
                {
                    rsp.releaseCode = colNullReleaseCode->Nested()->As<::clickhouse::ColumnUInt8>()->At(rowIx);
                }
                auto colNullCallDur = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                if (!colNullCallDur->IsNull(rowIx))
                {
                    rsp.callDuration = colNullCallDur->Nested()->As<::clickhouse::ColumnUInt16>()->At(rowIx);
                }
                auto colNullIdUvrT = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                if (!colNullIdUvrT->IsNull(rowIx))
                {
                    rsp.idUvrT = colNullIdUvrT->Nested()->As<::clickhouse::ColumnUInt16>()->At(rowIx);
                }
                rsp.callId = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                rsp.idRel = block[clmnIx]->As<::clickhouse::ColumnUInt8>()->At(rowIx);
                rsps.push_back(rsp);
                LOGD("Got record from ", this->db::interfaces::tables::calls::server::tableName, " table: ", rsp);
            }
        };

        if (!exec<Select>(query, selectCb))
        {
            Uvr::CallInformationResponse rsp(req.numA,
                                             req.numB,
                                             req.numC,
                                             req.numD,
                                             req.date,
                                             req.date,
                                             req.idReq,
                                             std::nullopt,
                                             std::nullopt,
                                             Uvr::ResponseCode::systemError,
                                             std::nullopt,
                                             Uvr::undefinedSessionId,
                                             std::nullopt,
                                             std::nullopt,
                                             std::nullopt,
                                             std::nullopt,
                                             req.callId,
                                             Uvr::VerdictTrigger::undefined);
            rsps.push_back(rsp);
        }

        if (rsps.empty())
        {
            Uvr::CallInformationResponse rsp(req.numA,
                                             req.numB,
                                             req.numC,
                                             req.numD,
                                             req.date,
                                             req.date,
                                             req.idReq,
                                             req.idSrc,
                                             req.idDst,
                                             Uvr::ResponseCode::notFound,
                                             Uvr::VerificationResponse::undefined,
                                             Uvr::undefinedSessionId,
                                             std::nullopt,
                                             std::nullopt,
                                             std::nullopt,
                                             std::nullopt,
                                             req.callId,
                                             Uvr::VerdictTrigger::undefined);
            rsps.push_back(rsp);
        }
    }

#define INSERT_FLUSHER_TIMER_ID 1
#define UPDATE_FLUSHER_TIMER_ID 2
    ClickHouseAcc::ClickHouseAcc(boost::asio::io_context &ioc,
                                 const Uvr::UvrId_t uvrId,
                                 const configuration::manager &config)
        : ClickHouse(uvrId, config),
          m_ioc(ioc)
    {
        this->db::interfaces::tables::calls::server::selectSettings = " FINAL";
        m_insertBufferSize =
            config.getValue<int>(Config::DB::ClckH::InsertBufferSize,
                                 Config::DB::ClckH::DefaultInsertBufferSize);
        m_updateBufferSize =
            config.getValue<int>(Config::DB::ClckH::UpdateBufferSize,
                                 Config::DB::ClckH::DefaultUpdateBufferSize);
        m_insertBufferFlushTimer =
            std::make_shared<timers::periodic>(m_ioc, "DB_INSERT_FLUSHER",
                                               config.getValue(Config::DB::ClckH::InsertBufferFlushTimer,
                                                               timers::Duration(Config::DB::ClckH::DefaultInsertBufferFlushTimer)),
                                               timers::id(INSERT_FLUSHER_TIMER_ID));
        m_updateBufferFlushTimer =
            std::make_shared<timers::periodic>(m_ioc, "DB_UPDATE_FLUSHER",
                                               config.getValue(Config::DB::ClckH::UpdateBufferFlushTimer,
                                                               timers::Duration(Config::DB::ClckH::DefaultUpdateBufferFlushTimer)),
                                               timers::id(UPDATE_FLUSHER_TIMER_ID));
    }

    ClickHouseAcc::~ClickHouseAcc()
    {
        disconnect();
    }

    bool ClickHouseAcc::connect(void)
    {
        bool result = ClickHouse::connect();
        // Start timer for writting into DB
        if (result)
        {
            m_insertBufferFlushTimer->::timers::interfaces::manager::client1::bind(shared_from_this());
            m_insertBufferFlushTimer->start();
            m_updateBufferFlushTimer->::timers::interfaces::manager::client1::bind(shared_from_this());
            m_updateBufferFlushTimer->start();
        }
        return result;
    }

    void ClickHouseAcc::disconnect(void)
    {
        ::timers::interfaces::manager::server::stop_server();
        if (m_insertBufferFlushTimer)
        {
            m_insertBufferFlushTimer->stop();
            m_insertBufferFlushTimer.reset();
        }
        if (m_updateBufferFlushTimer)
        {
            m_updateBufferFlushTimer->stop();
            m_updateBufferFlushTimer.reset();
        }
        flushAllBuffers("disconnection from DB", NotRestartTimer);
        ClickHouse::disconnect();
    }

    void ClickHouseAcc::put(const Uvr::CallRecord &rec)
    {
        if (rec.callType == Uvr::CallType::undefined)
        {
            LOGW("Failed to insert record=\"", rec, "\": undefined Call Type");
            return;
        }

        std::lock_guard<std::recursive_mutex> guard(m_gInsert);
        BufKey bufKey(rec.callId, rec.callType);
        m_recToInsert[bufKey].push_back(rec);
        ++m_numOfInsertReqs;
        LOGD("ADDED TO INSERT BUFFER[", m_numOfInsertReqs, "/", m_insertBufferSize, "]: ", rec);

        // Flush records from buffer to DB
        if (m_numOfInsertReqs == m_insertBufferSize)
        {
            boost::asio::post(m_ioc,
                              std::bind(&ClickHouseAcc::flushInsertBuffer,
                                        shared_from_this(),
                                        "insert buffer overflow",
                                        RestartTimer));
        }
    }

    bool ClickHouseAcc::update(const Uvr::CallRecord &rec)
    {
        bool noCallId = rec.callId.empty();
        bool noCallType = (rec.callType == Uvr::CallType::undefined);
        if (noCallId || noCallType)
        {
            LOGW("Failed to update record=\"", rec, "\": ",
                 (noCallId ? "empty Call ID" : ""),
                 (noCallId && noCallType ? ", " : ""),
                 (noCallType ? "undefined Call Type" : ""));
            return false;
        }
        BufKey bufKey(rec.callId, rec.callType);

        // Check buffer with insert SQL records
        std::unique_lock<std::recursive_mutex> guardI(m_gInsert);
        auto bufIter = m_recToInsert.find(bufKey);
        uint32_t count = 0;
        if (bufIter != m_recToInsert.end())
        {
            count++;
            for (auto &bufRec : bufIter->second)
            {
                bufRec.update(rec);
                LOGD("CHANGED IN INSERT BUFFER: key[", bufKey.key,
                     "] \'", rec, "\' >> \'", bufRec, "\'");
            }
        }
        guardI.unlock();
        // Check buffer with update SQL records
        std::lock_guard<std::recursive_mutex> guardU(m_gUpdate);

        auto recIter = m_recToUpdate.find(bufKey);
        if (recIter != m_recToUpdate.end())
        {
            recIter->second.first.update(rec);
            LOGD("CHANGED IN UPDATE BUFFER: key[", bufKey.key,
                 "] \'", rec, "\' >> \'", recIter->second.first, "\'");
        }
        else
        {
            m_recToUpdate.emplace(bufKey, std::make_pair(rec, count));
            m_numOfUpdateReqs++;
            LOGD("ADDED TO UPDATE BUFFER[", m_numOfUpdateReqs, "/", m_updateBufferSize, "]: ", rec);
        }
        // Flush records from buffer to DB
        if (m_numOfUpdateReqs == m_updateBufferSize)
        {
            boost::asio::post(m_ioc,
                              std::bind(&ClickHouseAcc::flushUpdateBuffer,
                                        shared_from_this(),
                                        "update buffer overflow",
                                        RestartTimer));
        }
        return true;
    }

    void ClickHouseAcc::get(Uvr::CallInformationResponses &rsps,
                            const Uvr::CallInformationRequest &req)
    {
        flushAllBuffers("select", RestartTimer);
        ClickHouse::get(rsps, req);
    }

    void ClickHouseAcc::manage_timeout(const timers::id &id,
                                       const timers::time_point &start_time)
    {
        if (id == INSERT_FLUSHER_TIMER_ID)
        {
            flushInsertBuffer("insert timer expiration", NotRestartTimer);
        }
        else if (id == UPDATE_FLUSHER_TIMER_ID)
        {
            flushUpdateBuffer("update timer expiration", NotRestartTimer);
        }
        else
        {
            LOGW("Unknown ClickHouse DB timer: id=", id);
        }
    }

    void ClickHouseAcc::manage_stopped(const ::timers::id &id)
    {
        m_insertBufferFlushTimer.reset();
        m_updateBufferFlushTimer.reset();
        flushAllBuffers("stop of all timers", NotRestartTimer);
    }

    void ClickHouseAcc::flushInsertBuffer(const std::string &event,
                                          bool restartTimer)
    {
        std::lock_guard<std::recursive_mutex> guard(m_gInsert);
        if (m_recToInsert.empty())
        {
            LOGT("No SQL insert requests to flush on ", event);
        }
        else
        {
            LOGD("Flushing ", m_numOfInsertReqs, " insert records on ", event);
            std::string insertQuery("insert into "
                                    + this->db::interfaces::tables::calls::server::tableName
                                    + " (*) values ");
            bool notFirstValue = false;
            for (auto const &buf : m_recToInsert)
            {
                for (auto const &bufRec : buf.second)
                {
                    insertQuery += notFirstValue ? ", " : (notFirstValue = true, "");
                    std::string sqlValue;
                    toSqlValue(bufRec, sqlValue);
                    insertQuery += sqlValue;
                }
            }

            if (exec<Query>(insertQuery))
            {
                LOGD("Flushed ", m_numOfInsertReqs, " insert requests on ", event);
                m_recToInsert.clear();
                m_numOfInsertReqs = 0;
            }
            else
            {
                LOGW("Failed to insert ", m_numOfInsertReqs,
                     " records on ", event, ". Try again soon.");
            }
        }
        // Restart timer if needed
        if (restartTimer)
        {
            timers::restartTimer<timers::periodic>(m_insertBufferFlushTimer);
        }
    }

    void ClickHouseAcc::flushUpdateBuffer(const std::string &event,
                                          bool restartTimer)
    {
        std::lock_guard<std::recursive_mutex> guard(m_gUpdate);
        if (m_recToUpdate.empty())
        {
            LOGT("No update SQL requests to flush on ", event);
        }
        else
        {
            LOGD("Flushing ", m_numOfUpdateReqs, " update records on ", event);
            auto recIter = m_recToUpdate.begin();
            std::string selectQuery = "select * from "
                                      + this->db::interfaces::tables::calls::server::tableName
                                      + " FINAL where (CALL_TYPE = " + std::to_string(recIter->second.first.callType)
                                      + " AND CALL_ID = \'" + recIter->second.first.callId + "\')";
            ++recIter;
            for (; recIter != m_recToUpdate.end(); recIter++)
            {
                selectQuery += " or (CALL_TYPE = " + std::to_string(recIter->second.first.callType)
                               + " AND CALL_ID = \'" + recIter->second.first.callId + "\')";
            }

            auto selectCb = [this](const ::clickhouse::Block &block)
            {
                int numOfRows = block.GetRowCount();
                LOGDEBUG("Found ", numOfRows, " records for ", m_numOfUpdateReqs, " updates");
                for (int rowIx = 0; rowIx < numOfRows; rowIx++)
                {
                    Uvr::CallRecord row;
                    std::size_t clmnIx = 0;
                    row.numA = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                    row.numB = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                    row.numD = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                    row.numC = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                    uint64_t date = block[clmnIx++]->As<::clickhouse::ColumnDateTime64>()->At(rowIx);
                    row.date.tv_sec = date / timers::giga;
                    row.date.tv_nsec = date % timers::giga;
                    uint64_t dateAct = block[clmnIx++]->As<::clickhouse::ColumnDateTime64>()->At(rowIx);
                    row.dateAct.tv_sec = dateAct;
                    row.dateAct.tv_nsec = 0;
                    auto colNullIdSrc = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                    if (!colNullIdSrc->IsNull(rowIx))
                    {
                        row.idSrc = colNullIdSrc->Nested()->As<::clickhouse::ColumnUInt32>()->At(rowIx);
                    }
                    auto colNullIdDst = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                    if (!colNullIdDst->IsNull(rowIx))
                    {
                        row.idDst = colNullIdDst->Nested()->As<::clickhouse::ColumnUInt32>()->At(rowIx);
                    }
                    row.sessionId = block[clmnIx++]->As<::clickhouse::ColumnUInt64>()->At(rowIx);
                    row.idRel = block[clmnIx++]->As<::clickhouse::ColumnUInt8>()->At(rowIx);
                    row.response = block[clmnIx++]->As<::clickhouse::ColumnInt16>()->At(rowIx);
                    auto colNullEndCode = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                    if (!colNullEndCode->IsNull(rowIx))
                    {
                        row.releaseCode = colNullEndCode->Nested()->As<::clickhouse::ColumnUInt8>()->At(rowIx);
                    }
                    auto colNullCallDur = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                    if (!colNullCallDur->IsNull(rowIx))
                    {
                        row.callDuration = colNullCallDur->Nested()->As<::clickhouse::ColumnUInt16>()->At(rowIx);
                    }
                    auto colNullIdUvrT = block[clmnIx++]->As<::clickhouse::ColumnNullable>();
                    if (!colNullIdUvrT->IsNull(rowIx))
                    {
                        row.idUvrT = colNullIdUvrT->Nested()->As<::clickhouse::ColumnUInt16>()->At(rowIx);
                    }
                    row.callId = block[clmnIx++]->As<::clickhouse::ColumnString>()->At(rowIx);
                    row.callType = block[clmnIx]->As<::clickhouse::ColumnUInt8>()->At(rowIx);

                    BufKey bufKey(row.callId, row.callType);
                    LOGD("Taken record from ", this->db::interfaces::tables::calls::server::tableName,
                         " table by key[", bufKey.key, "]: ", row);
                    auto recIter = m_recToUpdate.find(bufKey);
                    if (recIter != m_recToUpdate.end())
                    {
                        row.update(recIter->second.first);
                        recIter->second.second++;
                        put(row);
                    }
                    else
                    {
                        LOGW("Extra record with key[", bufKey.key,
                             "] is found in select results: ", row);
                    }
                }
            };

            exec<Select>(selectQuery, selectCb);

            for (auto const &recInfo : m_recToUpdate)
            {
                if (recInfo.second.second == 0)
                {
                    LOGW("Failed to update record: ", recInfo.second.first);
                }
            }
            LOGD("Flushed ", m_numOfUpdateReqs, " update requests into insert buffer on ", event);
            m_numOfUpdateReqs = 0;
            m_recToUpdate.clear();
        }
        // Restart timer if needed
        if (restartTimer)
        {
            timers::restartTimer<timers::periodic>(m_updateBufferFlushTimer);
        }
    }

    void ClickHouseAcc::flushAllBuffers(const std::string &event,
                                        bool restartTimer)
    {
        flushUpdateBuffer(event, restartTimer);
        flushInsertBuffer(event, restartTimer);
    }

    void ClickHouseAcc::toSqlValue(const Uvr::CallRecord &rec,
                                   std::string &sqlVal)
    {
        sqlVal = "(\'" + rec.numA + "\', "
                 + "\'" + rec.numB + "\', "
                 + "\'" + rec.numD + "\', "
                 + "\'" + rec.numC + "\', "
                 + std::to_string(rec.date.tv_sec * timers::giga + rec.date.tv_nsec) + ", "
                 + std::to_string(rec.dateAct.tv_sec) + ", "
                 + Uvr::toString(rec.idSrc, s_sqlNullValue) + ", "
                 + Uvr::toString(rec.idDst, s_sqlNullValue) + ", "
                 + std::to_string(rec.sessionId) + ", "
                 + std::to_string(rec.idRel) + ", "
                 + std::to_string(rec.response) + ", "
                 + Uvr::toString(rec.releaseCode, s_sqlNullValue) + ", "
                 + Uvr::toString(rec.callDuration, s_sqlNullValue) + ", "
                 + Uvr::toString(rec.idUvrT, s_sqlNullValue) + ", "
                 + "\'" + rec.callId + "\', "
                 + std::to_string(rec.callType) + ")";
    }

} // namespace db
