#pragma once

// AFAP includes.
#include "components/config/src/config_manager.hpp"
#include "components/db/src/client.hpp"
#include "components/timers/src/timer_periodic.hpp"

// Boost includes.
#include <boost/thread/tss.hpp>

// clickhouse includes.
#include <clickhouse/client.h>

namespace db
{
    class ClickHouse : public client
    {
    public:
        ClickHouse(const Uvr::UvrId_t uvrId,
                   const configuration::manager &config);
        virtual ~ClickHouse() = default;

        bool connect() override;
        void disconnect() override;

        int getActualSchemaVersion();
        int getExpectedSchemaVersion() const;
        void putSchemaVersion(const int version);

        // tables
        void create(const ::db::table::name &tbName,
                    const ::db::table::fields &tbFields,
                    const ::db::table::settings &tbSettings) override;
        void drop(const ::db::table::name &tbName) override;
        void recreate(const ::db::table::name &tbName,
                      const ::db::table::fields &tbFields,
                      const ::db::table::settings &tbSettings) override{/* Not used */};
        void fill_in(const ::db::table::name &tbName,
                     const std::string &filename) override{/* Not used */};
        void subscribe_on(const ::db::table::name &tbName) override{/* Not used */};
        ::db::limit_t size(const ::db::table::name &tbName) override;
        bool is_exist(const ::db::table::name &tbName) override;

        // calls
        void put(const Uvr::CallRecord &rec) override;
        bool update(const Uvr::CallRecord &rec) override;
        void get(Uvr::CallInformationResponses &rsps,
                 const Uvr::CallInformationRequest &req) override;

        // incidents
        void get(::db::incident::rows &rows,
                 const timers::time_point &start,
                 const timers::time_point &stop) override{/* Not used */};
        void put(const ::db::incident::record &rec) override{/* Not used */};
        // incidentsA
        void get(::db::incidentA::rows &rows,
                 const timers::time_point &start,
                 const timers::time_point &stop) override{/* Not used */};
        void put(const ::db::incidentA::record &rec) override{/* Not used */};
        // subscribers
        void populate(const ::db::subscriber::rows &rows) override{/* Not used */};
        void getall(const std::function<void(::db::subscriber::record &&)> &fn) override{/* Not used */};
        void get(::db::subscriber::rows &rows,
                 ::db::limit_t first,
                 ::db::limit_t last) override{/* Not used */};
        void put(const ::db::subscriber::record &rec) override{/* Not used */};
        void update(const ::db::subscriber::record &) override{/* Not used */};
        void remove(const Uvr::Number &) override{/* Not used */};

        // uvrs
        void populate(const ::db::uvr::rows &rows) override{/* Not used */};
        bool get(::db::uvr::record &rec,
                 const Uvr::UvrId_t &uvrId) override { /* Not used */ return false; };
        void get(::db::uvr::rows &rows,
                 const ::Uvr::Ip_t &ipAddress) override{/* Not used */};
        void put(const ::db::uvr::record &rec) override{/* Not used */};
        // hubs
        void populate(const ::db::hub::rows &rows) override{/* Not used */};
        void get(::db::hub::rows &rows,
                 const ::db::gts_t &list) override{/* Not used */};
        void put(const ::db::hub::record &rec) override{/* Not used */};
        // operators
        void populate(const ::db::Operator::Rows &rows) override{/* Not used */};
        void get(::db::Operator::Rows &rows,
                 const ::db::OperatorsIds &list) override{/* Not used */};
        void put(const ::db::Operator::Record &rec) override{/* Not used */};

    protected:
        const Uvr::UvrId_t m_uvrId; ///< UVR ID for which records are stored.

        using ClientPtr = boost::thread_specific_ptr<::clickhouse::Client>;
        ClientPtr m_clientPtr;  ///< Pointer to connection to Clckh DB
        std::string m_db;       ///< Database name for Clckh DB
        std::string m_user;     ///< User name for Clckh DB
        std::string m_password; ///< Password for Clckh DB
        std::string m_host;     ///< Host or IP address of Clckh DB
        int m_port;             ///< Port of Clckh DB
        int m_retries{2};       ///< The number of attempts to execute SQL request

        /// @brief The function creates separate connection to Clickhouse DB for each thread,
        ///        if there is no the connection for the thread yet.
        /// @return Thread local pointer for the connection
        /// @throws Exception if the connection can't be created.
        virtual ::clickhouse::Client *getClient();

        void setSettings(const configuration::manager &config);

        using SqlType = int;
        static constexpr SqlType Query = 0;
        static constexpr SqlType Select = 1;

        /// @brief The function executes SQL request with specified number of retries.
        ///        If the request couldn't be performed, client reconnects to DB and tries again.
        /// @tparam type SQL request type, see @SqlType
        /// @tparam ...QueryArgs Type represents the list of additional agruments
        ///                      for ClickHouse client native function to execute SQL request
        /// @param query The SQL request to be executed
        /// @param ...arguments The list of additional agruments for ClickHouse client native function
        /// @return @c true if SQL request was successfuly executed, otherwise @c false
        template <SqlType type, typename... QueryArgs>
        bool exec(const std::string &query,
                  QueryArgs &...arguments)
        {
            for (int attemptCounter = 1; attemptCounter <= m_retries; attemptCounter++)
            {
                try
                {
                    LOGD("[Attempt=", attemptCounter, "] Executing query=\"", query, "\"");
                    switch (type)
                    {
                        case Query:
                        {
                            getClient()->Execute(query);
                            break;
                        }
                        case Select:
                        {
                            getClient()->Select(query, arguments...);
                            break;
                        }
                        default:
                        {
                            LOGW("Unknown query type(", type, ") for query=\"", query, "\"");
                            return false;
                        }
                    }
                    break; /* query was successfuly executed */
                }
                catch (const std::exception &execEx)
                {
                    /* check the number of attempts on unsuccessful execution */
                    if (attemptCounter >= m_retries)
                    {
                        LOGW("Failed to execute DB query to DB after ", attemptCounter, "/", m_retries,
                             " attempts: query=\"", query, "\", error=\"", execEx.what(), "\"");
                        return false;
                    }
                    /* trying to reconnect to DB again before new attempt */
                    try
                    {
                        LOGD("Reconnecting to DB after failure on : ", execEx.what());
                        m_clientPtr->ResetConnection();
                    }
                    catch (const std::exception &reconnEx)
                    {
                        LOGW("Failed to reconnect to DB on ",
                             attemptCounter, "/", m_retries, " attempt: ", reconnEx.what());
                        LOGW("Failed to execute query to DB after ", attemptCounter, "/", m_retries,
                             " attempts: query=\"", query);
                        return false;
                    }
                }
            }
            return true;
        }
    };

    class ClickHouseAcc : public ClickHouse,
                          public std::enable_shared_from_this<ClickHouseAcc>,
                          public timers::interfaces::manager::server
    {
    public:
        ClickHouseAcc(boost::asio::io_context &ioc,
                      const Uvr::UvrId_t uvrId,
                      const configuration::manager &config);
        virtual ~ClickHouseAcc();

        bool connect() override;
        void disconnect() override;

        // calls
        void put(const Uvr::CallRecord &rec) override;
        bool update(const Uvr::CallRecord &rec) override;
        void get(Uvr::CallInformationResponses &rsps,
                 const Uvr::CallInformationRequest &req) override;

        // timer interface
        void manage_timeout(const timers::id &id,
                            const timers::time_point &start_time) override;
        void manage_stopped(const ::timers::id &id) override;

    protected:
        void flushInsertBuffer(const std::string &event, bool restartTimer);
        void flushUpdateBuffer(const std::string &event, bool restartTimer);
        void flushAllBuffers(const std::string &event, bool restartTimer);

        void toSqlValue(const Uvr::CallRecord &rec,
                        std::string &sqlVal);

        /// @brief Internal type for buffer key to be able to use in unordered map
        struct BufKey
        {
            std::string key;

            BufKey(const Uvr::CallId_t &callId,
                   const Uvr::CallType_t &type)
            {
                switch (type)
                {
                    case Uvr::CallType::outgoing:
                    {
                        key += "out-";
                    }
                    case Uvr::CallType::incoming:
                    {
                        key += "in-";
                    }
                }
                key += callId;
            }

            bool operator==(const BufKey &o) const
            {
                return key == o.key;
            }

            struct Hash
            {
                std::size_t operator()(const BufKey &k) const
                {
                    return std::hash<std::string>()(k.key);
                }
            };
        };
        using BufUpdateInfo = std::pair<Uvr::CallRecord, uint32_t>;

        boost::asio::io_context &m_ioc;                                            ///< Context for async operations
        timers::timer_ptr m_insertBufferFlushTimer;                                ///< Timer to flush insert records into DB
        int m_numOfInsertReqs{0};                                                  ///< Number of insert records in buffer
        int m_insertBufferSize{0};                                                 ///< Buffer size of insert records
        mutable std::recursive_mutex m_gInsert;                                    ///< Protector of shared data in multiple threads
        std::unordered_map<BufKey, Uvr::CallRecords, BufKey::Hash> m_recToInsert; ///< Buffer to store insert records with non-empty callId
        timers::timer_ptr m_updateBufferFlushTimer;                                ///< Timer to flush update records into DB
        int m_numOfUpdateReqs{0};                                                  ///< Number of update records in buffer
        int m_updateBufferSize{0};                                                 ///< Buffer size of update records
        mutable std::recursive_mutex m_gUpdate;                                    ///< Protector of shared data in multiple threads
        std::unordered_map<BufKey, BufUpdateInfo, BufKey::Hash> m_recToUpdate;     ///< Buffer to store update records
    };

} // namespace db
