#pragma once

// AFAP includes.
#include "components/timers/src/timer_types.hpp"
#include "UvrTypes.hpp"

// Boost includes.
#include "boost/json.hpp"

// STL includes.
#include <string>
#include <vector>

namespace db
{
    using limit_t = uint64_t;

    using gts_t = std::vector<Uvr::Gt_t>;
    using OperatorsIds = std::vector<Uvr::OperatorId_t>;

    const limit_t unlimited = -1;

    const ::timers::time_point no_get{0, 0};
    const ::timers::time_point last_get{0, 1};
    const ::timers::time_point latest{0, 2};

    namespace table
    {
        using name = std::string;
        using fields = std::string;
        using settings = std::string;
        using selectSettings = std::string;
    } // namespace table

    // Namespace related to the storage of calls records. Used for call information by request.
    namespace Calls
    {
        const table::name tableName = "calls";
    } // namespace Calls

    namespace incident
    {
        const table::name tableName = "incidents";

        /*
        +----------+--------------+------------------------------------+---------------------------------+
        | ID поля  | Обязательное | Описание                           | Возможные значения и примеры    |
        +----------+--------------+------------------------------------+---------------------------------+
        | NUM_A    | Обязательное | Номер абонента А,                  | Пример: 79251234567             |
        |          |              | в формате E.164                    |                                 |
        +----------+--------------+------------------------------------+---------------------------------+
        | NUM_B    | Обязательное | Номер абонента B,                  | Пример: 79251234567             |
        |          |              | в формате E.164                    |                                 |
        +----------+--------------+------------------------------------+---------------------------------+
        | NUM_D    | Опциональное | Дополнительный номер,              | Пример: 79251234567             |
        |          |              | в формате E.164                    | При отсутствии – пустое поле    |
        +----------+--------------+------------------------------------+---------------------------------+
        | NUM_C    | Опциональное | Номер переадресации,               | Пример: 79251234567             |
        |          |              | в формате E.164                    | При отсутствии – пустое поле    |
        +----------+--------------+------------------------------------+---------------------------------+
        | DATE     | Обязательное | Дата и время попытки соединения    | Пример:                         |
        |          |              | (UTC) с указанием временной зоны.  | 2021-06-06T01:05:49+03:00       |
        |          |              | Формат: YYYY-MM-DDTHH:MM:SS+HH:MM  |                                 |
        +----------+--------------+------------------------------------+---------------------------------+
        | ID_REL   | Обязательное | Идентификатор системы,             | Пример: 1                       |
        |          |              | принявшей решение об отбитии       | Для системы верификации         |
        |          |              | вызова, целое число                | зарезервировано значение “1”    |
        +----------+--------------+------------------------------------+---------------------------------+
        | RLC      | Обязательное | Код причины отбоя вызова,          | В соответствии с номенклатурой  |
        |          |              | целое число.                       | кодов отбоя                     |
        |          |              |                                    | для системы верификации.        |
        +----------+--------------+------------------------------------+---------------------------------+
        | ID_SRC   | Обязательное | Идентификатор оператора,           | Пример: 1                       |
        |          |              | от которого получен вызов          |                                 |
        +----------+--------------+------------------------------------+---------------------------------+
        | ID_UVR_T | Обязательное | Идентификатор УВр, на который      | Пример: 2                       |
        |          |              | был направлен запрос верификации,  |                                 |
        |          |              | приведший к инциденту.             |                                 |
        +----------+--------------+------------------------------------+---------------------------------+
        | CALL_ID  | Опциональное | Идентификатор соединения(вызова).  | Пример:                         |
        |          |              | Создается оператором обслуживающим | 245-1337486@192.168.119.50      |
        |          |              | вызываемого абонента               | При отсутствии – пустое поле    |
        +----------+--------------+------------------------------------+---------------------------------+
        */

        struct record
        {
            Uvr::OctetStringPtr m_numA;       ///< calling number
            Uvr::OctetStringPtr m_numB;       ///< called number
            Uvr::OctetStringPtr m_numC;       ///< origin called number
            Uvr::OctetStringPtr m_numD;       ///< generic number
            ::timers::time_point m_date;      ///< date of connection attempt
            Uvr::VerdictTrigger_t m_idRel;    ///< identifier of system made verdict
            Uvr::VerificationVerdict_t m_rlc; ///< verdict on verification request
            Uvr::OperatorId_t m_idSrc;        ///< source operator ID
            Uvr::UvrId_t m_idUvrT;            ///< UVR ID which verify request was sent to
            Uvr::CallId_t callId;             ///< call Id
        };

        std::ostream &operator<<(std::ostream &os, const record &r);

        using rows = std::vector<record>;
    } // namespace incident

    namespace incidentA
    {
        const table::name tableName = "incidentsA";

        /*
        +----------+--------------+-----------------------------------+---------------------------------+
        | ID поля  | Обязательное | Описание                          | Возможные значения и примеры    |
        +----------+--------------+-----------------------------------+---------------------------------+
        | NUM_A    | Обязательное | Номер абонента А,                 | Пример: 79251234567             |
        |          |              | в формате E.164                   |                                 |
        +----------+--------------+-----------------------------------+---------------------------------+
        | NUM_B    | Обязательное | Номер абонента B,                 | Пример: 79251234567             |
        |          |              | в формате E.164                   |                                 |
        +----------+--------------+-----------------------------------+---------------------------------+
        | NUM_C    | Опциональное | Номер переадресации,              | Пример: 79251234567             |
        |          |              | в формате E.164                   | При отсутствии – пустое поле    |
        +----------+--------------+-----------------------------------+---------------------------------+
        | DATE     | Обязательное | Дата и время попытки соединения   | Пример:                         |
        |          |              | (UTC) с указанием временной зоны. | 2021-06-06T01:05:49+03:00       |
        |          |              | Формат: YYYY-MM-DDTHH:MM:SS+HH:MM |                                 |
        +----------+--------------+-----------------------------------+---------------------------------+
        | ID_REL   | Обязательное | Идентификатор системы,            | Пример: 1                       |
        |          |              | принявшей решение об отбитии      | Для системы верификации         |
        |          |              | вызова, целое число               | зарезервировано значение “1”    |
        +----------+--------------+-----------------------------------+---------------------------------+
        | RLC      | Обязательное | Код причины отбоя вызова,         | В соответствии с номенклатурой  |
        |          |              | целое число.                      | кодов отбоя                     |
        |          |              |                                   | для системы верификации.        |
        +----------+--------------+-----------------------------------+---------------------------------+
        | ID_UVR_O | Обязательное | Идентификатор УВр, с которого     | Пример: 1                       |
        |          |              | получен запрос верификации,       |                                 |
        |          |              | приведший к инциденту.            |                                 |
        +----------+--------------+-----------------------------------+---------------------------------+
        */

        struct record
        {
            Uvr::OctetStringPtr m_numA;       ///< calling number
            Uvr::OctetStringPtr m_numB;       ///< called number
            Uvr::OctetStringPtr m_numC;       ///< origin called number
            ::timers::time_point m_date;      ///< date of connection attempt
            Uvr::VerdictTrigger_t m_idRel;    ///< identifier of system made verdict
            Uvr::VerificationVerdict_t m_rlc; ///< verdict on verification request
            Uvr::UvrId_t m_idUvrO;            ///< UVR ID which verify request was sent from
        };

        std::ostream &operator<<(std::ostream &os, const record &r);

        using rows = std::vector<record>;
    } // namespace incidentA

    namespace subscriber
    {
        const table::name tableName = "subscribers";

        /*
        +-----------+--------------+-----------------------------+-------------------------------+
        |           |              |                             |                               |
        | ID поля   | Обязательное | Описание                    | Возможные значения и примеры  |
        +-----------+--------------+-----------------------------+-------------------------------+
        | NUMBER    | Обязательное | Номер телефона абонента,    | Пример: 79251234567           |
        |           |              | в формате E.164             |                               |
        +-----------+--------------+-----------------------------+-------------------------------+
        | ID_SRC    | Обязательное | Уникальный идентификатор    | Пример: ПАО Оператор Один     |
        |           |              | оператора связи             |                               |
        |           |              | владельца А-номера          |                               |
        +-----------+--------------+-----------------------------+-------------------------------+
        | ID_UVR_P  | Обязательное | Уникальный идентификатор    | Целое число,                  |
        |           |              | основного узла верификации  | назначаемое при регистрации.  |
        |           |              |                             | Возможные   значения 1..16383 |
        +-----------+--------------+-----------------------------+-------------------------------+
        | ID_UVR_S  | Опциональное | Уникальный идентификатор    | Целое число,                  |
        |           |              | резервного узла верификации | назначаемое при регистрации.  |
        |           |              |                             | Возможные   значения 1..16383 |
        +-----------+--------------+-----------------------------+-------------------------------+
        | META_INFO | Опциональное | Мета-информация             | Дополнительная информация     |
        |           |              |                             | по номеру, в формате          |
        |           |              |                             | KEY1=VALUE1, KEY2=VALUE2      |
        +-----------+--------------+-----------------------------+-------------------------------+
        */

        using tuple = std::tuple<Uvr::Number, Uvr::OperatorId_t, Uvr::UvrId_t, Uvr::UvrId_t, Uvr::MetaInfo::String>;

        struct record
        {
            Uvr::Number m_number;
            Uvr::OperatorId_t m_idSrc;
            Uvr::UvrId_t m_idUvrP;
            Uvr::UvrId_t m_idUvrS;
            Uvr::MetaInfo::String m_metaInfo;

            record() = default;
            record(const Uvr::Number &number,
                   const Uvr::OperatorId_t &idSrc,
                   Uvr::UvrId_t idUvrP,
                   Uvr::UvrId_t idUvrS,
                   const Uvr::MetaInfo::String &metaInfo)
                : m_number(number),
                  m_idSrc(idSrc),
                  m_idUvrP(idUvrP),
                  m_idUvrS(idUvrS),
                  m_metaInfo(metaInfo)
            {
            }
            explicit record(const tuple &tpl)
                : m_number(std::get<0>(tpl)),
                  m_idSrc(std::get<1>(tpl)),
                  m_idUvrP(std::get<2>(tpl)),
                  m_idUvrS(std::get<3>(tpl)),
                  m_metaInfo(std::get<4>(tpl))
            {
            }
            record(const char *tmp, const char delimeter = ';');
        };

        record tag_invoke(boost::json::value_to_tag<record>, boost::json::value const &jv);

        using rows = std::vector<record>;
    } // namespace subscriber

    namespace uvr
    {
        const table::name tableName = "uvrs";

        /*
        +-----------+--------------+---------------------------------+----------------------------------+
        | ID поля   | Обязательное | Описание                        | Возможные значения и примеры     |
        +-----------+--------------+---------------------------------+----------------------------------+
        | ID_UVR    | Обязательное | Уникальный идентификатор        | Целое число,                     |
        |           |              | узла верификации                | назначаемое при регистрации.     |
        |           |              |                                 | Возможные значения 1..16383      |
        +-----------+--------------+---------------------------------+----------------------------------+
        | GT_UVR    | Опциональное | Глобальный заголовок (GT)       | Пример: 79209090901              |
        |           |              | в формате E.164, используемый   | Присутствует только если         |
        |           |              | для SCCP маршрутизации          | центр верификации подключен      |
        |           |              |                                 | к сети ОКС7                      |
        +-----------+--------------+---------------------------------+----------------------------------+
        | IP_UVR_P  | Обязательное | Основной IP v4 адрес            | Пример: 89.109.44.77             |
        |           |              | узла верификации                |                                  |
        +-----------+--------------+---------------------------------+----------------------------------+
        | IP_UVR_S  | Опциональное | Резервный IP v4 адрес           | Пример: 89.109.44.78             |
        |           |              | узла верификации                | Пустое значение, при отсутствии  |
        +-----------+--------------+---------------------------------+----------------------------------+
        | DNS_UVR   | Опциональное | Доменное имя (DNS)              | Пример                           |
        |           |              | узла верификации                | <id узла верификации>.<домен>.ru |
        +-----------+--------------+---------------------------------+----------------------------------+
        | ID_HUB_P  | Опциональное | Уникальный идентификатор        | Пример: 79259090900              |
        |           |              | основного узла взаимодействия,  | В случае подключения             |
        |           |              | глобальный заголовок            | УВР напрямую к оператору         |
        |           |              | в формате E.164                 | связи поле опциональное          |
        +-----------+--------------+---------------------------------+----------------------------------+
        | ID_HUB_S  | Опциональное | Уникальный идентификатор        | Пример: 79259090901              |
        |           |              | резервного узла взаимодействия, |                                  |
        |           |              | глобальный заголовок            |                                  |
        |           |              | в формате E.164                 |                                  |
        +-----------+--------------+---------------------------------+----------------------------------+
        | GT_UVR1   | Опциональное | Глобальный заголовок (GT)       | Пример: 79209090901              |
        |           |              | в формате E.164, используемый   | Присутствует только если         |
        |           |              | для SCCP маршрутизации          | центр верификации подключен      |
        |           |              |                                 | к сети ОКС7                      |
        +-----------+--------------+---------------------------------+----------------------------------+
        | GT_UVR2   | Опциональное | Глобальный заголовок (GT)       | Пример: 79209090901              |
        |           |              | в формате E.164, используемый   | Присутствует только если         |
        |           |              | для SCCP маршрутизации          | центр верификации подключен      |
        |           |              |                                 | к сети ОКС7                      |
        +-----------+--------------+---------------------------------+----------------------------------+
        | ID_SRC    | Опциональное | Идентификатор оператора связи   | Целое число, начиная с нуля.     |
        |           |              | владельца УВр, согласно         | Возможные значения:              |
        |           |              | справочнику                     | 0..4294967295.                   |
        |           |              |                                 | Диапазон от 0 до 9999            |
        |           |              |                                 | зарезервирован под служебные     |
        |           |              |                                 | нужды.                           |
        +-----------+--------------+---------------------------------+----------------------------------+
        |           | Опциональное | Мета-информация                 | Дополнительная информация по     |
        | META_INFO |              |                                 | УВр, в формате:                  |
        |           |              |                                 | KEY1=VALUE1, KEY2=VALUE2         |
        +-----------+--------------+---------------------------------+----------------------------------+
        */

        struct record
        {
            Uvr::UvrId_t m_idUvr{Uvr::UvrId::undefined};
            Uvr::Gt_t m_gtUvr;
            Uvr::Ip_t m_ipUvrP;
            Uvr::Ip_t m_ipUvrS;
            Uvr::Dns_t m_dnsUvr;
            Uvr::Gt_t m_idHubP;
            Uvr::Gt_t m_idHubS;
            Uvr::Gt_t m_gtUvr1;
            Uvr::Gt_t m_gtUvr2;
            Uvr::OperatorId_t m_idSrc{Uvr::undefinedOperatorId};
            Uvr::MetaInfo::BitFlags_t m_metaInfo{Uvr::MetaInfo::emptyBitFlags};

            record() = default;
            record(Uvr::UvrId_t idUvr,
                   const Uvr::Gt_t &gtUvr,
                   const Uvr::Ip_t &ipUvrP,
                   const Uvr::Ip_t &ipUvrS,
                   const Uvr::Dns_t &dnsUvr,
                   const Uvr::Gt_t &idHubP,
                   const Uvr::Gt_t &idHubS,
                   const Uvr::Gt_t &gtUvr1,
                   const Uvr::Gt_t &gtUvr2,
                   const Uvr::OperatorId_t idSrc)
                : m_idUvr(idUvr),
                  m_gtUvr(gtUvr),
                  m_ipUvrP(ipUvrP),
                  m_ipUvrS(ipUvrS),
                  m_dnsUvr(dnsUvr),
                  m_idHubP(idHubP),
                  m_idHubS(idHubS),
                  m_gtUvr1(gtUvr1),
                  m_gtUvr2(gtUvr2),
                  m_idSrc(idSrc)
            {
            }
            record(const char *line, const char delimeter = ';');

            bool operator==(const record &other) const;
        };

        std::ostream &operator<<(std::ostream &os, const record &r);

        using rows = std::vector<record>;
    } // namespace uvr

    namespace hub
    {
        const table::name tableName = "hubs";

        /*
        +-----------+--------------+--------------------------------------+----------------------------------+
        | ID поля   | Обязательное | Описание                             | Возможные значения и примеры     |
        +-----------+--------------+--------------------------------------+----------------------------------+
        | ID_HUB    | Обязательное | Уникальный идентификатор             | Пример:                          |
        |           |              | узла взаимодействия,                 | 79259090909                      |
        |           |              | глобальный заголовок в формате E.164 |                                  |
        +-----------+--------------+--------------------------------------+----------------------------------+
        | IP_HUB_P  | Обязательное | IP v4 адрес основного хоста          | Пример:                          |
        |           |              | узла взаимодействия                  | 89.109.44.55                     |
        +-----------+--------------+--------------------------------------+----------------------------------+
        | IP_HUB_P  | Обязательное | IP v4 адрес резервного хоста         | Пример:                          |
        |           |              | узла взаимодействия                  | 89.109.44.55                     |
        +-----------+--------------+--------------------------------------+----------------------------------+
        | DNS_HUB_P | Обязательное | Доменное имя (DNS) основного хоста   | Пример:                          |
        |           |              | узла взаимодействия                  | <id узла верификации>.<домен>.ru |
        +-----------+--------------+--------------------------------------+----------------------------------+
        | DNS_HUB_S | Обязательное | Доменное имя (DNS) резервного хоста  | Пример:                          |
        |           |              | узла взаимодействия                  | <id узла верификации>.<домен>.ru |
        +-----------+--------------+--------------------------------------+----------------------------------+
        */

        struct record
        {
            Uvr::Gt_t m_idHub;
            Uvr::Ip_t m_ipHubP;
            Uvr::Ip_t m_ipHubS;
            Uvr::Dns_t m_dnsHubP;
            Uvr::Dns_t m_dnsHubS;

            record() = default;
            record(const Uvr::Gt_t &idHub,
                   const Uvr::Ip_t &ipHubP,
                   const Uvr::Ip_t &ipHubS,
                   const Uvr::Dns_t &dnsHubP,
                   const Uvr::Dns_t &dnsHubS)
                : m_idHub(idHub),
                  m_ipHubP(ipHubP),
                  m_ipHubS(ipHubS),
                  m_dnsHubP(dnsHubP),
                  m_dnsHubS(dnsHubS)
            {
            }
            record(const char *line, const char delimeter = ';');
        };

        using rows = std::vector<record>;
    } // namespace hub

    namespace Operator
    {
        const table::name tableName = "operators";

        /*
        +------------+--------------+------------------------------+--------------------------------------+
        | ID поля    | Обязательное | Описание                     | Возможные значения и примеры         |
        +------------+--------------+------------------------------+--------------------------------------+
        | ID_SRC     | Обязательное | Идентификатор оператора      | Целое число, начиная с нуля.         |
        |            |              | связи, согласно справочнику  | Возможные значения: 0..4294967295.   |
        |            |              |                              |                                      |
        |            |              |                              | Диапазон от 0 до 9999 зарезервирован |
        |            |              |                              | под служебные нужды. См. таблицу 9   |
        +------------+--------------+------------------------------+--------------------------------------+
        | OPR_NAME   | Обязательное | Наименование оператора связи | Текстовое поле, от 1 символа.        |
        |            |              |                              |                                      |
        |            |              |                              | Пример: Публичное Акционерное        |
        |            |              |                              | Общество Оператор Один               |
        +------------+--------------+------------------------------+--------------------------------------+
        | OPR_NICK   | Обязательное | Укороченное наименование     | Текстовое поле, от 1 символа.        |
        |            |              | оператора связи              |                                      |
        |            |              |                              | Пример: ПАО Оператор Один            |
        +------------+--------------+------------------------------+--------------------------------------+
        | INN        | Обязательное | Индивидуальный номер         | Последовательность из цифр           |
        |            |              | налогоплательщика для        | длиною до 12 символов.               |
        |            |              | оператора связи              | Пустая строка – при                  |
        |            |              |                              | отсутствии номера.                   |
        |            |              |                              |                                      |
        |            |              |                              | Пример: 6658217170                   |
        +------------+--------------+------------------------------+--------------------------------------+
        | BDPN_CODE  | Опциональное | Список пар: код оператора    | Информация по оператору связи в      |
        |            |              | связи MNC и код организации  | формате MNC=ORG,MNC2=ORG2            |
        |            |              | OrgCode согласно БДПН        |                                      |
        +------------+--------------+------------------------------+--------------------------------------+
        | NAME_BRAND | Опциональное | Фирменное наименование       | Текстовое поле, от 1 символа         |
        |            |              | оператора связи              |                                      |
        +------------+--------------+------------------------------+--------------------------------------+
        */

        struct Record
        {
            Uvr::OperatorId_t idSrc;
            Uvr::Name oprName;
            Uvr::Name oprNick;
            Uvr::Inn inn;
            Uvr::OptionalBdpnCode bdpnCode;
            Uvr::OptionalName nameBrand;

            Record(const Uvr::OperatorId_t idSrc,
                   const Uvr::Name &oprName,
                   const Uvr::Name &oprNick,
                   const Uvr::Inn &inn,
                   const Uvr::OptionalBdpnCode &bdpnCode,
                   const Uvr::OptionalName &nameBrand);

            Record(const char *line, const char delimeter = ';');
        };

        using Rows = std::vector<Record>;
    }; // namespace Operator

} // namespace db