#pragma once

// AFAP includes.
#include "components/http/src/interfaces.hpp"
#include "components/http/src/types.hpp"
#include "components/log/src/Log.hpp"
#include "components/timers/src/timer_types.hpp"
#include "UvrTypes.hpp"

// Boost includes.
#include <boost/asio.hpp>
#include <boost/beast/core.hpp>
#include <boost/bind/bind.hpp>

namespace http
{
    const int ver1_1 = 11;
    using tcp = boost::asio::ip::tcp;
    using stream = boost::beast::tcp_stream;
    using stream_ptr = std::shared_ptr<stream>;

    std::string connState2str(connections::state state);
    template <bool isClient>
    std::ostream &operator<<(std::ostream &os, const connection<isClient> &c);

    template <bool isClient>
    class connection : public std::enable_shared_from_this<connection<isClient>>,
                       public interfaces::connections::notifier::client,
                       public interfaces::connections::resolver::client
    {
    public:
        template <bool isRequest = isClient,
                  typename = typename std::enable_if<isRequest>::type>
        connection(boost::asio::io_context &ioc,
                   const ::timers::duration &connectTimeout,
                   const std::string &port = "",
                   const std::string &address = "") : m_ioc(ioc),
                                                      m_address(address),
                                                      m_port(port),
                                                      m_connectTimeout(connectTimeout)
        {
        }

        template <bool isRequest = !isClient,
                  typename = typename std::enable_if<isRequest>::type>
        connection(boost::asio::io_context &ioc,
                   const std::string &port = "",
                   const std::string &address = "") : m_ioc(ioc),
                                                      m_address(address),
                                                      m_port(port),
                                                      m_connectTimeout(timers::duration())
        {
        }

        virtual ~connection() = default;

        template <bool isRequest = isClient,
                  typename = typename std::enable_if<isRequest>::type>
        bool startSessionAndQuery(const std::string &target,
                                  const boost::beast::string_view &content,
                                  const std::string &payloadData,
                                  bool keepAlive = false)
        {
            // Prepare request to send
            m_outMsg = {};
            m_outMsg.method(boost::beast::http::verb::post);
            m_outMsg.target(target);
            m_outMsg.version(http::ver1_1);
            m_outMsg.set(boost::beast::http::field::content_type, content);
            m_outMsg.keep_alive(keepAlive);
            m_outMsg.body() = payloadData;
            m_outMsg.prepare_payload();
            return connect();
        }

        template <bool isRequest = isClient,
                  typename = typename std::enable_if<isRequest>::type>
        bool rerouteRequest(void)
        {
            bool hasNextAddress = connect();
            if (hasNextAddress)
            {
                // reset as new attempt is started
                m_firstMsgRcvd = false;
            }
            return hasNextAddress;
        }

        template <bool isResponse = !isClient,
                  typename = typename std::enable_if<isResponse>::type>
        void reply(messages::result result,
                   const boost::beast::string_view &content,
                   const std::string &payloadData,
                   bool keepAlive = false)
        {
            http::message<http::messages::RESPONSE> res{result, http::ver1_1};
            res.set(boost::beast::http::field::accept,
                    boost::beast::http::to_string(boost::beast::http::verb::post));
            res.set(boost::beast::http::field::content_type, content);
            res.keep_alive(keepAlive);
            res.body() = payloadData;
            res.prepare_payload();
            send(res);
            // Close connection if needed
            if (!keepAlive)
            {
                shutdown();
            }
        }

        void shutdown(void)
        {
            boost::asio::post(m_streamPtr->get_executor(),
                              std::bind(&connection<isClient>::shutdown_int,
                                        this->shared_from_this()));
        }

        void shutdown_int(void)
        {
            // Initiate graceful connection closure.
            changeState(connections::state::DOWN);
            if (m_streamPtr->socket().is_open())
            {
                boost::system::error_code ignored_err;
                m_streamPtr->socket().cancel(ignored_err);
                m_streamPtr->socket().shutdown(tcp::socket::shutdown_both, ignored_err);
                m_streamPtr->socket().close(ignored_err);
            }
            http::interfaces::connections::notifier::client::connectionDown(m_connId, isClient ? role::CLIENT : role::SERVER);
        }

        void changeState(connections::state state)
        {
            LOGD("Change state ", connState2str(m_state), "->", connState2str(state), " for ", *this);
            m_state = state;
        }

        http::role role(void) const
        {
            return isClient ? role::CLIENT : role::SERVER;
        }
        connections::id connId(void) const
        {
            return m_connId;
        }
        void connId(connections::id connId)
        {
            m_connId = connId;
        }
        connections::state state(void) const
        {
            return m_state;
        }
        const std::string &address(void) const
        {
            return m_address;
        }
        const std::string &port(void) const
        {
            return m_port;
        }
        template <bool isRequest = isClient,
                  typename = typename std::enable_if<isRequest>::type>
        const http::message<isRequest> &outgoingRequest(void) const
        {
            return m_outMsg;
        }
        Uvr::SessionId_t sessionId(void) const
        {
            return m_sessionId;
        }
        void sessionId(const Uvr::SessionId_t &sessionId)
        {
            m_sessionId = sessionId;
        }
        std::shared_ptr<void> userData(void) const
        {
            return m_userData;
        }
        void userData(const std::shared_ptr<void> &userDataPtr)
        {
            m_userData = userDataPtr;
        }

        void onAccept(tcp::socket &&sock)
        {
            boost::system::error_code err;
            auto rem_ep = sock.remote_endpoint(err);
            if (err)
            {
                LOGW("Cannot get remote address of ", *this, ": ", err.message());
                resetSock();
                return;
            }
            m_address = rem_ep.address().to_string();
            m_port = std::to_string(rem_ep.port());

#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
            boost::asio::socket_base::receive_buffer_size rcvOpt;
            sock.get_option(rcvOpt);
            int rcvSize = rcvOpt.value();
            boost::asio::socket_base::send_buffer_size sndOpt;
            sock.get_option(sndOpt);
            int sndSize = sndOpt.value();
#endif
            LOGD("On socket accept for ", *this,
                 ": rsvbuf=", rcvSize, ", sndbuf=", sndSize);
            m_streamPtr.reset(new stream(std::move(sock)));
            changeState(connections::state::AUTHENTICATING);
            if (http::interfaces::connections::notifier::client::connectionAccept(this->shared_from_this()))
            {
                changeState(connections::state::ESTBLISHED);
                http::interfaces::connections::notifier::client::connectionUp(this->shared_from_this());
                startAsyncWait();
            }
            else
            {
                resetSock();
            }
        }

    protected:
        void onConnect(const boost::system::error_code &err)
        {
            if (!err)
            {
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
                boost::asio::socket_base::receive_buffer_size rcvOpt;
                m_streamPtr->socket().get_option(rcvOpt);
                int rcvSize = rcvOpt.value();
                boost::asio::socket_base::send_buffer_size sndOpt;
                m_streamPtr->socket().get_option(sndOpt);
                int sndSize = sndOpt.value();
#endif
                LOGD("On socket connect for ", *this,
                     ": rsvbuf=", rcvSize, ", sndbuf=", sndSize);
                changeState(connections::state::ESTBLISHED);
                http::interfaces::connections::notifier::client::connectionUp(this->shared_from_this());
                if (send(m_outMsg))
                {
                    // Unset timeout on async operations
                    m_streamPtr->expires_never();
                    startAsyncWait();
                    return;
                }
            }
            // In case of error on connect or on sending request above
            // or couldn't connect using all configured addresses
            if (err != boost::asio::error::operation_aborted && !connect())
            {
                LOGW("Cannot establish ", *this, ": ", err.message());
                if (m_firstMsgSent)
                    resendFailure(this->shared_from_this());
                else
                    connectFailure(this->shared_from_this());
                resetSock();
            }
        }

        void onRead(const boost::system::error_code &err, std::size_t bytesTransferred)
        {
            if (!err)
            {
                LOGD("Received message:\n", m_inMsg);
                m_firstMsgRcvd = true;
                bool isWaiting = false;
                if (isClient && !m_inMsg.keep_alive())
                    shutdown();
                else
                    isWaiting = true;

                if (m_inMsg.payload_size().value_or(0) > 0)
                    newMessage(this, m_inMsg);

                if (isWaiting && m_state == connections::state::ESTBLISHED) // check that no rerouting was called on newMessage() fire
                    startAsyncWait();

                return;
            }

            if (err != boost::beast::http::error::end_of_stream && std::strcmp("beast.http", err.category().name()) == 0)
            {
                LOGW("[on read] Bad http messages is received on ", *this, ": err(", err, "), ", err.message(),
                     "\nMessage:\n", std::string(static_cast<char *>(m_inBuf.data().data()), m_inBuf.size()));
                m_inBuf = {}; // cleanup incoming message buffer
                if (isClient && !m_firstMsgRcvd)
                {
                    LOGI("[on read] Trying to resend request on ", *this, " to another address");
                    shutdown(); // Close current connection
                    if (!connect())
                        resendFailure(this->shared_from_this());
                }
                else if (!isClient)
                {
                    // Indicates request originator about bad http message
                    http::message<http::messages::RESPONSE> res{http::messages::result::bad_request, http::ver1_1};
                    res.set(boost::beast::http::field::content_type, messages::content_text);
                    res.keep_alive(false);
                    res.body() = err.message();
                    res.prepare_payload();
                    send(res);
                    shutdown(); // Close current connection
                }
            }
            else if (err != boost::asio::error::operation_aborted)
            {
                http::interfaces::connections::notifier::client::connectionDown(m_connId, isClient ? role::CLIENT : role::SERVER);
                // If outgoing connection was closed while waiting for first reply
                bool newConnectionInit = true;
                if (isClient && !m_firstMsgRcvd && (newConnectionInit = connect()))
                {
                    return;
                }
                else if (!newConnectionInit) // reconnect above failed
                {
                    resendFailure(this->shared_from_this());
                }
                else // it is SERVER or message was already received on this connection
                {
                    resetSock();
                }
                LOGW("[on read] Lost ", *this, ": err(", err, "), ", err.message());
            }
        }

        bool connect(void)
        {
            resetSock();
            std::string newAddress;
            http::interfaces::connections::resolver::client::getAddress(m_attemptNumber, m_address, newAddress);
            if (newAddress.empty())
            {
                LOGW("Cannot establish ", *this,
                     ": no available addresses after ", m_attemptNumber, " attempts");
                return false;
            }
            m_attemptNumber++;
            m_address = newAddress;
            LOGD("Start to establish ", *this, ", attempt=", m_attemptNumber);
            boost::system::error_code err;
            tcp::resolver resolver(m_ioc);
            tcp::resolver::query query(m_address, m_port);
            auto const endpoints = *resolver.resolve(query, err);
            if (err)
            {
                LOGW("Cannot resolve peer address for ", *this,
                     ", try next one: ", err.message());
                connect(); // try next address
            }

            // create socket descriptor to set socket options
            m_streamPtr->socket().open(boost::asio::ip::tcp::v4());
            // Increase TCP socket buffer size
            LOGD("Change the current value of TCP socket receive buffer size to ",
                 connections::tcp_buf_size, " for ", *this);
            boost::asio::socket_base::receive_buffer_size newRcvBufOpt(connections::tcp_buf_size);
            m_streamPtr->socket().set_option(newRcvBufOpt, err);
            if (err)
            {
                LOGW("Cannot set RCVBUF option for socket of ",
                     *this, ": ", err.message());
            }

            LOGD("Change the current value of TCP socket send buffer size to ",
                 connections::tcp_buf_size, " for ", *this);
            boost::asio::socket_base::send_buffer_size newSndBufOpt(connections::tcp_buf_size);
            m_streamPtr->socket().set_option(newSndBufOpt, err);
            if (err)
            {
                LOGW("Cannot set SNDBUF option for socket of ",
                     *this, ":", err.message());
            }
            m_streamPtr->expires_after(m_connectTimeout);
            m_streamPtr->async_connect(endpoints,
                                       boost::beast::bind_front_handler(&connection<isClient>::onConnect,
                                                                        this->shared_from_this()));
            changeState(connections::state::STARTED);
            return true;
        }

        void startAsyncWait()
        {
            m_inMsg = {};
            boost::beast::http::async_read(*m_streamPtr, m_inBuf, m_inMsg,
                                           boost::beast::bind_front_handler(&connection<isClient>::onRead,
                                                                            this->shared_from_this()));
        }

        template <bool isRequest>
        bool send(http::message<isRequest> &msg)
        {
            if (m_state < connections::state::AUTHENTICATING)
            {
                LOGW("Cannot send message via ", *this, ": connection should be in \"",
                     connState2str(connections::state::AUTHENTICATING), "\" or \"",
                     connState2str(connections::state::ESTBLISHED), "\" state. ",
                     "Current state: \"", connState2str(m_state), "\".\n",
                     "Http message:\n", msg);
                return false;
            }

            // Add IP address and port of nearest node into http header
            msg.set(boost::beast::http::field::host, m_address + ":" + m_port);

            boost::beast::error_code err;
            boost::beast::http::write(*m_streamPtr, msg, err);
            LOGD("Sent message:\n", msg);

            if (err && err != boost::asio::error::operation_aborted)
            {
                LOGW("[on send] Lost ", *this, ": ", err.message());
                resetSock();
                http::interfaces::connections::notifier::client::connectionDown(m_connId, isClient ? role::CLIENT : role::SERVER);
                return false;
            }
            m_firstMsgSent = true;
            return true;
        }

        void resetSock(void)
        {
            changeState(connections::state::INIT);
            m_streamPtr.reset(new stream(boost::asio::make_strand(m_ioc)));
        }

    private:
        boost::asio::io_context &m_ioc;                       ///< Context for asynchronous operations
        connections::id m_connId;                             ///< Connection ID
        std::string m_address;                                ///< Remote IP address
        std::string m_port;                                   ///< Remote port
        connections::state m_state{connections::state::INIT}; ///< Connection state
        stream_ptr m_streamPtr{nullptr};                      ///< Pointer to TCP socket
        http::message<isClient> m_outMsg;                     ///< Container to store the outgoing http message
        http::message<!isClient> m_inMsg;                     ///< Container to store the incomming http message
        boost::beast::flat_buffer m_inBuf;                    ///< Buffer for additional bytes read by the implementation from the stream
        Uvr::SessionId_t m_sessionId{0};                      ///< Connection specific data
        uint m_attemptNumber{0};                              ///< The attempt number to establish connection or retransmit
        bool m_firstMsgSent{false};                           ///< Indicates that first message was sent
        bool m_firstMsgRcvd{false};                           ///< Indicates that first message was received
        std::shared_ptr<void> m_userData;                     ///< Http user specific data for connection
        timers::duration m_connectTimeout;                    ///< TCP connect timeout

        friend std::ostream &operator<< <>(std::ostream &os, const connection<isClient> &c);

    }; // connection

    inline std::string connState2str(connections::state state)
    {
        switch (state)
        {
            case connections::state::NONE:
                return "NONE";
            case connections::state::INIT:
                return "INIT";
            case connections::state::STARTED:
                return "STARTED";
            case connections::state::DOWN:
                return "DOWN";
            case connections::state::AUTHENTICATING:
                return "AUTHENTICATING";
            case connections::state::ESTBLISHED:
                return "ESTABLISHED";
            default:
                return "UNKWN(" + std::to_string(static_cast<int>(state)) + ")";
        }
    }

    template <bool isClient>
    std::ostream &operator<<(std::ostream &os, const connection<isClient> &c)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);
        logger << "http conn[id:" << c.m_connId << ";sid:" << c.m_sessionId;
        if (isClient)
        {
            logger << ";CL;";
        }
        else
        {
            logger << ";SR;";
        }
        logger << connState2str(c.m_state)
               << ";ep=" << c.m_address << ":" << c.m_port << "]";
        os << trace.str();
        return os;
    }

} // namespace http
