#include "http/src/connection_manager.hpp"

// AFAP includes.
#include "components/http/src/connection.hpp"

namespace http
{
    namespace connections
    {
        manager::manager(const TypeInst &owner,
                         boost::asio::io_context &ioc)
            : m_ownInst(owner),
              m_ioc(ioc)
        {
            m_connIdCntr.init();
        }

        manager::~manager()
        {
            destroyAllConnections();
        }

        connection_ptr<role::CLIENT> manager::createOutgoingConnection(Uvr::SessionId_t reqId,
                                                                       const std::string &port,
                                                                       const timers::duration &connectTimeout)
        {
            auto connPtr =
                std::make_shared<connection<role::CLIENT>>(m_ioc, connectTimeout, port);
            connPtr->::http::interfaces::connections::notifier::client::bind(this);
            connPtr->::http::interfaces::connections::resolver::client::bind(this);
            connPtr->sessionId(reqId);
            return connPtr;
        }

        id manager::addConnection(connection_ptr<role::CLIENT> connPtr)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            id connId(connPtr->sessionId());
            connPtr->connId(connId);
            auto result = m_outgoing.emplace(connId, connPtr);
            if (!result.second)
            {
                LOGERROR(m_ownInst,
                         " The outgoing connection with id=", connId, " already exists: old ",
                         *(result.first->second), ", new ", *connPtr);
            }
            LOGDEBUG(m_ownInst, " Added ", *connPtr);
            return connId;
        }

        id manager::addConnection(connection_ptr<role::SERVER> connPtr)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            id connId = m_connIdCntr++;
            connPtr->connId(connId);
            m_incoming.emplace(connId, connPtr);
            LOGDEBUG(m_ownInst, " Added ", *connPtr);
            return connId;
        }

        connection_ptr<role::CLIENT> manager::getOutgoingConnection(id connId)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            auto connIter = m_outgoing.find(connId);
            if (connIter == m_outgoing.end())
                return nullptr;
            return connIter->second;
        }

        connection_ptr<role::SERVER> manager::getIncomingConnection(id connId)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            auto connIter = m_incoming.find(connId);
            if (connIter == m_incoming.end())
                return nullptr;
            return connIter->second;
        }

        void manager::destroyAllConnections(void)
        {
            LOGDEBUG(m_ownInst,
                     " Destroying all connections: outgoing=", m_outgoing.size(),
                     ", incoming=", m_incoming.size());
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            m_outgoing.clear();
            m_incoming.clear();
        }

        bool manager::connectionAccept(connection_ptr<role::CLIENT> connPtr)
        {
            return true;
        }

        bool manager::connectionAccept(connection_ptr<role::SERVER> connPtr)
        {
            return true;
        }

        void manager::connectFailure(connection_ptr<role::CLIENT> connPtr)
        {
            // Do something with an ancestor
        }

        void manager::connectFailure(connection_ptr<role::SERVER> connPtr)
        {
            // Do something with an ancestor
        }

        void manager::resendFailure(connection_ptr<role::CLIENT> connPtr)
        {
            // Do something with an ancestor
        }

        void manager::resendFailure(connection_ptr<role::SERVER> connPtr)
        {
            // Do something with an ancestor
        }

        void manager::connectionUp(connection_ptr<role::CLIENT> connPtr)
        {
            addConnection(connPtr);
        }

        void manager::connectionUp(connection_ptr<role::SERVER> connPtr)
        {
            addConnection(connPtr);
        }

        void manager::connectionDown(connections::id connId, role role)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            if (role == role::CLIENT)
            {
                auto result = m_outgoing.find(connId);
                if (result == m_outgoing.end())
                {
                    LOGERROR(m_ownInst,
                             " Outgoing connection with id=", connId,
                             " is not found on connection down event");
                    return;
                }
                // clean connection map
                m_outgoing.erase(result);
            }
            else
            {
                auto result = m_incoming.find(connId);
                if (result == m_incoming.end())
                {
                    LOGWARN(m_ownInst,
                            " Incoming connection with id=", connId,
                            " is not found on connection down event");
                    return;
                }
                // clean connection map
                m_incoming.erase(result);
            }
        }

        void manager::newMessage(const http::connection<role::SERVER> *connPtr,
                                 const http::message<messages::type::REQUEST> &msg)
        {
            std::string payloadData = msg.body();
            if (isServiceMsg(msg))
            {
                // handle here service requests if any
            }
            else
            {
                requestSpecificHandling(connPtr->connId(), msg.target(), payloadData);
            }
        }

        void manager::newMessage(const http::connection<role::CLIENT> *connPtr,
                                 const http::message<messages::type::RESPONSE> &msg)
        {
            std::string payloadData = msg.body();
            if (isServiceMsg(msg))
            {
                // handle here service responses if any
            }
            else
            {
                responseSpecificHandling(connPtr, msg.result(), payloadData, msg.keep_alive());
            }
        }

        void manager::getAddress(uint attemptNumber,
                                 const std::string &prevAddress,
                                 std::string &nextAddress)
        {
            // only one attempt to establish connection by default
            if (attemptNumber == 0)
                nextAddress = prevAddress;
            else
                nextAddress.clear();
        }

    } // namespace connections

} // namespace http
