#pragma once

// AFAP includes.
#include "components/http/src/interfaces.hpp"
#include "components/worker_node/src/typeInst.hpp"
#include "components/timers/src/timer_types.hpp"
#include "UvrTypes.hpp"

// STL includes.
#include <unordered_map>
#include <mutex>

namespace http
{
    namespace connections
    {
        class manager : public http::interfaces::connections::notifier::server,
                        public http::interfaces::connections::resolver::server
        {
        public:
            manager(const TypeInst &owner,
                    boost::asio::io_context &ioc);
            ~manager() override;

            connection_ptr<role::CLIENT> createOutgoingConnection(Uvr::SessionId_t reqId,
                                                                  const std::string &port,
                                                                  const ::timers::duration &connectTimeout);

            connection_ptr<role::CLIENT> getOutgoingConnection(id connId);
            connection_ptr<role::SERVER> getIncomingConnection(id connId);

            void destroyAllConnections(void);

        protected:
            virtual void requestSpecificHandling(connections::id connId,
                                                 const std::string_view &target,
                                                 const std::string &payloadData) = 0;
            virtual void responseSpecificHandling(const http::connection<role::CLIENT> *connPtr,
                                                  messages::result result,
                                                  const std::string &payloadData,
                                                  bool keepAlive) = 0;

            id addConnection(connection_ptr<role::CLIENT> connPtr);
            id addConnection(connection_ptr<role::SERVER> connPtr);

            // Implementation of http::connections::notifier::server interface

            /// @brief Do nothing.
            /// @returns Always @c true.
            bool connectionAccept(connection_ptr<role::CLIENT> connPtr) override;

            /// @brief Simply adds connection to internal map.
            /// @returns Always @c true.
            bool connectionAccept(connection_ptr<role::SERVER> connPtr) override;

            void connectFailure(connection_ptr<role::CLIENT> connPtr) override;
            void connectFailure(connection_ptr<role::SERVER> connPtr) override;
            void resendFailure(connection_ptr<role::CLIENT> connPtr) override;
            void resendFailure(connection_ptr<role::SERVER> connPtr) override;
            void connectionUp(connection_ptr<role::CLIENT> connPtr) override;
            void connectionUp(connection_ptr<role::SERVER> connPtr) override;
            void connectionDown(connections::id connId, role role) override;
            void newMessage(const http::connection<role::SERVER> *connPtr,
                            const http::message<messages::type::REQUEST> &msg) override;
            void newMessage(const http::connection<role::CLIENT> *connPtr,
                            const http::message<messages::type::RESPONSE> &msg) override;
            // Implementation of http::connections::resolver::server interface
            void getAddress(uint attemptNumber,
                            const std::string &prevAddress,
                            std::string &nextAddress) override;

            template <bool isRequest>
            bool isServiceMsg(const http::message<isRequest> &msg) const
            {
                auto field = msg.find(boost::beast::http::field::connection);
                if (field == msg.end())
                    return false;

                std::string_view value = msg[boost::beast::http::field::connection];
                if (value == "Upgrade")
                    return true;

                return false;
            }

            const TypeInst &m_ownInst;                                             ///< VFN ID (for debug purpose)
            boost::asio::io_context &m_ioc;                                        ///< Context for asynchronous operations
            mutable std::recursive_mutex m_guard;                                  ///< Protector of shared data in multiple threads
            std::unordered_map<id, http::connection_ptr<role::SERVER>> m_incoming; ///< Storage of incoming connections
            std::unordered_map<id, http::connection_ptr<role::CLIENT>> m_outgoing; ///< Storage of outgoing connections
            id m_connIdCntr;                                                       ///< Connection ID counter
        };
    } // namespace connections
} // namespace http
