#include "http/src/listener.hpp"

namespace http
{
    listener::listener(const TypeInst &owner,
                       boost::asio::io_context &ioc)
        : m_owner(owner),
          m_ioc(ioc),
          m_acceptor(m_ioc)
    {
    }

    void listener::start(const std::string &port)
    {
        m_stopped = false;
        prepareAcceptor(port);
        startAccept();
    }

    void listener::stop(void)
    {
        boost::system::error_code err;
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        m_acceptor.close(err);
        if (err)
        {
            LOGW(m_owner,
                 " Failed to close HTTP acceptor: ", err.message());
        }
        m_stopped = true;
        m_nextConnPtr.reset();
    }

    void listener::prepareAcceptor(const std::string &port)
    {
        boost::system::error_code err;
        tcp::endpoint endpoint(tcp::v4(),
                               static_cast<boost::asio::ip::port_type>(std::stoi(port)));
        m_acceptor.open(endpoint.protocol(), err);
        if (err)
        {
            LOGE(m_owner,
                 " Cannot open socket on ", port, " port: ", err.message());
            return;
        }
        m_acceptor.set_option(tcp::acceptor::reuse_address(true), err);
        if (err)
        {
            LOGE(m_owner,
                 " Cannot set REUSE_ADDRESS option for socket on ", port, " port: ", err.message());
            return;
        }
        boost::asio::socket_base::receive_buffer_size rcvBufOpt;
        m_acceptor.get_option(rcvBufOpt, err);
        if (err)
        {
            LOGW(m_owner,
                 " Cannot get receive buffer size for TCP socket on ",
                 port, " port: ", err.message());
        }

        LOGI(m_owner,
             " Change the current value of receive buffer size for TCP socket on ",
             port, " port: ", rcvBufOpt.value(), " -> ", connections::tcp_buf_size);
        boost::asio::socket_base::receive_buffer_size newRcvBufOpt(connections::tcp_buf_size);
        m_acceptor.set_option(newRcvBufOpt, err);
        if (err)
        {
            LOGW(m_owner,
                 " Cannot set RCVBUF option for socket on ", port, " port: ", err.message());
        }
        boost::asio::socket_base::send_buffer_size sndBufOpt;
        m_acceptor.get_option(sndBufOpt, err);
        if (err)
        {
            LOGW(m_owner,
                 " Cannot get send buffer size for TCP socket on ",
                 port, " port: ", err.message());
        }

        LOGI(m_owner,
             " Change the current value of send buffer size for TCP socket on ",
             port, " port: ", sndBufOpt.value(), " -> ", connections::tcp_buf_size);
        boost::asio::socket_base::send_buffer_size newSndBufOpt(connections::tcp_buf_size);
        m_acceptor.set_option(newSndBufOpt, err);
        if (err)
        {
            LOGW(m_owner,
                 " Cannot set SNDBUF option for socket on ", port, " port: ", err.message());
        }

        m_acceptor.bind(endpoint, err);
        if (err)
        {
            LOGE(m_owner,
                 " Cannot bind to socket on ", port, " port: ",
                 err.message());
            return;
        }
        m_acceptor.listen(boost::asio::socket_base::max_listen_connections, err);
        if (err)
        {
            LOGE(m_owner,
                 " Cannot listen to socket on ", port, " port: ", err.message());
            return;
        }
    }

    void listener::startAccept(void)
    {
        m_nextConnPtr.reset(new connection<role::SERVER>(m_ioc));
        m_nextConnPtr->interfaces::connections::notifier::client::bind_over(this);
        m_nextConnPtr->changeState(connections::state::STARTED);
        m_acceptor.async_accept(boost::asio::make_strand(m_ioc),
                                boost::beast::bind_front_handler(&listener::onAccept, this));
    }

    void listener::onAccept(const boost::system::error_code &err,
                            tcp::socket sock)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        if (m_stopped)
        {
            LOGD(m_owner,
                 " Rejected new connection as HTTP server stops listen to new ones");
            return;
        }

        if (!err)
        {
            m_nextConnPtr->onAccept(std::move(sock));
        }
        else if (err != boost::asio::error::operation_aborted)
        {
            LOGW(m_owner, " Cannot accept connection: ", err.message());
        }
        else
        {
            return;
        }
        startAccept();
    }

} // namespace http