#pragma once

// Log includes.
#include "Message.hpp"
#include "StreamFormatBackup.hpp"

// STL includes.
#include <iomanip>
#include <memory>
#include <sstream>
#include <unordered_set>

namespace Log
{
    class Logger
    {
    public:
        friend class SyncLogger;

    public:
        virtual ~Logger() = default;

    public:
        /// @name Configuration methods
        /// @{

        void setLevel(const Level level);

        void blockTag(const char *tag);
        void blockTag(const std::string &tag);
        void blockTag(const Tag &tag);

        void unblockTag(const char *tag);
        void unblockTag(const std::string &tag);
        void unblockTag(const Tag &tag);

        void setToPrintTime(const bool value);
        void setToPrintThreadId(const bool value);
        void setToPrintLevel(const bool value);
        void setToPrintSourceLocation(const bool value);
        void setToPrintTag(const bool value);

        /// @}

    public:
        /// @name Logging methods
        /// @{

        template <typename... Args>
        void f(const Args &...args);

        template <typename... Args>
        void fatal(const Args &...args);

        template <typename... Args>
        void e(const Args &...args);

        template <typename... Args>
        void error(const Args &...args);

        template <typename... Args>
        void w(const Args &...args);

        template <typename... Args>
        void warn(const Args &...args);

        template <typename... Args>
        void i(const Args &...args);

        template <typename... Args>
        void info(const Args &...args);

        template <typename... Args>
        void d(const Args &...args);

        template <typename... Args>
        void debug(const Args &...args);

        template <typename... Args>
        void t(const Args &...args);

        template <typename... Args>
        void trace(const Args &...args);

        template <typename... Args>
        void log(const Level level, const Args &...args);

        /// @}

    private:
        template <typename... Args>
        void composeMessage(Message &message, const Args &...args);

        template <typename Arg, typename... Args>
        void composeMessageFromArgs(Message &message,
                                    std::ostringstream &stringStream,
                                    const Arg &arg,
                                    const Args &...args);

        template <typename... Args>
        void composeMessageFromArgs(Message &message,
                                    std::ostringstream &stringStream,
                                    const SourceLocation &sourceLocation,
                                    const Args &...args);

        template <typename... Args>
        void composeMessageFromArgs(Message &message,
                                    std::ostringstream &stringStream,
                                    const Tag &tag,
                                    const Args &...args);

        void composeMessageFromArgs(Message &message,
                                    const std::ostringstream &stringStream);

        std::string formatMessage(const Message &message);
        virtual void processMessage(std::string &&message) = 0;

    private:
        Level m_level = Level::All;

        std::unordered_set<std::string> m_blockedTags;

        bool m_isToPrintTime = true;
        bool m_isToPrintThreadId = true;
        bool m_isToPrintLevel = true;
        bool m_isToPrintSourceLocation = true;
        bool m_isToPrintTag = true;
    };

    using LoggerPtr = std::shared_ptr<Logger>;
} // namespace Log

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Log
{
    // Configuration methods:

    inline void Logger::setLevel(const Level level)
    {
        m_level = level;
    }

    inline void Logger::blockTag(const char *tag)
    {
        m_blockedTags.insert(tag);
    }

    inline void Logger::blockTag(const std::string &tag)
    {
        m_blockedTags.insert(tag);
    }

    inline void Logger::blockTag(const Tag &tag)
    {
        m_blockedTags.insert(tag);
    }

    inline void Logger::unblockTag(const char *tag)
    {
        m_blockedTags.erase(tag);
    }

    inline void Logger::unblockTag(const std::string &tag)
    {
        m_blockedTags.erase(tag);
    }

    inline void Logger::unblockTag(const Tag &tag)
    {
        m_blockedTags.erase(tag);
    }

    inline void Logger::setToPrintTime(const bool value)
    {
        m_isToPrintTime = value;
    }

    inline void Logger::setToPrintThreadId(const bool value)
    {
        m_isToPrintThreadId = value;
    }

    inline void Logger::setToPrintLevel(const bool value)
    {
        m_isToPrintLevel = value;
    }

    inline void Logger::setToPrintSourceLocation(const bool value)
    {
        m_isToPrintSourceLocation = value;
    }

    inline void Logger::setToPrintTag(const bool value)
    {
        m_isToPrintTag = value;
    }

    // Logging methods:

    template <typename... Args>
    inline void Logger::f(const Args &...args)
    {
        fatal(args...);
    }

    template <typename... Args>
    inline void Logger::fatal(const Args &...args)
    {
        log(Level::Fatal, args...);
    }

    template <typename... Args>
    inline void Logger::e(const Args &...args)
    {
        error(args...);
    }

    template <typename... Args>
    inline void Logger::error(const Args &...args)
    {
        log(Level::Error, args...);
    }

    template <typename... Args>
    inline void Logger::w(const Args &...args)
    {
        warn(args...);
    }

    template <typename... Args>
    inline void Logger::warn(const Args &...args)
    {
        log(Level::Warn, args...);
    }

    template <typename... Args>
    inline void Logger::i(const Args &...args)
    {
        info(args...);
    }

    template <typename... Args>
    inline void Logger::info(const Args &...args)
    {
        log(Level::Info, args...);
    }

    template <typename... Args>
    inline void Logger::d(const Args &...args)
    {
        debug(args...);
    }

    template <typename... Args>
    inline void Logger::debug(const Args &...args)
    {
        log(Level::Debug, args...);
    }

    template <typename... Args>
    inline void Logger::t(const Args &...args)
    {
        trace(args...);
    }

    template <typename... Args>
    inline void Logger::trace(const Args &...args)
    {
        log(Level::Trace, args...);
    }

    template <typename... Args>
    inline void Logger::log(Level level, const Args &...args)
    {
        // TODO[Log]: check for Level::Off and Level::All values?
        if (level > m_level)
        {
            return;
        }

        Message message(level);
        composeMessage(message, args...);
        if (m_blockedTags.find(message.tag) != m_blockedTags.cend())
        {
            return;
        }

        message.threadId = std::this_thread::get_id();
        std::string formattedMessage = formatMessage(message);
        processMessage(std::move(formattedMessage));
    }

    template <typename... Args>
    inline void Logger::composeMessage(Message &message, const Args &...args)
    {
        std::ostringstream stringStream;
        composeMessageFromArgs(message, stringStream, args...);
    }

    template <typename Arg, typename... Args>
    inline void Logger::composeMessageFromArgs(Message &message,
                                        std::ostringstream &stringStream,
                                        const Arg &arg,
                                        const Args &...args)
    {
        stringStream << arg;
        composeMessageFromArgs(message, stringStream, args...);
    }

    template <typename... Args>
    inline void Logger::composeMessageFromArgs(Message &message,
                                        std::ostringstream &stringStream,
                                        const SourceLocation &sourceLocation,
                                        const Args &...args)
    {
        message.sourceLocation = sourceLocation;
        composeMessageFromArgs(message, stringStream, args...);
    }

    template <typename... Args>
    inline void Logger::composeMessageFromArgs(Message &message,
                                        std::ostringstream &stringStream,
                                        const Tag &tag,
                                        const Args &...args)
    {
        message.tag = tag;
        composeMessageFromArgs(message, stringStream, args...);
    }

    inline void Logger::composeMessageFromArgs(Message &message,
                                        const std::ostringstream &stringStream)
    {
        message.text = std::move(stringStream.str());
    }

    inline std::string Logger::formatMessage(const Message &message)
    {
        std::ostringstream stringStream;

        if (m_isToPrintTime)
        {
            using namespace std::chrono;

            const auto time = Clock::to_time_t(message.time);
            tm localTime;
            localtime_r(&time, &localTime);
            const auto mseconds =
                duration_cast<milliseconds>(message.time.time_since_epoch()).count() % 1000;

            StreamFormatBackup sfb(stringStream);

            stringStream << '['
                         << std::put_time(&localTime, "%F %T.")
                         << std::right << std::setw(3) << std::setfill('0')
                         << mseconds
                         << ']';
        }

        if (m_isToPrintThreadId)
        {
            StreamFormatBackup sfb(stringStream);

            stringStream << "[TID=0x"
                         << std::hex
                         << message.threadId
                         << ']';
        }

        if (m_isToPrintLevel)
        {
            StreamFormatBackup sfb(stringStream);

            stringStream << '['
                         << std::left << std::setw(5) << std::setfill(' ')
                         << toString(message.level)
                         << ']';
        }

        if (m_isToPrintTag && !message.tag.isEmpty())
        {
            stringStream << " <"
                         << message.tag
                         << ">";
        }

        stringStream << ' ' << message.text;

        if (m_isToPrintSourceLocation && !message.sourceLocation.isEmpty())
        {
            stringStream << " ("
                         << message.sourceLocation
                         << ')';
        }

        return stringStream.str();
    }
} // namespace Log
