#pragma once

// UVR includes.
#include "UvrTypes.hpp"


/// A set of methods for working with the Called/Calling Party Number stored
/// as an @ref Uvr::OctetString, in accordance with ITU-T Recommendation Q.763.
namespace TRecQ763Number
{
    /// The valid length of the octet string for the hashed number.
    constexpr size_t hashedNumberSize = 10;

    /// The valid length of the octet string for the not hashed number.
    constexpr size_t notHashedNumberSize = 8;

    /// @brief Converts a number from an octet string representation to unsigned integer.
    /// @param[in] number Octet string with a number.
    /// @returns Number as unsigned integer.
    uint64_t toUInt64(const Uvr::OctetStringPtr &number);

    /// @brief Converts a number from an octet string representation to string.
    /// @param[in] number Pointer to octet string with hashed or not hashed number.
    /// @returns String with a number or number hash (ex. "89871112233" or "b763332ddda5b62c").
    std::string toStdString(const Uvr::OctetStringPtr &number);

    /// @brief Converts a part of the number from an octet string representation to string.
    /// @param[in] number Pointer to octet string with hashed or not hashed number.
    /// @param[in] nbBytes The number of bytes (excluding the first 2 bytes) that will be converted.
    /// @returns String with a part of the number or part of the number hash.
    std::string toStdString(const Uvr::OctetStringPtr &number, const size_t nbBytes);

    /// @brief Converts a number from a string representation to an octet string.
    /// @param[in] number String with a number (ex. "89871112233").
    /// @returns Generated @ref OctetStringPtr. Take a note that the first byte will
    ///          contain only Odd/Even Indicator. The second byte will be empty,
    ///          All subsequent bytes will contain the number encoded in accordance
    ///          with the recommendation.
    Uvr::OctetStringPtr toOctetStringPtr(const Uvr::Number &number);

    /// @brief Converts a number from a string representation to an octet string
    ///        with hash.
    /// @param[in] number String with a number (ex. "89871112233").
    /// @returns Generated @ref OctetStringPtr. Take a note that the first byte will
    ///          contain only Nature Of Address Indicator. The second byte will be empty,
    ///          All subsequent bytes will contain the hashed number.
    Uvr::OctetStringPtr toHashedOctetStringPtr(const Uvr::Number &number);

    /// @brief Converts a number from an octet string representation to an hashed octet string
    ///        with hash.
    /// @param[in] number number in a binary form aka @ref OctetsString object.
    /// @returns The input object if it is a hashed number already or generated @ref OctetStringPtr otherwise.
    ///          Take a note that the first byte will contain only Nature Of Address Indicator.
    ///          The second byte will be empty, all subsequent bytes will contain the hashed number.
    Uvr::OctetStringPtr toHashedOctetStringPtr(const Uvr::OctetStringPtr &number);

    /// @brief Checks a bit in an octet string that is an Odd/Even Indicator.
    /// @param[in] number Called/Calling Party Number in octet string representation.
    /// @returns @c true if the number contains an even number of digits, @c false otherwise.
    bool isEvenNbOfDigits(const Uvr::OctetStringPtr &number);

    /// @brief Checks a bits in an octet string that are a Nature Of Address Indicator.
    /// @param[in] number Called/Original Called Party Number in octet string representation.
    /// @returns @c true if the number is hashed, @c false otherwise.
    bool isNumberHashed(const Uvr::OctetStringPtr &number);

    /// @brief Gets number hash.
    /// @param[in] number Called/Original Called Party Number in octet string representation.
    /// @returns Pointer to number hash.
    const uint64_t *getNumberHashPtr(const Uvr::OctetStringPtr &number);

    /// @brief Checks that the numbers are equal. If both numbers are hashed or
    ///        not hashed simply compares the octet strings, otherwise compares hashes for numbers.
    /// @param[in] firstNumber  Called/Original Called Party Number in octet string representation.
    /// @param[in] secondNumber Called/Original Called Party Number in octet string representation.
    /// @returns @c true if the numbers are equal, @c false otherwise.
    bool areEqual(const Uvr::OctetStringPtr &firstNumber, const Uvr::OctetStringPtr &secondNumber);

    /// @brief Checks that the hashed number are equal to not hashed.
    /// @param[in] firstNumber  Called/Original Called Party Number in octet string representation.
    ///                         Warning: no additional check that the number is hashed.
    /// @param[in] secondNumber Called/Original Called Party Number in octet string representation.
    ///                         Warning: no additional check that the number is not hashed.
    /// @returns @c true if the numbers are equal, @c false otherwise.
    bool areHashedEqual(const Uvr::OctetStringPtr &hashedFirstNumber, const Uvr::OctetStringPtr &secondNumber);

} // TRecQ763Number
