#include "sftp_client.hpp"

namespace sftp
{
    client::client(const TypeInst &ownerId,
                   const std::string &user,
                   const std::string &address,
                   const std::string &port,
                   const std::string &privateKey,
                   const std::string &extraOptions) : m_ownerId(ownerId)
    {
        std::string sftpLogin;
        std::string sftpPortOption;
        std::string sftpPrivateKeyPathOption;
        std::string sftpOptions;

        if (user.empty())
        {
            LOGE(m_ownerId, " SFTP USER parameter is missed.");
            return;
        }

        if (address.empty())
        {
            LOGE(m_ownerId, " SFTP NODE ADDRESS parameter is missed.");
            return;
        }

        sftpLogin = user + "@" + address;

        if (!privateKey.empty())
        {
            sftpOptions += std::string("-i ") + privateKey + std::string(" ");
        }
        if (!port.empty())
        {
            sftpOptions += std::string("-P ") + port + std::string(" ");
        }
        if (!extraOptions.empty())
        {
            sftpOptions += extraOptions + std::string(" ");
        }

        m_sftpAuth = "sftp -b - " + sftpOptions + sftpLogin + " 2> /dev/null << EOF\n";
    }

    bool client::sendFile(const std::string &fileOrMask, const std::string &dstDir) const
    {
        if (!isInitialized())
        {
            LOGE(m_ownerId, " Cannot send file=\"", fileOrMask, "\" to remote directory=\"",
                 dstDir, "\": SFTP parameters are absent");
            return false;
        }

        std::string cmd = m_sftpAuth + "put " + fileOrMask + " " + dstDir + "/\nEOF";
        int result = system(cmd.c_str());
        if (result != 0)
        {
            LOGE(m_ownerId, " Failed send file(s) to remote node: result=", result, ", command=\"", cmd, "\"");
            return false;
        }
        return true;
    }

    bool client::getFile(const std::string &fileOrMask, const std::string &dstDir) const
    {
        if (!isInitialized())
        {
            LOGE(m_ownerId, " Cannot put remote file=\"", fileOrMask, "\" to local directory=\"",
                 dstDir, "\": SFTP parameters are absent");
            return false;
        }

        std::string cmd = m_sftpAuth + "get " + fileOrMask + " " + dstDir + "/\nEOF";
        int result = system(cmd.c_str());
        if (result != 0)
        {
            LOGE(m_ownerId, " Failed get file(s) from remote node: result=", result, ", command=\"", cmd, "\"");
            return false;
        }
        return true;
    }

    bool client::listDir(const std::string &dirOrMask, fs::files::list &files) const
    {
        if (!isInitialized())
        {
            LOGE(m_ownerId, " Cannot list remote directory=\"", dirOrMask, "\": SFTP parameters are absent");
            return false;
        }

        std::string cmd = m_sftpAuth + "ls -1 " + dirOrMask + "\nEOF";
        return fs::command::list(m_ownerId, cmd, files);
    }

    void client::getLatestFile(const std::string &dirOrMask, std::string &filename) const
    {
        filename.clear();
        fs::files::list files;
        if (!listDir(dirOrMask, files))
            return;
        // Last file in alphabetic order should be with newest timestamp
        if (!files.empty())
            filename = files.back();
    }

    bool client::isInitialized() const
    {
        return !m_sftpAuth.empty();
    }
} // namespace sftp
