#pragma once

// AFAP includes.
#include "components/timers/src/timer_object.hpp"

// STL includes.
#include <forward_list>
#include <unordered_map>

namespace timers
{

    namespace interfaces
    {
        INTERFACE_S(notifier,
                    INTERFACE_SERVER(void, notify_timeout, const timers::id &, const ::timers::time_point &);)
        INTERFACE_C(notifier,
                    INTERFACE_CLIENT(void, notify_timeout, const timers::id &, const ::timers::time_point &);)
    }

    class manager : public std::enable_shared_from_this<manager>,
                    public interfaces::manager::server,
                    public interfaces::notifier::client
    {
    public:
        explicit manager(boost::asio::io_context &ioc);

        void delTimerAndStop(id &timer_id);
        void delAndStopAll(void);

        ~manager() override;

        template <class T>
        id makeTimerAndStart(std::string name,
                             timers::duration duration)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            size_t numOfTimers = m_timers.size();
            if (!timers::id::valid(numOfTimers))
            {
                LOGERROR("Too many timers: size=", numOfTimers);
                return id();
            }

            timers::id timerId = find_free_timerId();
            std::shared_ptr<T> timer =
                std::make_shared<T>(m_ioc, name, duration, timerId);
            timer->::timers::interfaces::manager::client1::bind(shared_from_this());
            m_timers.emplace(timerId, timer);
            timer->start();
            return timerId;
        }

        void restartTimer(const id &id);

    protected:
        id find_free_timerId(void);

        void manage_timeout(const ::timers::id &id,
                            const ::timers::time_point &start_time) override;
        void manage_stopped(const ::timers::id &id) override;

        // Protector of shared data in multiple threads
        mutable std::recursive_mutex m_guard;
        // Context for asynchronous operations
        boost::asio::io_context &m_ioc;
        // The container to store information about timers
        std::unordered_map<id, timer_ptr> m_timers;
        // Timer ID counter
        id m_timerId_cntr;
        // Releases timer IDs
        std::forward_list<timers::id> m_rel_timerIds;
    };
}
