#include "timer_object.hpp"

// Boost includes.
#include <boost/bind.hpp>

namespace timers
{
    timer::timer(boost::asio::io_context &ioc,
                 const std::string &name,
                 timers::duration duration,
                 id timerId) : PreTimer(boost::asio::make_strand(ioc.get_executor()), duration),
                               boost::asio::steady_timer(m_strand, m_duration),
                               m_timerId(timerId),
                               m_name(name)
    {
    }

    timer::timer(const boost::asio::strand<boost::asio::io_context::executor_type> &strand,
                 const std::string &name,
                 timers::duration duration,
                 id timerId) : PreTimer(strand, duration),
                               boost::asio::steady_timer(m_strand, m_duration),
                               m_timerId(timerId),
                               m_name(name)
    {
    }

    void timer::handle(const boost::system::error_code &error)
    {
        // The timer was cancelled. OR the timer could be queued
        // for invocation, when it was cancel. In this case
        // handler are passed an error code that indicates
        // the successful completion of the wait operation.
        // That is why we need additional check for timer cancellation.
        if (error == boost::asio::error::operation_aborted || m_wasCancelled)
        {
            LOGDEBUG("The ", this, " was canceled");
            return;
        }
        else if (error)
        {
            LOGERROR("The ", this, " was stopped due to error: ", error.message());
            interfaces::manager::client1::manage_stopped(m_timerId);
            return;
        }
        LOGT("The ", this, " was triggered");

        // Start timer specific handling
        handle_internal();
    }

    void timer::start()
    {
        clock_gettime(CLOCK_REALTIME, &m_startTime);
        async_wait(boost::bind(&timers::timer::handle,
                               shared_from_this(),
                               boost::asio::placeholders::error));
        LOGDEBUG("The ", this, " was started");
    }

    void timer::stop(void)
    {
        m_wasCancelled = true;
        boost::asio::post(m_strand,
                          std::bind(&timer::stop_int, shared_from_this()));
    }

    void timer::stop_int(void)
    {
        boost::system::error_code err;
        cancel(err);
        if (err)
        {
            LOGERROR("Cannot cancel ", this, ": ", err.message());
            return;
        }
    }

    std::ostream &operator<<(std::ostream &os, const timer *t)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);

        logger << "timer[" << t->m_timerId << ";"
               << t->m_name << ";";
        const type &type = t->getType();
        switch (type)
        {
            case timers::type::PERIODIC:
            {
                logger << "PERIODIC";
                break;
            }
            case timers::type::ONESHOT:
            {
                logger << "ONESHOT";
                break;
            }
            default:
            {
                logger << "UNKOWN(" + std::to_string(static_cast<int>(type)) + ")";
            }
        }
        logger << ";";

        auto duration = t->m_duration;
        auto hours = std::chrono::duration_cast<std::chrono::hours>(duration);
        duration -= hours;
        auto mins = std::chrono::duration_cast<std::chrono::minutes>(duration);
        duration -= mins;
        auto secs = std::chrono::duration_cast<std::chrono::seconds>(duration);
        duration -= secs;
        auto msecs = std::chrono::duration_cast<std::chrono::milliseconds>(duration);

        if (hours > std::chrono::hours::zero())
            logger << hours.count() << "h";
        if (mins > std::chrono::minutes::zero())
            logger << mins.count() << "m";
        if (secs > std::chrono::seconds::zero())
            logger << secs.count() << "s";
        if (msecs > std::chrono::milliseconds::zero())
            logger << msecs.count() << "ms";

        logger << "]";
        os << trace.str();
        return os;
    }
} // namespace timers
