#pragma once

// AFAP includes.
#include "components/support/identifier.hpp"

// Boost includes.
#include <boost/asio.hpp>

// STL includes.
#include <memory>
#include <chrono>

#define MS(v) (v * ::timers::mega)

namespace timers
{
    class timer;
    using timer_ptr = std::shared_ptr<timer>;
    using duration = boost::asio::steady_timer::duration;

    struct Duration : public duration
    {
        using duration::duration;
        Duration(const duration &dur) : duration(dur){};
    };

    enum class type
    {
        PERIODIC,
        ONESHOT
    };

    constexpr int kilo = 1000;
    constexpr int mega = kilo * kilo;
    constexpr int giga = mega * kilo;

    using id = identifier;

    struct time_point : public timespec
    {
        /// @brief Constructs @c time_point object from the number of seconds
        ///        and nanoseconds
        /// @param sec The number of seconds
        /// @param nsec The number of nanoseconds
        time_point(time_t sec = 0, long nsec = 0)
            : timespec{sec, nsec}
        {
        }

        /// @brief Constructs @c time_point object from the duration
        /// @tparam Duration unit: std::chrono::nanoseconds, std::chrono::microseconds, ...
        /// @param[in] duration Duration in integer number of units (nanoseconds, microseconds, ...)
        template <class T>
        time_point(std::chrono::duration<time_t, T> duration)
        {
            auto sec = std::chrono::duration_cast<std::chrono::seconds>(duration);
            auto nsec = std::chrono::duration_cast<std::chrono::nanoseconds>(duration);
            nsec = nsec - sec;
            tv_sec = sec.count();
            tv_nsec = nsec.count();
        }

        /// @brief Constructs @c time_point object from string
        ///        with generic date format (YYYY-MM-DDTHH:MM:SS+HH:MM)
        /// @param timeline The string with date in generic date format
        time_point(const std::string &timeline);

        /// @brief Constructs @c time_point object from string
        ///        with specified format.
        /// @note  The format should be valid for std::get_time() function.
        ///        As format std::get_time() function ignores time zone,
        ///        time zone is considered as UTC.
        /// @param timeline The string with date
        /// @param format_str The specified date format
        time_point(const std::string &timeline, const char *format_str);

        time_point operator-(const time_point &other) const;
        time_point operator+(const time_point &other) const;
        bool operator>(const time_point &other) const;
        bool operator>=(const time_point &other) const;
        bool operator<(const time_point &other) const;
        bool operator<=(const time_point &other) const;
        bool operator==(const time_point &other) const;
        bool operator!=(const time_point &other) const;

        /// @brief Converts @c time_point object to string with
        ///        generic date format  (YYYY-MM-DDTHH:MM:SS+HH:MM)
        /// @return The date as string in generic date format
        std::string format() const;

        /// @brief Converts @c time_point object to string with
        ///        specified date format
        /// @note The format should be valid for std::put_time() function
        /// @param format_str The date format
        /// @return The date as string in specified date format
        std::string format(const char *format_str) const;

        /// @brief Converts @c time_point object to string with
        ///        specified UTC date format
        /// @note The format should be valid for std::put_time() function
        /// @param format_str The date format
        /// @return The date as string in specified UTC date format
        std::string utcFormat(const char *format_str) const;
    };
    std::ostream &operator<<(std::ostream &os, const time_point &obj);
    void now(time_point &obj);
    time_point now();
} // namespace timers
