#pragma once

// STL includes.
#include <future>

/// @brief This class is used to synchronize the receipt on request.
/// @tparam TKey Type of identifier for promise.
/// @tparam TValue Type of value for promise.
/// @note All functions of this class are thread-safe with unique lock.
template <class TKey, class TValue>
class PromisesMap
{
public:
    using Future = std::future<TValue>;

public:
    PromisesMap() = default;
    ~PromisesMap();

public:
    /// @brief Creates a promise associated with passed @a key if
    ///        it doesn't already exist.
    /// @param[in] key The identifier by which the promise can be
    ///                fulfilled later.
    /// @returns A valid @ref Future for the created promise, or an invalid
    ///          one if a promise with the same @a key already exists.
    Future add(const TKey& key);

    /// @brief Fulfills the promise if it was added.
    /// @param[in] key The identifier for which the promise was added.
    ///                After the promise is fulfilled, the identifier
    ///                can be reused.
    /// @param[in] result Promise result.
    void fulfill(const TKey& key, const TValue& result);

public:
    PromisesMap(const PromisesMap &) = delete;
    void operator=(const PromisesMap &) = delete;

private:
    using Promise = std::promise<TValue>;
    using PromisesInternalMap = std::unordered_map<TKey, Promise>;

private:
    mutable std::mutex m_promisesMutex;
    PromisesInternalMap m_promises;
};

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

template <class TKey, class TValue>
PromisesMap<TKey, TValue>::~PromisesMap()
{
    const std::lock_guard<std::mutex> lock(m_promisesMutex);
    m_promises.clear();
}

template <class TKey, class TValue>
typename PromisesMap<TKey, TValue>::Future PromisesMap<TKey, TValue>::add(const TKey& key)
{
    const std::lock_guard<std::mutex> lock(m_promisesMutex);

    const auto result = m_promises.emplace(key, Promise());
    if (result.second == false)
    {
        return Future();
    }

    return result.first->second.get_future();
}

template <class TKey, class TValue>
void PromisesMap<TKey, TValue>::fulfill(const TKey& key, const TValue& result)
{
    const std::lock_guard<std::mutex> lock(m_promisesMutex);

    const auto promisesIterator = m_promises.find(key);
    promisesIterator->second.set_value(result);
    m_promises.erase(promisesIterator);
}
