// Own include.
#include "UvrTypes.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/types/src/KeyValue.hpp"

#define TO_BIT_FLAGS(VAL) static_cast<Uvr::MetaInfo::BitFlags_t>(VAL)

namespace
{
    // The bit flags map for Meta Info is:
    //
    //   7   6   5   4   3   2   1   0   7   6   5   4   3   2   1   0
    // +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    // |           |DEF_POL| S |   CAPABILITY  | S |MNT| S |  PLC  | S | 0
    // +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    //
    // PLACE:      [0,  2 ] bits, 0  bit - isSet flag, [1,  2 ] bits - value
    // MAINT:      [3,  4 ] bits, 3  bit - isSet flag, [4,  4 ] bits - value
    // CAPABILITY: [5,  9 ] bits, 5  bit - isSet flag, [6,  9 ] bits - value
    // DEF_POLICY: [10, 12] bits, 10 bit - isSet flag, [11, 12] bits - value

    constexpr Uvr::MetaInfo::BitFlags_t s_isSetBitFlag = 0b1;
    constexpr Uvr::MetaInfo::BitFlags_t s_isSetBitShift = 1;

    const std::string s_placeKey = "PLACE";
    constexpr auto s_placeMinValue = TO_BIT_FLAGS(Uvr::UvrInfo::Place::External);
    constexpr auto s_placeMaxValue = TO_BIT_FLAGS(Uvr::UvrInfo::Place::Vipnet);
    constexpr uint8_t s_placeBitShift = 0;
    constexpr Uvr::MetaInfo::BitFlags_t s_placeBitMask = 0b11;

    const std::string s_statusKey = "MAINT";
    const std::vector<std::string> s_statusValues = {"FALSE", "TRUE"};
    constexpr uint8_t s_statusBitShift = 3;
    constexpr Uvr::MetaInfo::BitFlags_t s_statusBitMask = 0b1;

    const std::string s_capabilityKey = "CAPABILITY";
    constexpr auto s_capabilityMinValue = TO_BIT_FLAGS(Uvr::UvrInfo::Capability::Universal);
    constexpr auto s_capabilityMaxValue = TO_BIT_FLAGS(Uvr::UvrInfo::Capability::All);
    constexpr uint8_t s_capabilityBitShift = 5;
    constexpr Uvr::MetaInfo::BitFlags_t s_capabilityBitMask = 0b1111;

    const std::string s_defPolicyKey = "DEF_POLICY";
    constexpr auto s_defPolicyMinValue = TO_BIT_FLAGS(Uvr::UvrInfo::DefPolicy::NoChange);
    constexpr auto s_defPolicyMaxValue = TO_BIT_FLAGS(Uvr::UvrInfo::DefPolicy::Allowed);
    constexpr uint8_t s_defPolicyBitShift = 10;
    constexpr Uvr::MetaInfo::BitFlags_t s_defPolicyBitMask = 0b11;

    bool isValueInRange(const Uvr::MetaInfo::BitFlags_t value,
                        const Uvr::MetaInfo::BitFlags_t minValue,
                        const Uvr::MetaInfo::BitFlags_t maxValue,
                        const Uvr::MetaInfo::String &metaInfoStr,
                        const std::string &key);

    size_t getValueIndex(const std::string &value,
                         const std::vector<std::string> &allowedValues,
                         const Uvr::MetaInfo::String &metaInfoStr,
                         const std::string &key);

    template <class ValueType>
    std::optional<ValueType> getValue(Uvr::MetaInfo::BitFlags_t metaInfoBitFlags,
                                      const uint8_t valueBitShift,
                                      const Uvr::MetaInfo::BitFlags_t valueBitMask);

} // namespace

namespace Uvr
{
    CallInformationRequest::CallInformationRequest(const Number &numA,
                                                   const Number &numB,
                                                   const Number &numC,
                                                   const Number &numD,
                                                   const Date &date,
                                                   const RequestId_t &idReq,
                                                   const OptionalOperatorId &idSrc,
                                                   const OptionalOperatorId &idDst,
                                                   const OptionalCallMatchInterval &interval,
                                                   const CallId_t &callId,
                                                   const OptionalCallDuration &callDuration)
        : numA(numA),
          numB(numB),
          numC(numC),
          numD(numD),
          date(date),
          idReq(idReq),
          idSrc(idSrc),
          idDst(idDst),
          interval(interval),
          callId(callId),
          callDuration(callDuration)
    {
        if ((interval.has_value()) && (interval.value() < CallMatchInterval::min || interval.value() > CallMatchInterval::max))
        {
            this->interval.reset();
            throw std::range_error(std::string("Interval value of  ")
                                   + std::to_string(interval.value())
                                   + " is out of ["
                                   + std::to_string(CallMatchInterval::min)
                                   + ";"
                                   + std::to_string(CallMatchInterval::max)
                                   + "] range");
        }
    }

    CallInformationRequest::CallInformationRequest(const char *req, const char delimeter)
    {
        Csv::File::OptionalValue<Uvr::Number> numDOpt = {.value = numD,
                                                         .defaultValue = ""};
        Csv::File::OptionalValue<Uvr::Number> numCOpt = {.value = numC,
                                                         .defaultValue = ""};
        Csv::File::OptionalValue<Uvr::CallId_t> idCallOpt = {.value = callId,
                                                             .defaultValue = ""};
        Csv::File::parseRow(req,
                            delimeter,
                            numA,
                            numB,
                            numDOpt,
                            numCOpt,
                            date,
                            idReq,
                            idSrc,
                            idDst,
                            interval,
                            idCallOpt,
                            callDuration);
        if ((interval.has_value()) && (interval.value() < CallMatchInterval::min || interval.value() > CallMatchInterval::max))
        {
            interval.reset();
            throw std::range_error(std::string("Interval value of  ")
                                   + std::to_string(interval.value())
                                   + " is out of ["
                                   + std::to_string(CallMatchInterval::min)
                                   + ";" + std::to_string(CallMatchInterval::max)
                                   + "] range");
        }
    }

    CallInformationResponse::CallInformationResponse(const Number &numA,
                                                     const Number &numB,
                                                     const Number &numC,
                                                     const Number &numD,
                                                     const Date &date,
                                                     const Date &dateAct,
                                                     const RequestId_t &idReq,
                                                     const OptionalOperatorId &idSrc,
                                                     const OptionalOperatorId &idDst,
                                                     const ResponseCode_t &rspCode,
                                                     const OptionalVerificationRsp &vrfRsp,
                                                     const SessionId_t &sessionId,
                                                     const OptionalTransitActionCode &tActionCode,
                                                     const OptionalReleaseCode &releaseCode,
                                                     const OptionalCallDuration &callDuration,
                                                     const OptionalUvrId &idUvrT,
                                                     const CallId_t &callId,
                                                     const Uvr::VerdictTrigger_t &idRel)
        : numA(numA),
          numB(numB),
          numC(numC),
          numD(numD),
          date(date),
          dateAct(dateAct),
          idReq(idReq),
          idSrc(idSrc),
          idDst(idDst),
          rspCode(rspCode),
          vrfRsp(vrfRsp),
          sessionId(sessionId),
          tActionCode(tActionCode),
          releaseCode(releaseCode),
          callDuration(callDuration),
          idUvrT(idUvrT),
          callId(callId),
          idRel(idRel)
    {
    }

    bool CallInformationResponse::operator==(const CallInformationResponse &rhs) const
    {
        return ((numA == rhs.numA)
                && (numB == rhs.numB)
                && (numC == rhs.numC)
                && (numD == rhs.numD)
                && (date == rhs.date)
                && (dateAct == rhs.dateAct)
                && (idReq == rhs.idReq)
                && (idSrc == rhs.idSrc)
                && (idDst == rhs.idDst)
                && (rspCode == rhs.rspCode)
                && (vrfRsp == rhs.vrfRsp)
                && (sessionId == rhs.sessionId)
                && (tActionCode == rhs.tActionCode)
                && (releaseCode == rhs.releaseCode)
                && (callDuration == rhs.callDuration)
                && (idUvrT == rhs.idUvrT)
                && (callId == rhs.callId)
                && (idRel == rhs.idRel));
    }

    void CallRecord::update(const CallRecord &other)
    {
        if (other.idSrc.has_value())
        {
            idSrc = other.idSrc;
        }

        if (other.idDst.has_value())
        {
            idDst = other.idDst;
        }

        if (other.releaseCode.has_value())
        {
            releaseCode = other.releaseCode;
        }

        if (other.callDuration.has_value())
        {
            callDuration = other.callDuration;
        }
    }

    MetaInfo::BitFlags_t UvrInfo::parseMetaInfo(const MetaInfo::String &metaInfoStr)
    {
        KeyValue::PairsList keyValueList = KeyValue::parse(metaInfoStr);

        MetaInfo::BitFlags_t resultBitFlags = MetaInfo::emptyBitFlags;
        for (const auto &keyValue : keyValueList)
        {
            auto key = keyValue.first;
            std::transform(key.begin(), key.end(), key.begin(), ::toupper);

            auto value = keyValue.second;
            std::transform(value.begin(), value.end(), value.begin(), ::toupper);

            if (key.empty())
            {
                LOGW("Uvr Meta Info \"", metaInfoStr, "\" has empty key with value=\"", value, "\", will be skipped.");
                continue;
            }

            if (value.empty())
            {
                LOGW("Uvr Meta Info \"", metaInfoStr, "\" has empty value for key=\"", key, "\", will be skipped.");
                continue;
            }

            MetaInfo::BitFlags_t valueBitFlag = MetaInfo::emptyBitFlags;
            uint8_t valueBitShift = 0;
            try
            {
                if (key == s_placeKey)
                {
                    valueBitFlag = boost::lexical_cast<Uvr::MetaInfo::BitFlags_t>(value);
                    if (!isValueInRange(valueBitFlag, s_placeMinValue, s_placeMaxValue, metaInfoStr, key))
                    {
                        continue;
                    }
                    valueBitShift = s_placeBitShift;
                }
                else if (key == s_statusKey)
                {
                    const int valueIndex = getValueIndex(value, s_statusValues, metaInfoStr, key);
                    if (valueIndex < 0)
                    {
                        continue;
                    }
                    valueBitFlag = static_cast<Uvr::MetaInfo::BitFlags_t>(valueIndex);
                    valueBitShift = s_statusBitShift;
                }
                else if (key == s_capabilityKey)
                {
                    valueBitFlag = boost::lexical_cast<Uvr::MetaInfo::BitFlags_t>(value);
                    if (!isValueInRange(valueBitFlag, s_capabilityMinValue, s_capabilityMaxValue, metaInfoStr, key))
                    {
                        continue;
                    }
                    valueBitShift = s_capabilityBitShift;
                }
                else if (key == s_defPolicyKey)
                {
                    valueBitFlag = boost::lexical_cast<Uvr::MetaInfo::BitFlags_t>(value);
                    if (!isValueInRange(valueBitFlag, s_defPolicyMinValue, s_defPolicyMaxValue, metaInfoStr, key))
                    {
                        continue;
                    }
                    valueBitShift = s_defPolicyBitShift;
                }
                else
                {
                    LOGW("Uvr Meta Info \"", metaInfoStr, "\" has unsupported key=\"", key, "\" with value=\"", value, "\", will be skipped.");
                    continue;
                }
            }
            catch (...)
            {
                LOGW("Uvr Meta Info \"", metaInfoStr, "\" has invalid value=\"", value, "\" for key=\"", key, "\", will be skipped.");
                continue;
            }

            // Append isSet bit flag.
            valueBitFlag <<= s_isSetBitShift;
            valueBitFlag |= s_isSetBitFlag;

            // Shift value bit flag to its position.
            valueBitFlag <<= valueBitShift;

            // Append value bit flag to other flags.
            resultBitFlags |= valueBitFlag;
        }

        return resultBitFlags;
    }

    UvrInfo::OptionalPlace UvrInfo::getPlace(const MetaInfo::BitFlags_t metaInfoBitFlags)
    {
        return getValue<UvrInfo::Place>(metaInfoBitFlags, s_placeBitShift, s_placeBitMask);
    }

    UvrInfo::OptionalStatus UvrInfo::getStatus(const MetaInfo::BitFlags_t metaInfoBitFlags)
    {
        return getValue<UvrInfo::Status>(metaInfoBitFlags, s_statusBitShift, s_statusBitMask);
    }

    UvrInfo::OptionalCapability UvrInfo::getCapability(const MetaInfo::BitFlags_t metaInfoBitFlags)
    {
        return getValue<UvrInfo::Capability>(metaInfoBitFlags, s_capabilityBitShift, s_capabilityBitMask);
    }

    UvrInfo::OptionalDefPolicy UvrInfo::getDefPolicy(const MetaInfo::BitFlags_t metaInfoBitFlags)
    {
        return getValue<UvrInfo::DefPolicy>(metaInfoBitFlags, s_defPolicyBitShift, s_defPolicyBitMask);
    }

    std::ostream &operator<<(std::ostream &os, const CallInformationResponse &r)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);

        logger << r.numA << Csv::delimiter
               << r.numB << Csv::delimiter
               << r.numD << Csv::delimiter
               << r.numC << Csv::delimiter
               << r.date << Csv::delimiter
               << r.dateAct << Csv::delimiter
               << r.idReq << Csv::delimiter
               << toString(r.idSrc, Csv::no_value) << Csv::delimiter
               << toString(r.idDst, Csv::no_value) << Csv::delimiter
               << r.rspCode << Csv::delimiter
               << toString(r.vrfRsp, Csv::no_value) << Csv::delimiter
               << (r.sessionId == undefinedSessionId ? Csv::no_value : std::to_string(r.sessionId)) << Csv::delimiter
               << toString(r.tActionCode, Csv::no_value) << Csv::delimiter
               << toString(r.releaseCode, Csv::no_value) << Csv::delimiter
               << toString(r.callDuration, Csv::no_value) << Csv::delimiter
               << toString(r.idUvrT, Csv::no_value) << Csv::delimiter
               << r.callId << Csv::delimiter
               << (r.idRel == Uvr::VerdictTrigger::undefined ? Csv::no_value : std::to_string(r.idRel));
        os << trace.str();
        return os;
    }

    std::ostream &operator<<(std::ostream &os, const NodeStatus &status)
    {
        switch (status)
        {
            case NodeStatus::NotInitialized:
            {
                os << "NotInitialized";
                break;
            }
            case NodeStatus::Initalized:
            {
                os << "Initalized";
                break;
            }
            case NodeStatus::Starting:
            {
                os << "Starting";
                break;
            }
            case NodeStatus::Started:
            {
                os << "Started";
                break;
            }
            case NodeStatus::Stopping:
            {
                os << "Stopping";
                break;
            }
            case NodeStatus::Stopped:
            {
                os << "Stopped";
                break;
            }
        }
        return os;
    }

    std::ostream &operator<<(std::ostream &os, const CallRecord &callRecord)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);

        timers::time_point date;
        date.tv_sec = callRecord.date.tv_sec;
        date.tv_nsec = callRecord.date.tv_nsec;

        timers::time_point dateAct;
        dateAct.tv_sec = callRecord.dateAct.tv_sec;
        dateAct.tv_nsec = callRecord.dateAct.tv_nsec;

        logger << callRecord.numA << Csv::delimiter
               << callRecord.numB << Csv::delimiter
               << callRecord.numC << Csv::delimiter
               << callRecord.numD << Csv::delimiter
               << date.format() << Csv::delimiter
               << dateAct.format() << Csv::delimiter
               << Uvr::toString(callRecord.idSrc) << Csv::delimiter
               << Uvr::toString(callRecord.idDst) << Csv::delimiter
               << (callRecord.sessionId != Uvr::undefinedSessionId ? std::to_string(callRecord.sessionId) : Csv::no_value) << Csv::delimiter
               << static_cast<int>(callRecord.idRel) << Csv::delimiter
               << (callRecord.response != Uvr::VerificationResponse::undefined ? std::to_string(callRecord.response) : Csv::no_value) << Csv::delimiter
               << Uvr::toString(callRecord.releaseCode) << Csv::delimiter
               << Uvr::toString(callRecord.callDuration) << Csv::delimiter
               << Uvr::toString(callRecord.idUvrT) << Csv::delimiter
               << callRecord.callId << Csv::delimiter
               << (callRecord.callType != Uvr::CallType::undefined ? std::to_string(callRecord.callType) : Csv::no_value);
        os << trace.str();
        return os;
    }

} // namespace Uvr

namespace
{
    bool isValueInRange(const Uvr::MetaInfo::BitFlags_t value,
                        const Uvr::MetaInfo::BitFlags_t minValue,
                        const Uvr::MetaInfo::BitFlags_t maxValue,
                        const Uvr::MetaInfo::String &metaInfoStr,
                        const std::string &key)
    {
        if (value < minValue
            || value > maxValue)
        {
            LOGW("Uvr Meta Info \"", metaInfoStr, "\" has out of range value for key=\"", key, "\", will be skipped.",
                 "\n Expected: [", minValue, ", ", maxValue, "]",
                 "\n Actual: ", value, "\n");
            return false;
        }

        return true;
    }

    size_t getValueIndex(const std::string &value,
                         const std::vector<std::string> &allowedValues,
                         const Uvr::MetaInfo::String &metaInfoStr,
                         const std::string &key)
    {
        std::string allowedValuesStr;

        for (size_t index = 0; index < allowedValues.size(); ++index)
        {
            const auto &allowedValue = allowedValues.at(index);
            if (value == allowedValue)
            {
                return index;
            }

            allowedValuesStr += "\"" + allowedValue + "\"; ";
        }

        LOGW("Uvr Meta Info \"", metaInfoStr, "\" has invalid value for key=\"", key, "\", will be skipped.",
             "\n Expected: ", allowedValuesStr,
             "\n Actual: \"", value, "\"\n");

        return -1;
    }

    template <class ValueType>
    std::optional<ValueType> getValue(Uvr::MetaInfo::BitFlags_t metaInfoBitFlags,
                                      const uint8_t valueBitShift,
                                      const Uvr::MetaInfo::BitFlags_t valueBitMask)
    {
        metaInfoBitFlags >>= valueBitShift;

        // Check isSet flag.
        if ((metaInfoBitFlags & s_isSetBitFlag) == 0)
        {
            return std::nullopt;
        }

        metaInfoBitFlags >>= s_isSetBitShift;

        // Apply value bit mask.
        metaInfoBitFlags &= valueBitMask;

        return static_cast<ValueType>(metaInfoBitFlags);
    }

} // namespace
