#pragma once

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "components/worker_node/src/primitive.hpp"

// Boost includes.
#include <boost/endian/conversion.hpp>

#define COUNTERS_TAG "c"
#define RESET_TAG "r"

namespace primitive
{
    namespace Codec
    {
        /// Used to indicate how the primitive data is encoded.
        enum class Type : uint8_t
        {
            Binary, // 0
            Json,   // 1
            Unknown
        };

        std::string typeToStr(const Type type);

        /// @brief Encodes the passed primitive with the specified codec.
        /// @param[in] prim Pointer to primitive to encode. Warning: no check for nullptr.
        /// @param[in] codecType Type of codec that the primitive should be encoded with.
        /// @param[in] prependSize If @c true, then the first two bytes of the result will
        ///                        contain the size of the message in bytes. @c false by default.
        /// @returns Encoded data in case of success. If @a prependSize was set to @c true then
        ///          the first two bytes will containt the size of the rest message, the third byte
        ///          will indicate the type of codec used, all subsequent bytes will contain the
        ///          encoded primitive. If @a prependSize was set to @c false, then the first byte
        ///          will indicate the type of codec used, all subsequent bytes will contain the
        ///          encoded primitive. In case of an error, the return value will be empty.
        raw_data_t encode(const generic *prim, const Type codecType, const bool prependSize = false);

        /// @brief Decodes the passed data into a primitive.
        /// @param[in] rawData Data to decode. The first byte is expected to contain
        ///                    the type of codec used.
        /// @returns Decoded primitive in case of success, or @c nullptr in case of error.
        GenericPtr decode(const raw_data_t &rawData);

        namespace Binary
        {
            template <class T>
            void writeToBuff(uint8_t *&buff, const T &var)
            {
                boost::endian::endian_store<T, sizeof(T), boost::endian::order::little>(buff, var);
                buff += sizeof(T);
            }

            template <class T>
            void readFromBuff(const uint8_t *&buff, T &var)
            {
                var = boost::endian::endian_load<T, sizeof(T), boost::endian::order::little>(buff);
                buff += sizeof(T);
            }
        } // namespace Binary

        namespace Json
        {
            raw_data_t encode(const generic *prim);
            GenericPtr decode(const uint8_t *buff, const size_t buffSize);

            boost::json::value encodeHeader(const generic::Header &header);
            generic::HeaderPtr decodeHeader(const boost::json::value *headerValue);

            boost::json::value encodePmProviderReqisterReq(const generic *prim);
            boost::json::value encodeGetCountersReq(const generic *prim);
            boost::json::value encodeGetCountersRsp(const generic *prim);
            boost::json::value encodeGetStatisticsReq(const generic *prim);
            boost::json::value encodeGetStatisticsRsp(const generic *prim);
            boost::json::value encodeGetIncidentsRsp(const generic *prim);
            boost::json::value encodeGetIncidentsARsp(const generic *prim);
            boost::json::value encodeUploadNumbersInd(const generic *prim);
            boost::json::value encodeUploadNodesInd(const generic *prim);
            boost::json::value encodeUploadOperatorsInd(const generic *prim);
            boost::json::value encodeUploadPublicKeysInd(const generic *prim);
            boost::json::value encodeNotifyCallInfoRequest(const generic *prim);
            boost::json::value encodeNotifyCallInfoResponse(const generic *prim);
            boost::json::value encodeVerifyCallQuery(const generic *prim);
            boost::json::value encodeVerifyCallRequest(const generic *prim);
            boost::json::value encodeVerifyCallIndication(const generic *prim);
            boost::json::value encodeVerifyCallConfirm(const generic *prim);
            boost::json::value encodeVerifyCallResponse(const generic *prim);
            boost::json::value encodeVerifyCallVerdict(const generic *prim);
            boost::json::value encodeOutgoingCallIndication(const generic *prim);
            boost::json::value encodeQueryUvrIdsReq(const generic *prim);
            boost::json::value encodeQueryUvrIdsRsp(const generic *prim);
            boost::json::value encodeQueryUvrInfoReq(const generic *prim);
            boost::json::value encodeQueryUvrInfoRsp(const generic *prim);
            boost::json::value encodeQueryHubInfoReq(const generic *prim);
            boost::json::value encodeQueryHubInfoRsp(const generic *prim);
            boost::json::value encodeQueryOperatorInfoReq(const generic *prim);
            boost::json::value encodeQueryOperatorInfoRsp(const generic *prim);
            boost::json::value encodeExecCmdReq(const generic *prim);
            boost::json::value encodeExecCmdRsp(const generic *prim);
            boost::json::value encodeRegisterCmdInd(const generic *prim);
            boost::json::value encodeStatusChangeInd(const generic *prim);

            GenericPtr decodeGetStatisticsReq(const boost::json::value *bodyValue);
            GenericPtr decodeGetStatisticsRsp(const boost::json::value *bodyValue);
            GenericPtr decodeExecCmdReq(const boost::json::value *bodyValue);
            GenericPtr decodeExecCmdRsp(const boost::json::value *bodyValue);
            GenericPtr decodeRegisterCmdInd(const boost::json::value *bodyValue);

            /// @brief Extracts an optional value.
            /// @param[in] parentValue Parsed JSON values to search with.
            /// @param[in] tag Tag for the value to search.
            /// @param[in] var This the value to return.
            /// @returns The return value states if the the value was found or not.
            template <class T>
            bool parseOptionalValue(const boost::json::value *parentValue,
                                    const std::string &tag,
                                    T &var)
            {
                boost::json::error_code errorCode;
                const boost::json::value *tagValue = parentValue->find_pointer("/" + tag, errorCode);
                if (errorCode)
                {
                    return false;
                }
                else
                {
                    var = boost::json::value_to<T>(*tagValue);
                    return true;
                }
            }

            /// @brief Extracts a mandatory value.
            /// @param[in] parentValue Parsed JSON values to search with.
            /// @param[in] tag Tag for the value to search.
            /// @param[in] var This the value to return.
            /// @returns The return value states if the the value was found or not.
            template <class T>
            bool parseValue(const boost::json::value *parentValue, const std::string &tag, T &var)
            {
                bool wasFound = parseOptionalValue(parentValue, tag, var);
                if (!wasFound)
                {
                    LOGERROR("primitive::Codec::Json::parseValue can't find a tag:\"", tag, "\"",
                             "\n Value: ", boost::json::serialize(*parentValue));
                }

                return wasFound;
            }

        } // namespace Json

    } // namespace Codec

} // namespace primitive
