#include "components/worker_node/src/node.hpp"

// AFAP includes.
#include "components/worker_node/src/config_parameters.hpp"

namespace workers
{
    node::NodesMap node::s_nodes;
    std::shared_mutex node::s_nodesMutex;

    node::node(const TypeInst &ownTypeInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config,
               const Uvr::UvrId_t &uvrId)
        : m_ownTypeInst(ownTypeInst),
          m_logTag(m_ownTypeInst),
          m_config(config),
          m_uvrId(uvrId),
          m_ioContext(ioc),
          m_timersManager(std::make_shared<timers::manager>(m_ioContext))
    {
        m_timersManager->::timers::interfaces::notifier::client::bind(this);

        if (m_uvrId != 0)
        {
            return; // the value was passed via constructor parameter.
        }
        m_uvrId = m_config.getValue<Uvr::UvrId_t>(UVR_ID,
                                                  Uvr::undefinedUserId,
                                                  Log::Level::Error);
        if (m_uvrId == Uvr::undefinedUserId)
        {
            LOGERROR(m_ownTypeInst, " Cannot read UVR_ID.");
        }
    }

    void node::start(void)
    {
        std::unique_lock<std::shared_mutex> lock(s_nodesMutex);
        s_nodes.emplace(m_ownTypeInst, shared_from_this());
    }

    void node::stop()
    {
        {
            std::unique_lock<std::recursive_mutex> sLock(s_stopMutex);
            if (stopped)
            {
                return;
            }
            m_timersManager->delAndStopAll();
            stopped = true;
        }
        std::unique_lock<std::shared_mutex> lock(s_nodesMutex);
        auto nodeSearch = s_nodes.find(m_ownTypeInst);
        if (nodeSearch == s_nodes.end())
        {
            LOGDEBUG(m_ownTypeInst,
                     " Failed to delete own node on stop");
            return;
        }
        s_nodes.erase(nodeSearch);
        LOGINFO(m_ownTypeInst, " node::stop done.");
    }

    const TypeInst &node::getTypeInst()
    {
        return m_ownTypeInst;
    }

    void node::send(primitive::GenericPtr prim)
    {
        {
            std::unique_lock<std::recursive_mutex> sLock(s_stopMutex);
            if (stopped)
            {
                LOGDEBUG(m_ownTypeInst, " Drop primitive with id=", prim->m_header->id,
                         " because STOPPED");
                return;
            }
        }
        prim->m_header->from = m_ownTypeInst;
        std::unique_lock<std::recursive_mutex> sLock(s_strandsMutex);
        auto strandSearch = s_strands.find(prim->m_header->to);
        if (strandSearch == s_strands.end())
        {
            std::shared_lock<std::shared_mutex> nLock(s_nodesMutex);
            auto nodeSearch = s_nodes.find(prim->m_header->to);
            if (nodeSearch == s_nodes.end())
            {
                LOGINFO(m_ownTypeInst,
                        " Can't find pointer to receiver node with id=", prim->m_header->to,
                        " on sending primitive with id=", prim->m_header->id);
                return;
            }

            auto result = s_strands.emplace(prim->m_header->to,
                                            StrandNodePair(boost::asio::make_strand(m_ioContext.get_executor()),
                                                           nodeSearch->second));
            LOGINFO(m_ownTypeInst,
                    " Added strand towards ", prim->m_header->to,
                    " on sending primitive with id=", prim->m_header->id);
            strandSearch = result.first;
        }

        boost::asio::post(strandSearch->second.first,
                          std::bind(&node::handlePrimitive, strandSearch->second.second,
                                    prim));
        LOGTRACE(m_ownTypeInst,
                 " Forward primitive=", primitive::id2str(prim->m_header->id), " to ", prim->m_header->to);
    }

    void node::stopEx()
    {
        auto prim = std::make_shared<primitive::WorkerStopReq>();
        prim->m_header->from = m_ownTypeInst;
        prim->m_header->to = m_ownTypeInst;
        send(prim);
    }

    void node::handlePrimitive(const primitive::GenericPtr &prim)
    {
        if (prim->m_header->id == primitive::WORKER_STOP_REQ)
        {
            LOGDEBUG(m_ownTypeInst, " handle WORKER_STOP_REQ from ",
                     prim->m_header->from);
            stop();
        }
        else
        {
            {
                std::unique_lock<std::recursive_mutex> sLock(s_stopMutex);
                if (stopped)
                {
                    LOGDEBUG(m_ownTypeInst, " Drop handlePrimitive with id=", prim->m_header->id,
                             " because STOPPED");
                    return;
                }
            }
            handleSpecificPrimitive(prim);
        }
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        LOGERROR(m_ownTypeInst,
                 " Has no handler for primitive with id: ", prim->m_header->id);
    }

    void node::notify_timeout(const ::timers::id &id,
                              const ::timers::time_point &start_time)
    {
        handle_timeout(id, start_time);
    }

    void node::handle_timeout(const ::timers::id &id,
                              const ::timers::time_point &start_time)
    {
    }

    void killAll()
    {
        // sending SIGUSR1 itself to stop UVR
        kill(getpid(), SIGUSR1);
    }

} // namespace workers
