#pragma once

// AFAP includes.
#include "components/config/src/config_manager.hpp"
#include "components/timers/src/timer_manager.hpp"
#include "components/worker_node/src/sender.hpp"

// STL includes
#include <map>
#include <shared_mutex>

namespace workers
{
    class node : public std::enable_shared_from_this<node>,
                 public workers::interfaces::sender::server,
                 public ::timers::interfaces::notifier::server
    {
    public:
        virtual void start();
        void stopEx();

        const TypeInst &getTypeInst() override;

    protected:
        virtual void stop();
        explicit node(const TypeInst &ownTypeInst,
                      boost::asio::io_context &ioc,
                      const configuration::manager &config,
                      const Uvr::UvrId_t &uvrId = 0);

        /// @brief  Sends a primitive to specified node.
        void send(primitive::GenericPtr prim) override;

        /// @brief Should be overriden to handle worker specific primitives.
        virtual void handleSpecificPrimitive(const primitive::GenericPtr &prim) = 0;

    private:
        /// @brief Handles base primitives for any kind of worker or calls
        ///        handleSpecificPrimitive(const primitive::GenericPtr &prim).
        void handlePrimitive(const primitive::GenericPtr &prim);

        void notify_timeout(const ::timers::id &id,
                            const ::timers::time_point &start_time) override;
        virtual void handle_timeout(const ::timers::id &id,
                                    const ::timers::time_point &start_time);

    protected:
        const TypeInst m_ownTypeInst;                     ///< Own type instance.
        const Log::Tag m_logTag;                          ///< Tag for logs.
        const configuration::manager &m_config;           ///< Reference to configuration.
        Uvr::UvrId_t m_uvrId;                             ///< Own VFN identifier.
        boost::asio::io_context &m_ioContext;             ///< Reference to context for asynchronous operations.
        std::shared_ptr<timers::manager> m_timersManager; ///< An entity to handle timers.
        bool stopped{false};                              ///< Flag on the node stop procedure was triggered

    private:
        using Strand = boost::asio::strand<boost::asio::io_context::executor_type>;
        using StrandNodePair = std::pair<Strand, std::shared_ptr<node>>;
        std::recursive_mutex s_strandsMutex;
        std::recursive_mutex s_stopMutex;
        std::map<TypeInst, StrandNodePair> s_strands;

        using NodesMap = std::unordered_map<TypeInst, std::shared_ptr<node>>;
        static std::shared_mutex s_nodesMutex;
        static NodesMap s_nodes;
    };

    template <class Node, int MaxNumOfInsts>
    using NodesArray = std::array<std::shared_ptr<Node>, MaxNumOfInsts>;

    template <class Node, int NodeType, int MaxNumOfInsts>
    class api
    {
    protected:
        static NodesArray<Node, MaxNumOfInsts> s_instances;

    public:
        api() = delete;
        api(api const &) = delete;
        api &operator=(api const &) = delete;

        static Node *getNode(std::size_t instId = 0)
        {
            return s_instances[instId].get();
        }

        static int init(boost::asio::io_context &ioc,
                        configuration::manager &config)
        {
            // Find first free instance.
            std::size_t instId = 0;
            while ((s_instances[instId] != nullptr) && (instId < s_instances.size()))
            {
                ++instId;
            }

            if (instId == MaxNumOfInsts)
            {
                LOGERROR("Cannot create the instance of ", nodeType2str(NodeType),
                         ": the maximum number of instances(", MaxNumOfInsts,
                         ") is reached");
                return workers::undefinedInstId;
            }
            s_instances[instId] = std::make_shared<Node>(TypeInst(NodeType, instId), ioc, config);

            return instId;
        }

        static bool start(std::size_t instId = 0)
        {
            if (instId >= s_instances.size() || !(s_instances[instId]))
            {
                LOGERROR("[on start] ", TypeInst(NodeType, instId),
                         " node was not initialized ");
                return false;
            }
            s_instances[instId]->start();

            return true;
        }

        static void stop(std::size_t instId = 0)
        {
            if (instId < s_instances.size() && s_instances[instId])
            {
                s_instances[instId]->stopEx();
            }
        }

        static void reset(std::size_t instId = 0)
        {
            if (instId < s_instances.size() && s_instances[instId])
            {
                s_instances[instId].reset();
            }
        }
    };

    void killAll();

    template <class Node, int NodeType, int MaxNumOfInsts>
    NodesArray<Node, MaxNumOfInsts> api<Node, NodeType, MaxNumOfInsts>::s_instances;

} // namespace workers
