#pragma once

// AFAP includes.
#include "components/db/src/types.hpp"
#include "components/worker_node/src/typeInst.hpp"

// STL includes.
#include <string>
#include <vector>

namespace primitive
{
    enum type
    {
        WHOAMI,                   // 0
        WORKER_STOP_REQ,          // 1
        GET_INCIDENTS_REQ,        // 2
        GET_INCIDENTS_RSP,        // 3
        GET_STATISTICS_REQ,       // 4
        GET_STATISTICS_RSP,       // 5
        PM_PROVIDER_REGISTER_REQ, // 6
        GET_COUNTERS_REQ,         // 7
        GET_COUNTERS_RSP,         // 8
        UPLOAD_NUMBERS_IND,       // 9
        UPLOAD_NODES_IND,         // 10
        OUTGOING_CALL_IND,        // 11
        VERIFY_CALL_QUERY,        // 12
        VERIFY_CALL_VRD,          // 13
        VERIFY_CALL_REQ,          // 14
        VERIFY_CALL_RSP,          // 15
        VERIFY_CALL_IND,          // 16
        VERIFY_CALL_CNF,          // 17
        NOTIFY_CALL_INFO_REQ,     // 18
        NOTIFY_CALL_INFO_RSP,     // 19
        VERIFY_REQ_ABORT,         // 20
        UPLOAD_PUBLIC_KEYS_IND,   // 21
        SEND_INCIDENTSA_IND,      // 22
        QUERY_UVR_IDS_REQ,        // 23
        QUERY_UVR_IDS_RSP,        // 24
        QUERY_UVR_INFO_REQ,       // 25
        QUERY_UVR_INFO_RSP,       // 26
        QUERY_HUB_INFO_REQ,       // 27
        QUERY_HUB_INFO_RSP,       // 28
        UPLOAD_OPERATORS_IND,     // 29
        QUERY_OPERATOR_INFO_REQ,  // 30
        QUERY_OPERATOR_INFO_RSP,  // 31
        EXEC_CMD_REQ,             // 32
        EXEC_CMD_RSP,             // 33
        REGISTER_CMD_IND,         // 34
        STATUS_CHANGE_IND,        // 35
        UNKNOWN
    };

    using hash_t = std::string;
    using raw_data_t = std::string;
    using ServiceKey_t = uint8_t;
    using SessionKey_t = uint32_t;
    using SessionKeyOptional_t = std::optional<SessionKey_t>;
    using ReleaseCode_t = uint8_t;
    using AdditionalCode_t = uint8_t;
    using ErrorCode_t = uint8_t;
    using ConnectionId_t = uint64_t;

    std::string id2str(uint16_t id);

    class generic
    {
    public:
        using ulpKey_t = Uvr::SessionId_t;

        struct Header
        {
            uint16_t id{UNKNOWN};
            TypeInst from{0};
            TypeInst to{0};
            ulpKey_t ulpKey{0};

            explicit Header(const uint16_t &_id,
                            const TypeInst &_from,
                            const TypeInst &_to,
                            const ulpKey_t &_ulpKey);

            bool operator==(const Header &other) const;
        };

        using HeaderPtr = std::shared_ptr<Header>;

        HeaderPtr m_header;

        explicit generic(const uint16_t &_id,
                         const TypeInst &_from,
                         const TypeInst &_to,
                         const ulpKey_t &_ulpKey);

        explicit generic(const uint16_t &_id);

        explicit generic();

        virtual ~generic() = default;
    };

    using GenericPtr = std::shared_ptr<generic>;

    // WHOAMI
    struct Whoami : public generic
    {
        Whoami();
    };

    // WORKER_STOP_REQ
    struct WorkerStopReq : public generic
    {
        WorkerStopReq();
    };

    // GET_INCIDENTS_REQ
    struct GetIncidentsReq : public generic
    {
        GetIncidentsReq();
    };

    // GET_INCIDENTS_RSP
    struct GetIncidentsRsp : public generic
    {
        std::string filename{""};

        explicit GetIncidentsRsp(const std::string &filename = "");
    };

    // SEND_INCIDENTSA_IND
    struct GetIncidentsARsp : public generic
    {
        std::string filename{""};

        explicit GetIncidentsARsp(const std::string &filename = "");
    };

    // UPLOAD_NUMBERS_IND
    struct UploadNumbersInd : public generic
    {
        std::string numFile{""};
        Uvr::Strings deltas;

        UploadNumbersInd(const std::string &numFile = "",
                         const Uvr::Strings &deltas = {});
    };

    // UPLOAD_NODES_IND
    struct UploadNodesInd : public generic
    {
        Uvr::Strings files;

        explicit UploadNodesInd(const Uvr::Strings &files = {});
    };

    // UPLOAD_OPERATORS_IND
    struct UploadOperatorsInd : public generic
    {
        Uvr::Strings files;

        explicit UploadOperatorsInd(const Uvr::Strings &files = {});
    };

    // OUTGOING_CALL_IND: APIGW -> VFS
    struct OutgoingCallIndication : public generic
    {
        Uvr::OctetStringPtr callingNumber;
        Uvr::OctetStringPtr calledNumber;

        OutgoingCallIndication(Uvr::OctetStringPtr callingNumber,
                               Uvr::OctetStringPtr calledNumber);
    };

    // VERIFY_CALL_QUERY: APIGW -> VFC
    struct VerifyCallQuery : public generic
    {
        Uvr::OperatorId_t idSrc;
        Uvr::OctetStringPtr callingNumber;
        Uvr::OctetStringPtr calledNumber;
        Uvr::OctetStringPtr originalCalledNumber;
        Uvr::OctetStringPtr genericNumber;
        Uvr::CallId_t callId;

        VerifyCallQuery(const Uvr::OperatorId_t idSrc,
                        const Uvr::OctetStringPtr &callingNumber,
                        const Uvr::OctetStringPtr &calledNumber,
                        const Uvr::OctetStringPtr &originalCalledNumber,
                        const Uvr::OctetStringPtr &genericNumber,
                        const Uvr::CallId_t callId);
    };

    using VerifyCallQueryPtr = std::shared_ptr<VerifyCallQuery>;

    // VERIFY_CALL_VERDICT: VFC -> APIGW
    struct VerifyCallVerdict : public generic
    {
        Uvr::VerificationVerdict_t rlc;
        Uvr::UvrId_t uvrT;

        explicit VerifyCallVerdict(Uvr::VerificationVerdict_t rlc,
                                   Uvr::UvrId_t uvrT);
    };

    // VERIFY_CALL_REQUEST: VFC -> VFGW
    struct VerifyCallRequest : public generic
    {
        VerifyCallQueryPtr verifyCallQuery;
        Uvr::UvrId_t uvrP;
        Uvr::UvrId_t uvrS;

        VerifyCallRequest(VerifyCallQueryPtr verifyCallQuery,
                          Uvr::UvrId_t uvrP,
                          Uvr::UvrId_t uvrS);
    };

    // VERIFY_CALL_RESPONSE: VFGW -> VFC
    struct VerifyCallResponse : public generic
    {
        Uvr::VerificationVerdict_t verificationVerdict;
        Uvr::VerificationVerdict_t incidentVerdict;
        Uvr::UvrId_t uvrT;

        explicit VerifyCallResponse(Uvr::VerificationVerdict_t verificationVerdict,
                                    Uvr::VerificationVerdict_t incidentVerdict,
                                    Uvr::UvrId_t uvrT);
    };

    // VERIFY_CALL_IND: VFGW -> VFS
    struct VerifyCallIndication : public generic
    {
        ConnectionId_t connId;
        ServiceKey_t serviceKey;
        Uvr::UvrId_t uvrO;
        Uvr::UvrId_t uvrT;
        SessionKeyOptional_t sessionKey;
        Uvr::OctetStringPtr callingNumber;
        Uvr::OctetStringPtr calledNumber;
        Uvr::OctetStringPtr originalCalledNumber;

        VerifyCallIndication(ConnectionId_t connId,
                             ServiceKey_t serviceKey,
                             Uvr::UvrId_t uvrO,
                             Uvr::UvrId_t uvrT,
                             SessionKeyOptional_t sessionKey,
                             Uvr::OctetStringPtr callingNumber,
                             Uvr::OctetStringPtr calledNumber,
                             Uvr::OctetStringPtr originalCalledNumber);
    };

    using VerifyCallIndicationPtr = std::shared_ptr<VerifyCallIndication>;

    // VERIFY_CALL_CNF: VFS -> VFGW
    struct VerifyCallConfirm : public generic
    {
        VerifyCallIndicationPtr verifyCallIndication;
        ReleaseCode_t releaseCode;
        AdditionalCode_t additionalCode;
        ErrorCode_t errorCode;

        VerifyCallConfirm(VerifyCallIndicationPtr verifyCallIndication,
                          ReleaseCode_t releaseCode,
                          AdditionalCode_t additionalCode = 0,
                          ErrorCode_t errorCode = 0);
    };

    // NOTIFY_CALL_INFO_REQ: CCNGW -> APIGW
    struct NotifyCallInfoRequest : public generic
    {
        Uvr::UvrId_t uvrId;
        Uvr::Strings fileNamesList;

        explicit NotifyCallInfoRequest(const Uvr::UvrId_t uvrId,
                                       const Uvr::Strings &fileNamesList = {});
    };

    // NOTIFY_CALL_INFO_RSP: APIGW -> CCNGW
    struct NotifyCallInfoResponse : public generic
    {
        std::string filename{""};

        explicit NotifyCallInfoResponse(const std::string &filename = "");
    };

    // VERIFY_REQ_ABORT: VFC -> VFGW
    struct VerifyRequestAbort : public generic
    {
        VerifyRequestAbort();
    };

    // UPLOAD_PUBLIC_KEYS_IND
    struct UploadPublicKeysInd : public generic
    {
        Uvr::Strings hubFilesToCopy;
        Uvr::Strings nodeFilesToCopy;
        Uvr::Strings hubFilesToRemove;
        Uvr::Strings nodeFilesToRemove;

        UploadPublicKeysInd(const Uvr::Strings &hubFilesToCopy = {},
                            const Uvr::Strings &nodeFilesToCopy = {},
                            const Uvr::Strings &hubFilesToRemove = {},
                            const Uvr::Strings &nodeFilesToRemove = {});
    };

    // QUERY_UVR_IDS_REQ: APIGW -> VFC
    struct QueryUvrIdsReq : public generic
    {
        Uvr::OctetStringPtr number;

        QueryUvrIdsReq(const Uvr::OctetStringPtr &number);
    };

    // QUERY_UVR_IDS_RSP: VFC -> APIGW
    struct QueryUvrIdsRsp : public generic
    {
        Uvr::UvrIds uvrIds;

        QueryUvrIdsRsp(const Uvr::UvrId_t &primary,
                       const Uvr::UvrId_t &secondary);
    };

    struct QueryUvrInfoReq : public generic
    {
        Uvr::UvrId_t uvrId;

        QueryUvrInfoReq(const Uvr::UvrId_t &uvrId);
    };

    struct QueryUvrInfoRsp : public generic
    {
        Uvr::UvrInfo uvrInfo;

        QueryUvrInfoRsp(const ::db::uvr::record &rec);
    };

    struct QueryHubInfoReq : public generic
    {
        Uvr::Gt_t hubId;

        QueryHubInfoReq(const Uvr::Gt_t &uvrId);
    };

    struct QueryHubInfoRsp : public generic
    {
        Uvr::HubInfo hubInfo;

        QueryHubInfoRsp(const ::db::hub::record &rec);
    };


    struct QueryOperatorInfoReq : public generic
    {
        Uvr::OperatorId_t operatorId;

        QueryOperatorInfoReq(const Uvr::OperatorId_t &operatorId);
    };

    struct QueryOperatorInfoRsp : public generic
    {
        Uvr::OperatorInfo operatorInfo;

        QueryOperatorInfoRsp(const ::db::Operator::Record &rec);
    };

    struct ExecCmdReq : public generic
    {
        std::string cmd;
        Uvr::Strings cmdArgs;

        ExecCmdReq(const std::string &command,
                   const Uvr::Strings &commandArgs);
    };

    struct ExecCmdRsp : public generic
    {
        Uvr::Strings responses;

        ExecCmdRsp(Uvr::Strings &&responses);
    };

    struct RegisterCmdInd : public generic
    {
        std::string cmd;

        RegisterCmdInd(const std::string &command);
    };

    struct StatusChangeInd : public generic
    {
        Uvr::NodeStatus status;

        StatusChangeInd(const Uvr::NodeStatus &status);
    };

} // namespace primitive
