#pragma once

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "components/timers/src/timer_types.hpp"
#include "UvrTypes.hpp"
#include "components/worker_node/src/typeInst.hpp"

// STL includes.
#include <unordered_map>

namespace transaction
{
    using id = Uvr::SessionId_t;
    using manager_id = uint8_t;

    enum class errors
    {
        OK,
        FAIL
    };

    /// Transaction ID consists of:
    ///   7  6  5  4  3  2  1  0  7  6  5  4  3  2  1  0
    /// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    /// |   Type    |Inst |    ManagerId    |  ...      | 48
    /// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    /// |          VFN ID             |      ...        | 32
    /// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    /// |              ...Transaction ID                | 16
    /// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    /// |                                               | 0
    /// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    constexpr id transId_mask = 0x3FFFFFFFFF;

    constexpr id undef_value = 0;
    constexpr id min_cntr = 1;
    constexpr id max_cntr = 0x3FFFFFFFFF;

    class base_data
    {
        timers::time_point m_start;
        timers::time_point m_stop;

    public:
        base_data() = default;
        virtual ~base_data() = default;

        timers::time_point duration() const
        {
            return (m_stop - m_start);
        }

        void set_start()
        {
            timers::now(m_start);
        }

        const timers::time_point &get_start() const
        {
            return m_start;
        }

        void set_stop()
        {
            timers::now(m_stop);
        }

        const timers::time_point &get_stop() const
        {
            return m_stop;
        }
    };

    template <class T>
    class manager
    {
        id m_transId_cntr{min_cntr};

    protected:
        const TypeInst m_ownInst;
        manager_id m_mngrId;
        Uvr::UvrId_t m_uvrId;

        mutable std::recursive_mutex m_guard;

    private:
        std::unordered_map<id, T> m_register;

        id generateTransactionId()
        {
            id key = (m_transId_cntr++) & transId_mask; // 38 bits field
            key |= (id)(m_uvrId & 0x3FFF) << 38;        // 14 bits field
            key |= (id)(m_mngrId & 0x3F) << 52;         // 6 bits field
            key |= (id)(m_ownInst.value & 0x3F) << 58;  // 6 bits field
            if (m_transId_cntr == max_cntr)
            {
                m_transId_cntr = min_cntr;
            }
            return key;
        }

    public:
        manager(const TypeInst &ownInst,
                manager_id mngrId,
                Uvr::UvrId_t uvrId)
            : m_ownInst{ownInst},
              m_mngrId{mngrId},
              m_uvrId(uvrId)
        {
        }

        manager(const manager &) = delete;
        manager &operator=(const manager &) = delete;

        virtual ~manager()
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            m_register.clear();
        }

        void setUvrId(Uvr::UvrId_t uvrId)
        {
            m_uvrId = uvrId;
        }

        virtual id openTransaction(T &&value, errors &result, id key = undef_value)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            openTransactionInt(std::move(value), result, key);
            return key;
        }

        using GuardLock = std::lock_guard<std::recursive_mutex>;
        using TransactionGuard = std::unique_ptr<GuardLock>;

        virtual TransactionGuard
        openTransactionEx(T &&value, errors &result, id &key)
        {
            TransactionGuard guard = std::make_unique<GuardLock>(m_guard);
            openTransactionInt(std::move(value), result, key);
            return guard;
        }

        T *get_transaction_data(id key)
        {
            std::unique_lock<std::recursive_mutex> guard(m_guard);
            auto transactionIter = m_register.find(key);
            T value;
            if (transactionIter != m_register.end())
            {
                return &(transactionIter->second);
            }
            else
            {
                LOGERROR(m_ownInst, " Cannot get transaction: id=", key);
                return nullptr;
            }
        }

        virtual T closeTransaction(id key, errors &result)
        {
            std::unique_lock<std::recursive_mutex> guard(m_guard);
            auto transactionIter = m_register.find(key);
            T value;
            if (transactionIter != m_register.end())
            {
                value = transactionIter->second;
                result = errors::OK;
                m_register.erase(transactionIter);
                value.set_stop();
                LOGDEBUG(m_ownInst, " Transaction id=", key, " is closed!");
            }
            else
            {
                LOGERROR(m_ownInst, " Cannot close transaction: id=", key);
                result = errors::FAIL;
            }
            return value;
        }

    private:
        virtual void openTransactionInt(T &&value, errors &result, id &key)
        {
            result = errors::OK;
            if (key == undef_value)
            {
                key = generateTransactionId();
            }
            value.set_start();
            auto res = m_register.emplace(key, std::move(value));
            if (!res.second)
            {
                LOGERROR(m_ownInst, " The transaction already exists with id=", key);
                result = errors::FAIL;
                key = undef_value;
            }
            else
            {
                LOGDEBUG(m_ownInst, " Transaction id=", key, " is opened!");
            }
        }
    };

    class manager_novalue : public manager<base_data>
    {
    public:
        using manager::manager;

        id openTransaction(errors &result, id key = undef_value)
        {
            return manager<base_data>::openTransaction(base_data(), result, key);
        }
    };

}
