#include "UvrAPI.hpp"

// AFAP includes.
#include "config/src/config_manager.hpp"
#include "apigw/src/node.hpp"
#include "ccngw/src/node.hpp"
#include "vfgw/src/node.hpp"
#include "vfc/src/node.hpp"
#include "vfs/src/node.hpp"
#include "oam/src/node.hpp"
#include "cligw/src/Node.hpp"

// STL includes.
#include <fstream>

namespace
{
    using WorkGuard = boost::asio::executor_work_guard<boost::asio::io_context::executor_type>;

    std::unique_ptr<boost::asio::io_context> s_ioContext(nullptr);  ///< Pointer to context for asynchronous operations.
    std::unique_ptr<boost::asio::signal_set> s_signalsSet(nullptr); ///< Pointer to the signal set used to register for process termination notifications.
    std::unique_ptr<WorkGuard> s_workersGuard(nullptr);             ///< Thread guard to avoid its early termination.
    std::vector<std::shared_ptr<std::thread>> s_threadsPool;        ///< Pool of worker threads.
    std::unique_ptr<configuration::manager> s_config;               ///< Pointer to VFN configuration.

    Uvr::NodeStatus s_nodeStatus = Uvr::NodeStatus::NotInitialized;
    std::shared_mutex s_nodeStatusGuard;

    void setNodeStatus(const Uvr::NodeStatus &status)
    {
        std::unique_lock<std::shared_mutex> lock(s_nodeStatusGuard);
        LOGI("   <===>   UVR status change: ", s_nodeStatus, " -> ", status);
        s_nodeStatus = status;
    }
} // namespace

namespace Uvr
{
    namespace Troubleshoot
    {
        constexpr int columnWidth = 30;
        const std::string s_versionKvr("3.0.1");
        const std::string s_versionCsuUvr("18.17");
        const std::string s_versionUvrUvr("10.6");
    } // namespace Troubleshoot

} // namespace Uvr

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Uvr
{
    namespace API
    {
        bool init(const std::string &configPath,
                  const bool asyncMode)
        {
            s_ioContext = std::make_unique<boost::asio::io_context>();
            s_signalsSet = std::make_unique<boost::asio::signal_set>(*s_ioContext);
            // Register to handle the signals that indicate when the server should exit.
            // It is safe to register for the same signal multiple times in a program,
            // provided all registration for the specified signal is made through Asio.
            s_signalsSet->add(SIGUSR1);
            s_config.reset(new configuration::manager());
            if (s_config->loadConfigFile(configPath) == configuration::errors::FAIL)
            {
                LOGE("Configuration file is a malformed JSON one!");
                return false;
            }
            if (s_config->validateConfigParams() == configuration::errors::FAIL)
            {
                LOGE("Some mandatory parameters with configuration file are missing!");
                return false;
            }
            ccngw::api::init(*s_ioContext, *s_config);
            apigw::api::init(*s_ioContext, *s_config, asyncMode);
            apigw::api::getNode()->doRegisterSetNodeStatusCallback(
                apigw::SetNodeStatusCallback(setNodeStatus));
            vfc::api::init(*s_ioContext, *s_config);
            vfs::api::init(*s_ioContext, *s_config);
            oam::api::init(*s_ioContext, *s_config);
            vfgw::api::init(*s_ioContext, *s_config);
            CliGw::API::init(*s_ioContext, *s_config);

            setNodeStatus(NodeStatus::Initalized);
            return true;
        }

        void start(int threadPoolSize)
        {
            bool isStarted = true;
            setNodeStatus(NodeStatus::Starting);

            if (isStarted)
            {
                isStarted &= CliGw::API::start();
            }
            if (isStarted)
            {
                isStarted &= oam::api::start();
            }
            if (isStarted)
            {
                isStarted &= vfgw::api::start();
            }
            if (isStarted)
            {
                isStarted &= vfc::api::start();
            }
            if (isStarted)
            {
                isStarted &= vfs::api::start();
            }
            if (isStarted)
            {
                isStarted &= ccngw::api::start();
            }
            if (isStarted)
            {
                isStarted &= apigw::api::start();
            }

            if (!isStarted)
            {
                setNodeStatus(NodeStatus::Stopped);
                exit(1);
            }

            s_signalsSet->async_wait(std::bind(&Uvr::API::stop));
            s_workersGuard = std::make_unique<WorkGuard>(boost::asio::make_work_guard(*s_ioContext));

            auto auxThreadMain = []()
            {
                LOGDEBUG("VFN: Aux thread started.");
                s_ioContext->run();
                LOGDEBUG("VFN: Aux thread stopped.");
            };

            // Create a pool of threads to run all of the io_contexts.
            for (int i = 0; i < threadPoolSize; ++i)
            {
                auto t = std::make_shared<std::thread>(auxThreadMain);
                s_threadsPool.push_back(t);
            }
        }

        Uvr::NodeStatus getNodeStatus()
        {
            std::shared_lock<std::shared_mutex> lock(s_nodeStatusGuard);
            return s_nodeStatus;
        }

        void stop()
        {
            setNodeStatus(NodeStatus::Stopping);

            ccngw::api::stop();
            apigw::api::stop();
            vfc::api::stop();
            vfs::api::stop();
            oam::api::stop();
            vfgw::api::stop();
            CliGw::API::stop();

            if (s_workersGuard)
            {
                s_workersGuard->reset();
            }
            if (s_signalsSet)
            {
                s_signalsSet->cancel();
            }
        }

        void waitStop()
        {
            // Wait for all threads in the pool to exit.
            for (std::size_t i = 0; i < s_threadsPool.size(); ++i)
            {
                if (s_threadsPool[i]->joinable())
                {
                    s_threadsPool[i]->join();
                }
            }
            // Cleanup
            s_threadsPool.clear();
            ccngw::api::reset();
            apigw::api::reset();
            vfc::api::reset();
            vfs::api::reset();
            oam::api::reset();
            vfgw::api::reset();
            CliGw::API::reset();
            s_config.reset();
            s_signalsSet.reset();
            s_ioContext.reset();

            setNodeStatus(NodeStatus::Stopped);
        }

        SessionId_t verifyCall(UserId_t userId,
                               OperatorId_t idSrc,
                               const OctetStringPtr &callingNumber,
                               const OctetStringPtr &calledNumber,
                               const OctetStringPtr &originalCalledNumber,
                               const OctetStringPtr &genericNumber,
                               const CallId_t &callId,
                               bool isTrusted)
        {
            return apigw::api::getNode()->doVerifyCall(userId,
                                                       idSrc,
                                                       callingNumber,
                                                       calledNumber,
                                                       originalCalledNumber,
                                                       genericNumber,
                                                       callId,
                                                       isTrusted);
        }

        VerificationResult
        verifyCallSynchronously(const OperatorId_t idSrc,
                                const OctetStringPtr &callingNumber,
                                const OctetStringPtr &calledNumber,
                                const OctetStringPtr &originalCalledNumber,
                                const OctetStringPtr &genericNumber,
                                const CallId_t &callId,
                                const bool isTrusted)
        {
            return apigw::api::getNode()->doVerifyCallSynchronously(idSrc,
                                                                    callingNumber,
                                                                    calledNumber,
                                                                    originalCalledNumber,
                                                                    genericNumber,
                                                                    callId,
                                                                    isTrusted);
        }

        void notifyAboutOutgoingCall(const OctetStringPtr &callingNumber,
                                     const OctetStringPtr &calledNumber,
                                     const Uvr::CallId_t &callId,
                                     const Uvr::OptionalOperatorId &idDst)
        {
            apigw::api::getNode()->doNotifyAboutOutgoingCall(callingNumber, calledNumber, callId, idDst);
        }

        UserId_t registerVerifyCallVerdictCallback(VerifyCallVerdictCallback callback,
                                                   UserId_t userId)
        {
            return apigw::api::getNode()->doRegister(apigw::verifyCallVerdictCallbackId, callback, userId);
        }

        UserId_t registerCallInformationRequestCallback(CallInformationRequestCallback callback,
                                                        UserId_t userId)
        {
            return apigw::api::getNode()->doRegister(apigw::callInformationRequestCallbackId, callback, userId);
        }

        UserId_t registerStoreRecordCallback(const StoreRecordCallback callback,
                                             const UserId_t userId)
        {
            return apigw::api::getNode()->doRegister(apigw::storeRecordCallbackId, callback, userId);
        }

        UserId_t registerUpdateRecordsCallback(const UpdateRecordsCallback callback,
                                               const UserId_t userId)
        {
            return apigw::api::getNode()->doRegister(apigw::updateRecordsCallbackId, callback, userId);
        }

        void reportCallInformation(const CallInformationResponses &responses)
        {
            if (!responses.empty())
            {
                apigw::api::getNode()->processCallInformation(responses);
            }
        }

        namespace Digest
        {
            namespace Numbers
            {
                UvrIds getUvrIds(const OctetStringPtr &number)
                {
                    return ::apigw::api::getNode()->getUvrIds(number);
                }
            } // namespace Numbers

            namespace Uvrs
            {
                UvrInfo getUvrInfo(const UvrId_t &uvrId)
                {
                    return ::apigw::api::getNode()->getUvrInfo(uvrId);
                }

                UvrId_t getOwnUvrId()
                {
                    return ::apigw::api::getNode()->getOwnUvrId();
                }
            } // namespace Uvrs

            namespace Hubs
            {
                HubInfo getHubInfo(const Gt_t &hubId)
                {
                    return ::apigw::api::getNode()->getHubInfo(hubId);
                }
            } // namespace Hubs

            namespace Operators
            {
                OperatorInfo getOperatorInfo(const OperatorId_t &operatorId)
                {
                    return ::apigw::api::getNode()->getOperatorInfo(operatorId);
                }
            } // namespace Operators

        } // namespace Digest

    } // namespace API

    namespace Troubleshoot
    {
        std::string doVersions()
        {
            std::stringbuf buf;
            std::ostream out(&buf);

            out << "=================================================================" << std::endl;
            out << std::setw(columnWidth) << "Versions"
                << " "
                << "information"
                << "\n"
                << "-----------------------------------------------------------------" << std::endl;

            out << std::setw(columnWidth) << "KVR"
                << ": "
                << s_versionKvr
                << "\n";
            out << std::setw(columnWidth) << "CSU-UVR"
                << ": "
                << s_versionCsuUvr
                << "\n";
            out << std::setw(columnWidth) << "UVR-UVR"
                << ": "
                << s_versionUvrUvr
                << "\n";
            out << std::setw(columnWidth) << "Configuration file"
                << ": "
                << configuration::getConfigFileVersion() << std::endl;

            out << "-----------------------------------------------------------------" << std::endl;
            out << std::setw(columnWidth) << "Make options"
                << " "
                << "information"
                << "\n"
                << "-----------------------------------------------------------------" << std::endl;

            out << std::setw(columnWidth) << "optim"
                << ": "
                << MAKE_OPTIM << std::endl;
            out << std::setw(columnWidth) << "log_level"
                << ": "
                << MAKE_LOG_LEVEL << std::endl;
            out << std::setw(columnWidth) << "tool"
                << ": "
                << MAKE_TOOL << std::endl;
            if (!std::string(GIT_COMMIT).empty())
            {
                out << std::setw(columnWidth) << "GIT_COMMIT"
                    << ": "
                    << GIT_COMMIT << std::endl;
            }

            out << "=================================================================" << std::endl;

            return buf.str();
        }

    } // namespace Troubleshoot

} // namespace Uvr
