// Own include.
#include "Cli.hpp"

// Cli includes.
#include "Common.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "components/tools/src/Tools.hpp"
#include "components/worker_node/src/Codec.hpp"
#include "components/worker_node/src/primitive.hpp"

// Boost includes.
#include <boost/asio.hpp>
#include <boost/bind.hpp>

#define START_ASYNC_TIMED_CONNECT(S, T, CR, TR)                                                         \
    {                                                                                                   \
        TR = std::make_shared<OptionalErrorCode>();                                                     \
        CR = std::make_shared<OptionalErrorCode>();                                                     \
                                                                                                        \
        T.expires_from_now(boost::posix_time::milliseconds(10));                                        \
        T.async_wait(boost::bind(setResultShared, TR, boost::asio::placeholders::error));               \
        boost::asio::async_connect(S, endpointIterator,                                                 \
                                   boost::bind(setResultShared, CR, boost::asio::placeholders::error)); \
    }

#define GET_IO_SERVICE(s) ((boost::asio::io_context &)(s).get_executor().context())

#ifndef MAX_CONN_BUFFER_SIZE
#define MAX_CONN_BUFFER_SIZE 2048
#endif

using boost::asio::ip::tcp;

namespace
{
    using OptionalErrorCode = std::optional<boost::system::error_code>;
    using SharedOptionalErrorCode = std::shared_ptr<OptionalErrorCode>;

    boost::asio::io_context s_ioContext;
    tcp::socket s_socket{s_ioContext};

    timers::time_point s_start;
    timers::time_point s_stop;

    bool connectWithTimeout(tcp::socket &sock,
                            const tcp::resolver::iterator &endpointIterator);

    template <typename MutableBufferSequence>
    void readWithTimeout(tcp::socket &sock,
                         const MutableBufferSequence &buffers,
                         const SharedOptionalErrorCode &read_result);

    void setResultShared(SharedOptionalErrorCode a, boost::system::error_code b);

    int sendPrimitive(tcp::socket &socket, primitive::generic *prim_p);
    primitive::GenericPtr readPrimitive(tcp::socket &socket);
}

namespace Cli
{
    void printShortHelp()
    {
        // TODO[HUB-350]
    }

    void printLongHelp()
    {
        // TODO[HUB-350]
    }

    bool connectToUvr(const std::string &uvrAddress, const std::string &uvrPort)
    {
        boost::system::error_code errorCode;
        tcp::resolver resolver(s_ioContext);
        tcp::resolver::query query(uvrAddress, uvrPort);
        tcp::resolver::iterator endpointIterator = resolver.resolve(query, errorCode);

        if (errorCode)
        {
            LOGERROR(s_logTag, "Failed to resolve ", uvrAddress, ":", uvrPort);
            return false;
        }

        if (!connectWithTimeout(s_socket, endpointIterator))
        {
            LOGERROR(s_logTag, "Failed to connect with ", uvrAddress, ":", uvrPort);
            return false;
        }

        LOGINFO(s_logTag, "Connected to ", uvrAddress, ":", uvrPort, " on socket ", s_socket.native_handle());

        primitive::Whoami prim;
        sendPrimitive(s_socket, &prim);
        return true;
    }

    bool sendExecCmdReq(const std::string &cmd,
                        const Strings &cmdArgs)
    {
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
        LOGDEBUG(Cli::s_logTag, " Command: ", cmd);
        std::string args;
        for (size_t i = 0; i < cmdArgs.size(); ++i)
        {
            args += cmdArgs.at(0);
            args += " ";
        }
        LOGDEBUG(Cli::s_logTag, " Arguments (", cmdArgs.size(), "): ", args);
#endif

        auto prim = std::make_shared<primitive::ExecCmdReq>(cmd, cmdArgs);

        sendPrimitive(s_socket, prim.get());
        return true;
    }

    bool handleExecCmdRsp()
    {
        auto prim = readPrimitive(s_socket);

        auto rsp =
            std::dynamic_pointer_cast<::primitive::ExecCmdRsp>(prim);

        if (rsp == nullptr)
        {
            return false;
        }

        std::stringbuf trace;
        std::ostream logger(&trace);

        for (size_t i = 0; i < rsp->responses.size(); ++i)
        {
            logger << rsp->responses.at(i) << "\n";
        }

        std::cout << trace.str();
        return true;
    }
}

namespace
{
    bool connectWithTimeout(tcp::socket &sock,
                            const tcp::resolver::iterator &endpointIterator)
    {
        boost::asio::deadline_timer timer(sock.get_executor());

        SharedOptionalErrorCode timerResult;
        SharedOptionalErrorCode connectResult;
        int attLeft = 10;

        START_ASYNC_TIMED_CONNECT(sock, timer, connectResult, timerResult);

        GET_IO_SERVICE(sock).restart();
        while (GET_IO_SERVICE(sock).run_one())
        {
            if (!attLeft)
            {
                LOGINFO(Cli::s_logTag, " No more attempts to connect left");
                return false;
            }
            if (*connectResult)
            {
                timer.cancel();
                --attLeft;

                if (connectResult->value())
                {
                    LOGINFO(Cli::s_logTag, " Do async connect error=", connectResult->value(), " attLeft=", attLeft);
                    START_ASYNC_TIMED_CONNECT(sock, timer, connectResult, timerResult);
                }
                else
                {
                    return true;
                }
            }
            else if (*timerResult)
            {
                sock.cancel();
                --attLeft;

                LOGINFO(Cli::s_logTag, " connect timeout attLeft=", attLeft);
                START_ASYNC_TIMED_CONNECT(sock, timer, connectResult, timerResult);
            }

            GET_IO_SERVICE(sock).restart();
        }

        return false;
    }

    template <typename MutableBufferSequence>
    void readWithTimeout(tcp::socket &sock,
                         const MutableBufferSequence &buffers,
                         const SharedOptionalErrorCode &read_result)
    {
        auto timer_result = std::make_shared<OptionalErrorCode>();
        boost::asio::deadline_timer timer(sock.get_executor());
        timer.expires_from_now(boost::posix_time::milliseconds(1100));
        timer.async_wait(boost::bind(setResultShared, timer_result, boost::asio::placeholders::error));

        boost::asio::async_read(sock, buffers,
                                boost::bind(setResultShared, read_result, boost::asio::placeholders::error));

        GET_IO_SERVICE(sock).restart();
        while (GET_IO_SERVICE(sock).run_one())
        {
            if (*read_result)
            {
                timer.cancel();
            }
            else if (*timer_result)
            {
                sock.cancel();
            }
        }
    }

    void setResultShared(SharedOptionalErrorCode a, boost::system::error_code b)
    {
        a->emplace(b);
    }

    int sendPrimitive(tcp::socket &socket, primitive::generic *prim_p)
    {
        prim_p->m_header->from = Cli::s_cliInst;
        prim_p->m_header->to = Cli::s_cliGwInst;
        primitive::raw_data_t msg = primitive::Codec::encode(prim_p, primitive::Codec::Type::Json, true);
        LOGDEBUG(Cli::s_logTag, " Sending ", doText(msg));
        return (int)boost::asio::write(socket, boost::asio::buffer(msg));
    }

    primitive::GenericPtr readPrimitive(tcp::socket &socket)
    {
        std::array<char, MAX_CONN_BUFFER_SIZE> buf;
        auto errorCode = std::make_shared<OptionalErrorCode>();

        readWithTimeout(socket, boost::asio::buffer(buf.data(), sizeof(uint16_t)), errorCode);
        if (errorCode->value())
        {
            LOGERROR(Cli::s_logTag, " Message size read_with_timeout failed.");
            return nullptr;
        }

        auto msgSize = *reinterpret_cast<const uint16_t *>(buf.data());

        readWithTimeout(socket, boost::asio::buffer(buf.data(), msgSize), errorCode);
        if (errorCode->value())
        {
            LOGERROR(Cli::s_logTag, " Message read_with_timeout failed.");
            return nullptr;
        }

        primitive::raw_data_t rawData(buf.data(), msgSize);
        auto prim = primitive::Codec::decode(rawData);
        if (prim == nullptr)
        {
            LOGERROR(Cli::s_logTag, " Primitive decode failed.");
            return nullptr;
        }

        LOGDEBUG(Cli::s_logTag,
                 " Receiving ", prim->m_header->id,
                 " From: ", prim->m_header->from,
                 " To: ", prim->m_header->to);

        return prim;
    }
}
