#include "vfc/src/matcher.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/log/src/Log.hpp"
#include "TRecQ763Number.hpp"

namespace vfc
{

    uint64_t matcher::reverseNumberBytes(uint64_t &number)
    {
        uint64_t ret = (number & LEVEL_MASK);
        number >>= LEVEL_BITS;
        for (int i = 0; i < (MAX_LEVELS - 2); i++)
        {
            ret <<= LEVEL_BITS;
            ret |= (number & LEVEL_MASK);
            number >>= LEVEL_BITS;
        }
        ret <<= LEVEL_BITS;
        ret |= (number & LEVEL_MASK);

        return ret;
    }

    uint64_t matcher::convertToNumber(const Uvr::Number &callingPartyNumber)
    {
        /* Convert number as a string to uint64_t
           for example "79875528794" -> 0x1298F3D8A5.
           The thing is that the order of the bytes
           with the return value should be reversed like
           0x1298F3D8A5 -> 0x5AD8F39812
           to match with matcher::lvls */
        uint64_t number = std::atol(callingPartyNumber.c_str());
        return reverseNumberBytes(number);
    }

    uint64_t matcher::convertToNumber(const Uvr::OctetStringPtr &callingPartyNumber)
    {
        uint64_t number = TRecQ763Number::toUInt64(callingPartyNumber);
        return reverseNumberBytes(number);
    };

    matcher::matcher(const TypeInst &owner,
                     boost::asio::io_context &ioc)
        : m_owner(owner),
          m_fs_local(owner),
          m_ioc(ioc)
    {
        reset();
    }

    void matcher::reset()
    {
        std::unique_lock<std::shared_mutex> guard(m_guard);
        lvls_p.reset();
        lvls_p = std::make_shared<MAX_LEVEL_TYPE>();
        m_number_of_subs = 0;
    }

    matcher::result_t matcher::match(const Uvr::OctetStringPtr &CallingPN,
                                     subscriber::info_t &rec)
    {
        std::shared_lock<std::shared_mutex> guard(m_guard);
        subscriber::info_t *subs_p = nullptr;
        lvls_p->match(convertToNumber(CallingPN), &subs_p);
        if (subs_p)
        {
            rec = *subs_p;
            return result_t::do_match;
        }
        else
        {
            return result_t::no_match;
        }
    }

    void matcher::sub_insert(const db::subscriber::record &rec)
    {
        subscriber::info_t subs_info;
        subs_info.m_idUvrP = rec.m_idUvrP;
        subs_info.m_idUvrS = rec.m_idUvrS;
        // Meta Information for numbers is not used yet.
        // subs_info.m_metaInfoFlags = Uvr::MetaInfo::toBitFlags(rec.m_metaInfo);

        if (lvls_p->push(convertToNumber(rec.m_number), subs_info) == 0)
        {
            m_number_of_subs++;
        }
        else
        {
            LOGWARN(m_owner,
                    " Information about subscriber [number=",
                    rec.m_number, "]: existed before insert procedure");
        }

        LOGDEBUG(m_owner,
                 " Added subscriber[number=", rec.m_number,
                 "]: primary uvrId=", subs_info.m_idUvrP,
                 ", secondary uvrId=", subs_info.m_idUvrS,
                 ", meta=", subs_info.m_metaInfoFlags,
                 "(", rec.m_metaInfo,
                 "), number of subscribers=", m_number_of_subs);
    }

    void matcher::subscribersParse(const StringsBuff &buff,
                                   std::shared_ptr<PopulateGuard> guard,
                                   StrandType strand)
    {
        std::shared_ptr<NumberInfoBuff> pp = std::make_shared<NumberInfoBuff>();
        pp->reserve(buff.size());

        for (const auto &line : buff)
        {
            db::subscriber::record rec(line.c_str());
            pp->emplace_back(NumberInfo{rec.m_idUvrP,
                                        rec.m_idUvrS,
                                        // Meta Information for numbers is not used yet.
                                        // Uvr::MetaInfo::toBitFlags(rec.m_metaInfo);
                                        Uvr::MetaInfo::emptyBitFlags,
                                        convertToNumber(rec.m_number)});
        }
        boost::asio::post(strand,
                          std::bind(&matcher::subscribersInsert,
                                    this,
                                    pp,
                                    guard));
    }

    void matcher::subscribersInsert(const std::shared_ptr<NumberInfoBuff> buff,
                                    std::shared_ptr<PopulateGuard> guard)
    {
        for (auto &info : *buff)
        {
            if (lvls_p->push(info.number, info.subInfo) == 0)
            {
                m_number_of_subs++;
            }
            else
            {
                LOGWARN(m_owner,
                        " Information about subscriber [number=",
                        info.number, "]: existed before insert procedure");
            }

            LOGDEBUG(m_owner,
                     " Added subscriber[number=", info.number,
                     "]: primary uvrId=", info.subInfo.m_idUvrP,
                     ", secondary uvrId=", info.subInfo.m_idUvrS,
                     ", meta=", info.subInfo.m_metaInfoFlags,
                     ", number of subscribers=", m_number_of_subs);
        }
    }

    void matcher::sub_update(const db::subscriber::record &rec)
    {
        subscriber::info_t *sub_p = nullptr;
        lvls_p->match(convertToNumber(rec.m_number), &sub_p);
        if (sub_p == nullptr)
        {
            LOGERROR(m_owner,
                     " Cannot update information about subscriber[number=",
                     rec.m_number, "]: record is not found");
            return;
        }

        sub_p->m_idUvrP = rec.m_idUvrP;
        sub_p->m_idUvrS = rec.m_idUvrS;
        // Meta Information for numbers is not used yet.
        // sub_p->m_metaInfoFlags = Uvr::MetaInfo::toBitFlags(rec.m_metaInfo);

        LOGDEBUG(m_owner,
                 " Updated subscriber[number=", rec.m_number,
                 "]: primary uvrId=", sub_p->m_idUvrP,
                 ", secondary uvrId=", sub_p->m_idUvrS,
                 ", meta=", sub_p->m_metaInfoFlags,
                 "(", rec.m_metaInfo, ")");
    }

    void matcher::sub_remove(const db::subscriber::record &rec)
    {
        subscriber::info_t subs_info;
        lvls_p->pop(convertToNumber(rec.m_number), subs_info);
        if (subs_info.m_idUvrP == 0)
        {
            LOGWARN(m_owner,
                    " Cannot delete information about subscriber[number=",
                    rec.m_number, "]: record is not found");
            return;
        }
        m_number_of_subs--;

        LOGDEBUG(m_owner, " Deleted subscriber[number=", rec.m_number,
                 "]: number of subscribers=", m_number_of_subs);
    }

#define BSIZE 100000

    void matcher::populate(const std::string &zipFilename)
    {
        reset();
        std::unique_lock<std::shared_mutex> guard(m_guard);

        LOGINFO(m_owner, " START populate from ", zipFilename);
        StringsBuff buff;
        buff.reserve(BSIZE);

        std::promise<bool> promiseObj;
        std::future<bool> futureObj = promiseObj.get_future();

        std::shared_ptr<PopulateGuard> guardP = std::make_shared<PopulateGuard>(promiseObj);

        StrandType strand = boost::asio::make_strand(m_ioc);

        Csv::File::parseZip(zipFilename,
                            [this, &buff, &guardP, &strand](std::string &&line)
                            {
                                buff.emplace_back(std::move(line));
                                if (buff.size() == BSIZE)
                                {
                                    boost::asio::post(m_ioc,
                                                      std::bind(&matcher::subscribersParse,
                                                                this,
                                                                std::move(buff),
                                                                guardP,
                                                                strand));
                                    buff.reserve(BSIZE);
                                }
                                return false;
                            });
        if (buff.size())
        {
            boost::asio::post(m_ioc,
                              std::bind(&matcher::subscribersParse,
                                        this,
                                        std::move(buff),
                                        guardP,
                                        strand));
        }

        guardP.reset();
        futureObj.get();

        LOGI(m_owner, " DONE populate from ", zipFilename,
             " subscribers=", number_of_subs());
    }

    void matcher::update(const std::string &filename)
    {
        std::unique_lock<std::shared_mutex> guard(m_guard);
        LOGINFO(m_owner, " START update from ", filename);
        std::string unzipFilename;
        m_fs_local.unzipFile(filename, unzipFilename);
        int cntr = 0;
        Csv::File::parse(unzipFilename,
                         [this, &cntr](const std::string &line)
                         {
                             cntr++;
                             std::size_t pos = line.find_first_of(';');
                             std::string operation = line.substr(0, pos);
                             db::subscriber::record rec(line.c_str() + pos + 1);
                             if (operation == "ADD")
                             {
                                 this->sub_insert(rec);
                             }
                             else if (operation == "MOD")
                             {
                                 this->sub_update(rec);
                             }
                             else if (operation == "DEL")
                             {
                                 this->sub_remove(rec);
                             }
                             else
                             {
                                 LOGERROR(m_owner,
                                          " Failed to update subscribers matcher: unknown operation=\"", operation,
                                          "\", line=\"", line, "\"");
                             }
                             return false;
                         });
        m_fs_local.deleteFile(unzipFilename);
        LOGINFO(m_owner, " DONE update ", cntr, " lines from ", filename,
                " subscribers=", m_number_of_subs);
    }

    ::db::limit_t matcher::number_of_subs() const
    {
        return m_number_of_subs;
    }

    std::ostream &operator<<(std::ostream &os, const matcher::result_t &obj)
    {
        switch (obj)
        {
            case matcher::result_t::do_match:
                os << "do_match";
                break;
            case matcher::result_t::no_match:
                os << "no_match";
                break;
        }
        return os;
    }

    matcher::PopulateGuard::PopulateGuard(std::promise<bool> &promiseObj)
        : m_promise(promiseObj)
    {
    }

    matcher::PopulateGuard::~PopulateGuard()
    {
        m_promise.set_value(true);
    }

} // namespace vfc
