#include "vfc/src/node.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/db/src/mimic.hpp"
#include "components/numbers/src/Validator.hpp"
#include "components/timers/src/timer_periodic.hpp"
#include "components/tools/src/Tools.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "TRecQ763Number.hpp"

namespace vfc
{
    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config)
        : pm::node(ownInst, ioc, config),
          m_matcher(ownInst, ioc),
          m_vf_req_mngr(ownInst, VERIFY_REQ_MNGR_ID, m_uvrId)
    {

        m_db_client_p = std::make_shared<::db::stub::mimic>(m_uvrId);
        db::interfaces::tables::client::bind(m_db_client_p.get());
        db::interfaces::tables::incidents::client::bind(m_db_client_p.get());
        m_vf_req_mngr.interfaces::verifier::client::bind(this);
    }

    node::~node()
    {
        m_db_client_p->disconnect();
    }

    void node::start(void)
    {
        start_node_timers();
        pm::node::start();
        // Create tables
        m_db_client_p->connect();
        db::interfaces::tables::client::create(m_db_client_p->db::interfaces::tables::incidents::server::tableName,
                                               m_db_client_p->db::interfaces::tables::incidents::server::tableFields,
                                               m_db_client_p->db::interfaces::tables::incidents::server::tableSettings);
    }

    void node::start_node_timers()
    {
        // TRANSACTIONS_WATCHDOG
        m_transactions_watchdog_timerId =
            m_timersManager->makeTimerAndStart<timers::periodic>("TRANSACTIONS_WATCHDOG",
                                                                 std::chrono::milliseconds(10));
    }

    void node::prepareCounters()
    {
        addCounter(pm::counter::calls::incoming, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::unverified, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::ATTMS, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::TBVRF, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::RJCTS, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::ERR1, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::ERR2, pm::counter::indexing::yes);
        addCounter(pm::counter::vfc::subscribers, pm::counter::indexing::no);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case ::primitive::VERIFY_CALL_QUERY:
            {
                handleVerifyCallQuery(prim);
                break;
            }
            case ::primitive::VERIFY_CALL_RSP:
            {
                auto rsp = std::dynamic_pointer_cast<::primitive::VerifyCallResponse>(prim);
                if (!rsp)
                {
                    break;
                }

                ::transaction::errors res;
                transaction::vf_req::data transaction_data(m_vf_req_mngr.closeTransaction(prim->m_header->ulpKey, res));
                if (::transaction::errors::OK == res)
                {
                    doVerdict(transaction_data,
                              rsp->m_header->ulpKey,
                              rsp->verificationVerdict,
                              rsp->incidentVerdict,
                              rsp->uvrT);
                }
                break;
            }
            case ::primitive::GET_INCIDENTS_REQ:
            {
                db::incident::rows incidents;
                db::interfaces::tables::incidents::client::get(incidents,
                                                               db::last_get,
                                                               db::latest);
                std::string report_name;
                std::string reportDirPath = m_config.getValue<std::string>(CCNGW_REPORT_PATH + INCIDENTS_REPORT,
                                                                           INCIDENTS_REPORT_DEFAULT_DIR)
                                            + "/";

                fs::local localFs(m_ownTypeInst);
                localFs.createDir(reportDirPath);

                if (Csv::File::generate(reportDirPath + Csv::Incident::prefix(m_uvrId),
                                        Csv::Incident::header,
                                        incidents,
                                        report_name))
                {
                    auto rsp = std::make_shared<primitive::GetIncidentsRsp>(report_name);
                    rsp->m_header->to = prim->m_header->from;
                    rsp->m_header->ulpKey = prim->m_header->ulpKey;
                    workers::node::send(rsp);
                }
                break;
            }
            case ::primitive::UPLOAD_NUMBERS_IND:
            {
                auto ind =
                    std::dynamic_pointer_cast<::primitive::UploadNumbersInd>(prim);
                if (ind)
                {
                    if (!ind->numFile.empty())
                    {
                        LOGINFO(m_ownTypeInst,
                                " Received indication about new numeration file,"
                                " recreate subscribers: file=\"",
                                ind->numFile, "\"");
                        m_matcher.populate(ind->numFile);
                    }
                    for (auto const &deltaFile : ind->deltas)
                    {
                        m_matcher.update(deltaFile);
                    }
                    setCounter(pm::counter::vfc::subscribers,
                               ::pm::counter::value(static_cast<::pm::counter::value_t>(m_matcher.number_of_subs())));

                    auto status = std::make_shared<primitive::StatusChangeInd>(Uvr::NodeStatus::Started);
                    status->m_header->to = TypeInst(workers::type::APIGW, 0);
                    workers::node::send(status);
                }
                break;
            }
            case ::primitive::QUERY_UVR_IDS_REQ:
            {
                auto req = std::dynamic_pointer_cast<::primitive::QueryUvrIdsReq>(prim);
                if (!req)
                {
                    break;
                }
                subscriber::info_t subInfo;
                m_matcher.match(req->number, subInfo);

                auto rsp = std::make_shared<primitive::QueryUvrIdsRsp>(subInfo.m_idUvrP, subInfo.m_idUvrS);
                rsp->m_header->to = prim->m_header->from;
                rsp->m_header->ulpKey = prim->m_header->ulpKey;
                workers::node::send(rsp);
                break;
            }
            default:
                pm::node::handleSpecificPrimitive(prim);
        }
    }

    void node::handleVerifyCallQuery(const primitive::GenericPtr &prim)
    {
        auto query = std::dynamic_pointer_cast<::primitive::VerifyCallQuery>(prim);

        increaseCounter(pm::counter::calls::incoming, std::to_string(query->idSrc));

        const auto sessionId = query->m_header->ulpKey;
        transaction::vf_req::data data(0, Uvr::UvrId::undefined, query);
        data.set_start();

        // First validate the calling number.
        if (!TRecQ763Number::Validator::verify(query->callingNumber))
        {
            increaseCounter(pm::counter::calls::ATTMS, std::to_string(query->idSrc));
            LOGI(m_logTag, "Calling number is not Russian");
            doVerdict(data, sessionId, Uvr::VerificationVerdict::callingNumberNotRussian);
            return;
        }

        // Then try to match the calling number.
        subscriber::info_t subInfo;
        const auto match = m_matcher.match(query->callingNumber, subInfo);
        if (match == matcher::result_t::no_match)
        {
            increaseCounter(pm::counter::calls::ATTMS, std::to_string(query->idSrc));
            LOGI(m_logTag, "Calling number is not not found");
            doVerdict(data, sessionId, Uvr::VerificationVerdict::callingNumberNotFound);
            return;
        }

        if (Digest::shouldIncreaseCounters(subInfo.m_idUvrP))
        {
            increaseCounter(pm::counter::calls::ATTMS, std::to_string(query->idSrc));
        }

        // Check if the number has a service UVR ID.
        if (!Digest::shouldSendVerifyRequest(subInfo.m_idUvrP))
        {
            switch (subInfo.m_idUvrP)
            {
                case Uvr::UvrId::numbersNotServiced:
                {
                    doVerdict(data, sessionId, Uvr::VerificationVerdict::failure, Uvr::VerificationVerdict::undefined, subInfo.m_idUvrP);
                    break;
                }
                default:
                {
                    doVerdict(data, sessionId, Uvr::VerificationVerdict::callingNumberNotServiced);
                }
            }
            return;
        }

        data.m_metaInfoFlags = subInfo.m_metaInfoFlags;
        data.m_idUvrT = subInfo.m_idUvrP;

        // Open transaction for verification request.
        ::transaction::errors result;
        m_vf_req_mngr.openTransaction(std::move(data), result, sessionId);
        if ((result != ::transaction::errors::OK) && Digest::shouldIncreaseCounters(subInfo.m_idUvrP))
        {
            // As no verification request is sent due to
            // transaction opening failure
            increaseCounter(pm::counter::calls::ERR2, std::to_string(query->idSrc));
            return;
        }

        // Finally, when all preconditions have been checked, a verification
        // request can be sent.
        const auto req = std::make_shared<primitive::VerifyCallRequest>(query,
                                                                        subInfo.m_idUvrP,
                                                                        subInfo.m_idUvrS);

        if (subInfo.m_idUvrP == m_uvrId || subInfo.m_idUvrS == m_uvrId)
        {
            req->m_header->to = TypeInst(workers::type::VFS, 0);
        }
        else
        {
            req->m_header->to = TypeInst(workers::type::VFGW, 0);
        }
        req->m_header->ulpKey = sessionId;
        workers::node::send(req);

        if (Digest::shouldIncreaseCounters(subInfo.m_idUvrP))
        {
            increaseCounter(pm::counter::calls::TBVRF, std::to_string(query->idSrc));
        }
    }

    void node::doTimeoutVerdict(transaction::vf_req::data &data,
                                ::transaction::id transaction_id)
    {
        doVerdict(data, transaction_id, Uvr::VerificationVerdict::timeout, Uvr::VerificationVerdict::undefined, data.m_idUvrT);

        auto abort = std::make_shared<primitive::VerifyRequestAbort>();
        abort->m_header->to = TypeInst(workers::type::VFGW, 0);
        abort->m_header->ulpKey = transaction_id;
        workers::node::send(abort);
    }

    void node::doVerdict(const transaction::vf_req::data &data,
                         const ::transaction::id transactionId,
                         const Uvr::VerificationVerdict_t verificationVerdict,
                         const Uvr::VerificationVerdict_t incidentVerdict,
                         const Uvr::UvrId_t idUvrT)
    {
        // Send reply.
        auto vrd = std::make_shared<primitive::VerifyCallVerdict>(verificationVerdict, idUvrT);
        vrd->m_header->to = TypeInst(workers::type::APIGW, 0);
        vrd->m_header->ulpKey = transactionId;
        workers::node::send(vrd);

        // Increase counters on real verdict.
        if (Digest::shouldIncreaseCounters(idUvrT))
        {
            switch (verificationVerdict)
            {
                case Uvr::VerificationVerdict::failure:
                {
                    increaseCounter(pm::counter::calls::RJCTS,
                                    std::to_string(data.m_query->idSrc));
                    break;
                }
                case Uvr::VerificationVerdict::timeout:
                {
                    increaseCounter(pm::counter::calls::ERR1,
                                    std::to_string(data.m_query->idSrc));
                    break;
                }
                case Uvr::VerificationVerdict::callingNumberNotRussian:
                case Uvr::VerificationVerdict::callingNumberNotFound:
                case Uvr::VerificationVerdict::callingNumberNotServiced:
                {
                    increaseCounter(pm::counter::calls::ERR2,
                                    std::to_string(data.m_query->idSrc));
                    break;
                }
                default:
                    break;
            }
        }

        // Generate incident.
        const bool makeIncident = (verificationVerdict != Uvr::VerificationVerdict::success
                                   || incidentVerdict != Uvr::VerificationVerdict::undefined)
                                  && Digest::shouldMakeIncident(idUvrT);
        if (makeIncident)
        {
            const auto verdict =
                incidentVerdict == Uvr::VerificationVerdict::undefined ? verificationVerdict
                                                                       : incidentVerdict;

            db::incident::record recIncident = {
                data.m_query->callingNumber,
                data.m_query->calledNumber,
                data.m_query->originalCalledNumber,
                data.m_query->genericNumber,
                data.get_start(),
                Uvr::VerdictTrigger::antifraud,
                verdict,
                data.m_query->idSrc,
                idUvrT,
                data.m_query->callId};

            db::interfaces::tables::incidents::client::put(recIncident);
        }
    }

    void node::handle_timeout(const ::timers::id &id,
                              const ::timers::time_point &start_time)
    {

        if (id == m_transactions_watchdog_timerId)
        {
            m_vf_req_mngr.check_watchdogs();
        }
    }

} // namespace vfc
