#include "vfgw/src/CryptoRsa.hpp"

// AFAP includes.
#include "components/tools/src/Tools.hpp"
#include "components/worker_node/src/config_parameters.hpp"

// OpenSSL includes.
#include <openssl/err.h>
#include <openssl/pem.h>

namespace Crypto
{
    Rsa::Rsa(const configuration::manager &config)
    {
        std::string privateKeyPath = config.getValue<std::string>(VFGW_PRIVATE_KEY_PATH);
        if (privateKeyPath.empty())
        {
            LOGINFO("Path to VFN private key is not configured!");
            return;
        }
        m_ownPrivateKey = readPrivateKey(privateKeyPath);
    }

    Rsa::~Rsa()
    {
        if (m_ownPrivateKey != nullptr)
        {
            RSA_free(m_ownPrivateKey);
        }
    }

    bool Rsa::encrypt(RSA *key,
                      const ::primitive::raw_data_t &inData,
                      ::primitive::raw_data_t &outData)
    {
        int blockSize = RSA_size(key);
        unsigned char blockBuf[blockSize] = {0};
        int blockLen = blockSize - 11; // limit for RSA_PKCS1_PADDING
        int dataLen = inData.size();
        const unsigned char *inBuf = reinterpret_cast<const unsigned char *>(inData.c_str());
        outData.clear();
        while (dataLen > 0)
        {
            int inLen = dataLen > blockLen ? blockLen : dataLen;
            int outLen = RSA_public_encrypt(inLen, inBuf, blockBuf, key, RSA_PKCS1_PADDING);
            if (outLen < 0)
            {
                LOGERROR("Failed to encrypt data[", doText(inData),
                         "]: ", ERR_error_string(ERR_get_error(), nullptr));
                return false;
            }
            outData.append(reinterpret_cast<const char *>(blockBuf), outLen);
            inBuf += inLen;
            dataLen -= inLen;
        }
        return true;
    }

    bool Rsa::decrypt(const ::primitive::raw_data_t &inData,
                      ::primitive::raw_data_t &outData)
    {
        if (m_ownPrivateKey == nullptr)
        {
            return false;
        }
        int blockSize = RSA_size(m_ownPrivateKey);
        unsigned char blockBuf[blockSize] = {0};
        int dataLen = inData.size();
        if ((dataLen % blockSize) != 0)
        {
            LOGERROR("Failed to decrypt data[", doText(inData), "]: data length=", dataLen,
                     " must be a multiple of RSA block size=", blockSize);
            return false;
        }
        const unsigned char *inBuf = reinterpret_cast<const unsigned char *>(inData.c_str());
        outData.clear();
        while (dataLen > 0)
        {
            int outLen = RSA_private_decrypt(blockSize, inBuf, blockBuf, m_ownPrivateKey, RSA_PKCS1_PADDING);
            if (outLen < 0)
            {
                LOGERROR("Failed to decrypt data[", doText(inData),
                         "]: ", ERR_error_string(ERR_get_error(), nullptr));
                return false;
            }
            outData.append(reinterpret_cast<const char *>(blockBuf), outLen);
            inBuf += blockSize;
            dataLen -= blockSize;
        }
        return true;
    }

    bool Rsa::signRecover(RSA *key,
                          const ::primitive::raw_data_t &signature,
                          ::primitive::raw_data_t &digest)
    {
        int blockSize = RSA_size(key);
        unsigned char blockBuf[blockSize] = {0};
        int sLen = signature.size();
        if ((sLen % blockSize) != 0)
        {
            LOGERROR("Failed to recover signature[", doText(signature), "]: its length=", sLen,
                     " must be a multiple of RSA block size=", blockSize);
            return false;
        }
        const unsigned char *inBuf = reinterpret_cast<const unsigned char *>(signature.c_str());
        digest.clear();
        while (sLen > 0)
        {
            int outLen = RSA_public_decrypt(blockSize, inBuf, blockBuf, key, RSA_PKCS1_PADDING);
            if (outLen < 0)
            {
                LOGERROR("Failed to recover signature[", doText(signature),
                         "]: ", ERR_error_string(ERR_get_error(), nullptr));
                return false;
            }
            digest.append(reinterpret_cast<const char *>(blockBuf), outLen);
            inBuf += blockSize;
            sLen -= blockSize;
        }
        return true;
    }

    RSA *Rsa::readPublicKey(const std::string &pubKeyFile)
    {
        FILE *fp = fopen(pubKeyFile.c_str(), "rb");
        if (fp == nullptr)
        {
            LOGERROR("Unable to open public key: file=", pubKeyFile);
            return nullptr;
        }
        RSA *rsa = PEM_read_RSA_PUBKEY(fp, nullptr, nullptr, nullptr);
        fclose(fp);
        return rsa;
    }

    RSA *Rsa::readPrivateKey(const std::string &privKeyFile)
    {
        FILE *fp = fopen(privKeyFile.c_str(), "rb");
        if (fp == nullptr)
        {
            LOGW("Unable to open private key: file=", privKeyFile);
            return nullptr;
        }
        RSA *rsa = PEM_read_RSAPrivateKey(fp, nullptr, nullptr, nullptr);
        fclose(fp);
        return rsa;
    }

} // namespace Crypto
