#include "vfgw/src/hub_manager.hpp"

// AFAP includes.
#include "components/worker_node/src/config_parameters.hpp"
#include "vfgw/src/types.hpp"

#define HTTP_DEFAULT_PORT "8081"
#define HTTP_REQS_MNGR_ID 1

namespace hubs
{
    manager::manager(const TypeInst &owner,
                     boost::asio::io_context &ioc,
                     const configuration::manager &config)
        : ::http::connections::manager(owner, ioc),
          m_config(config),
          m_listener(m_ownInst, m_ioc)
    {
        m_listener.http::interfaces::connections::notifier::client::bind(this);

        m_connectTimeout = m_config.getValue(HTTP_CONNECT_TIMEOUT,
                                             timers::Duration(std::chrono::milliseconds(50)));
    }

    manager::~manager()
    {
        m_listener.stop();
    }

    void manager::start(void)
    {
        if (m_isRunning)
        {
            LOGI(m_ownInst, " HUB manager is already running");
            return;
        }
        m_isRunning = true;

        m_remPort = m_config.getValue<std::string>(HTTP_REMOTE_PORT,
                                                   HTTP_DEFAULT_PORT);

        std::string port = m_config.getValue<std::string>(HTTP_LOCAL_PORT,
                                                          HTTP_DEFAULT_PORT);
        m_listener.start(port);

        LOGI(m_ownInst, " Started HUB manager");
    }

    void manager::stop(void)
    {
        m_listener.stop();
        m_hubConfig.clear();
        m_hubNodes.clear();
        destroyAllConnections();
        m_isRunning = false;
        LOGI(m_ownInst, " Stopped HUB manager");
    }

    void manager::updateHubs(::db::hub::rows &hubs)
    {
        std::unique_lock<std::shared_mutex> guard(m_hubGuard);
        if (m_hubConfig.empty())
        {
            for (auto const &hub : hubs)
            {
                addHub(hub);
            }
            return;
        }

        // Compare updated hubs data with previous one:
        // - Hosts' data
        // - priority order
        // - new/old nodes
        // NOTE: any changes in hub data, excluding node priority change or adding new node,
        // will lead to restarting of round-robin http requests distribution
        for (std::size_t i = 0; i < hubs.size(); i++)
        {
            std::size_t j = 0;
            for (; j < m_hubConfig.size(); j++)
            {
                if (hubs[i].m_idHub == m_hubConfig[j].m_idHub)
                {
                    // Check primary IP address
                    if (hubs[i].m_ipHubP != m_hubConfig[j].m_ipHubP
                        || hubs[i].m_dnsHubP != m_hubConfig[j].m_dnsHubP
                        || hubs[i].m_ipHubS != m_hubConfig[j].m_ipHubS
                        || hubs[i].m_dnsHubS != m_hubConfig[j].m_dnsHubS)
                    {
                        LOGD(m_ownInst,
                             " HUB data is changed: gt=", m_hubConfig[j].m_idHub);
                        replaceHub(j, hubs[i]);
                    }
                    // Check the order
                    if (i != j)
                    {
                        LOGD(m_ownInst,
                             " HUB order for gt=", m_hubConfig[j].m_idHub,
                             "is changed from ", j, "position to ", i);
                        if (i < m_hubConfig.size())
                            swapHubs(i, j);
                        else
                            addHub(hubs[i]);
                    }
                    break;
                }
            } // for (; j < m_hubConfig.size(); j++)
            // new hub node is detected
            if (j == m_hubConfig.size())
            {
                LOGD(m_ownInst, " New HUB is detected: gt=", hubs[i].m_idHub);
                if (i < m_hubConfig.size())
                    insertHub(i, hubs[i]);
                else
                    addHub(hubs[i]);
            }
        } // for (int i = 0; i < hubs.size(); i++)
        // After actions above hubs data is sorted according to updated data in m_hubConfig
        // All extra elements in m_hubConfig are old hubs and should be deleted
        size_t nelem = (hubs.size() < m_hubConfig.size()) ? (m_hubConfig.size() - hubs.size()) : 0;
        size_t elemIx = hubs.size();
        while (nelem)
        {
            LOGD(m_ownInst, " Old HUB is detected: gt=", (*(m_hubConfig.begin() + elemIx)).m_idHub);
            removeHub(elemIx);
            nelem--;
        }
    }

    bool manager::getHubIdByIpAddress(const Uvr::Ip_t &ipAddress, Uvr::Gt_t &hubId)
    {
        std::shared_lock<std::shared_mutex> guard(m_hubGuard);
        for (const auto &hubNode : m_hubNodes)
        {
            if (hubNode.hasIpAddress(ipAddress))
            {
                hubId = hubNode.m_hubId;
                LOGD(m_ownInst,
                     " IP=", ipAddress,
                     " was found among HUBs: hub=", hubId);
                return true;
            }
        }
        return false;
    }

    bool manager::startAndSendRequest(Uvr::SessionId_t reqId,
                                      const std::string &target,
                                      const std::string &payloadData,
                                      const std::shared_ptr<void> &user_data = nullptr,
                                      bool keepAlive)
    {
        http::connection_ptr<http::role::CLIENT> connPtr =
            createOutgoingConnection(reqId, m_remPort, m_connectTimeout);
        connPtr->userData(user_data);
        return connPtr->startSessionAndQuery(target, http::messages::content_json, payloadData, keepAlive);
    }

    bool manager::sendReply(http::messages::result result,
                            const http::connections::id &connId,
                            const std::string &payloadData,
                            bool keepAlive)
    {
        http::connection_ptr<http::role::SERVER> connPtr = getIncomingConnection(connId);
        if (!connPtr)
        {
            LOGW(m_ownInst,
                 " Cannot send http response ",
                 " incoming connection with id=",
                 connId, "is not found");
            return false;
        }
        connPtr->reply(result,
                       result == http::messages::result::ok
                           ? http::messages::content_json
                           : http::messages::content_text,
                       payloadData,
                       keepAlive);
        return true;
    }

    bool manager::resendRequestOnNextHost(const http::connection<http::role::CLIENT> *connPtr) const
    {
        auto ptrConn =
            std::make_shared<http::connection<http::role::CLIENT>>(*connPtr);
        // Reroute request on nest address if it exists
        if (!ptrConn->rerouteRequest())
        {
            pm::interfaces::provider::client::increaseCounter(pm::counter::http::request_routing_to_hub_failed,
                                                              ::pm::counter::value(1));
            return false;
        }
        return true;
    }

    bool manager::connectionAccept(http::connection_ptr<http::role::SERVER> connPtr)
    {
        // Check IP address among HUBs
        for (const auto &hubNode : m_hubNodes)
        {
            if (hubNode.hasIpAddress(connPtr->address()))
            {
                LOGD(m_ownInst,
                     " IP=", connPtr->address(),
                     " was found among HUBs: hub=", hubNode.m_hubId);
                return true;
            }
        }
        // Check IP address among UVRs
        db::uvr::rows uvrs;
        db::interfaces::tables::uvrs::client::get(uvrs, connPtr->address());
        if (!uvrs.empty())
        {
            LOGD(m_ownInst,
                 " IP=", connPtr->address(),
                 " was found among ", uvrs.size(),
                 " UVRs: first uvr=", uvrs[0].m_idUvr);
            return true;
        }

        std::string error = std::string("Unauthorized IP address: ") + connPtr->address();
        connPtr->reply(http::messages::result::forbidden,
                       http::messages::content_text,
                       error,
                       true);
        pm::interfaces::provider::client::increaseCounter(pm::counter::http::http_auth_reject,
                                                          ::pm::counter::value(1));
        return false;
    }

    void manager::connectFailure(http::connection_ptr<http::role::CLIENT> connPtr)
    {
        ::http::connections::manager::connectFailure(connPtr);
        if (connPtr->outgoingRequest().target().compare(HTTP_VERIFY_CALL_TARGET) == 0)
        {
            std::shared_ptr<http::VerifyCallData> dataPtr =
                std::static_pointer_cast<http::VerifyCallData>(connPtr->userData());
            assert(dataPtr != nullptr);
            pm::interfaces::provider::client::increaseCounter(pm::counter::calls::ERR2,
                                                              std::to_string(dataPtr->m_idSrc),
                                                              ::pm::counter::value(1));
        }
    }

    void manager::connectionUp(http::connection_ptr<http::role::CLIENT> connPtr)
    {
        ::http::connections::manager::connectionUp(connPtr);
        pm::interfaces::provider::client::increaseCounter(pm::counter::http::http_up,
                                                          ::pm::counter::value(1));
    }

    void manager::connectionUp(http::connection_ptr<http::role::SERVER> connPtr)
    {
        ::http::connections::manager::connectionUp(connPtr);
        pm::interfaces::provider::client::increaseCounter(pm::counter::http::http_up,
                                                          ::pm::counter::value(1));
    }

    void manager::connectionDown(http::connections::id connId, http::role role)
    {
        ::http::connections::manager::connectionDown(connId, role);
        pm::interfaces::provider::client::increaseCounter(pm::counter::http::http_down,
                                                          ::pm::counter::value(1));
    }

    void manager::getAddress(uint attemptNumber,
                             const std::string &prevAddress,
                             std::string &nextAddress)
    {
        // The algorithm is based on the order,
        // where firstly hosts of primary HUB is used,
        // then hosts of secondary HUB and etc
        LOGD(m_ownInst,
             " Getting IP address from HUB configuration: attempt=", attemptNumber,
             " prev IP=", (prevAddress.empty() ? "\'\'" : prevAddress));
        std::shared_lock<std::shared_mutex> guard(m_hubGuard);
        uint accNumOfHosts = 0;
        const std::string *addressPtr = nullptr;
        for (std::size_t hubIx = 0; hubIx < m_hubNodes.size(); hubIx++)
        {
            // Check whether switching to next HUB for next connect attempt
            if ((attemptNumber - accNumOfHosts) == 0)
            {
                // Get first host of this HUB by round robin
                addressPtr = m_hubNodes[hubIx].getIpOfNextHost();
                break;
            }
            // Check whether this HUB should be choosen for next connect attempt
            accNumOfHosts += m_hubNodes[hubIx].m_hosts.size();
            if (attemptNumber < accNumOfHosts)
            {
                addressPtr = m_hubNodes[hubIx].getIpOfAfterHost(prevAddress);
                break;
            }
        }
        if (addressPtr)
        {
            LOGD(m_ownInst,
                 " Got IP address: prev IP=", prevAddress,
                 ", next IP=", *addressPtr);
            nextAddress = *addressPtr;
        }
        else
        {
            LOGD(m_ownInst,
                 " No available IP addresses: attempt=", attemptNumber,
                 ", prev IP=", prevAddress);
            nextAddress.clear();
        }
    }

    void manager::requestSpecificHandling(http::connections::id connId,
                                          const std::string_view &target,
                                          const std::string &payloadData)
    {
        http::interfaces::messages::handler::client::handleRequest(connId, target, payloadData);
    }

    void manager::responseSpecificHandling(const http::connection<http::role::CLIENT> *connPtr,
                                           http::messages::result result,
                                           const std::string &payloadData,
                                           bool keepAlive)
    {
        http::interfaces::messages::handler::client::handleResponse(result,
                                                                    connPtr,
                                                                    payloadData);
    }

    void manager::addHub(const db::hub::record &hub)
    {
        m_hubConfig.emplace_back(hub);
        m_hubNodes.emplace_back(node(hub.m_idHub,
                                     {host(hub.m_ipHubP, hub.m_dnsHubP),
                                      host(hub.m_ipHubS, hub.m_dnsHubS)}));
    }

    void manager::insertHub(size_t pos, const db::hub::record &hub)
    {
        m_hubConfig.insert(m_hubConfig.begin() + pos, hub);
        m_hubNodes.emplace(m_hubNodes.begin() + pos,
                           node(hub.m_idHub,
                                {host(hub.m_ipHubP, hub.m_dnsHubP),
                                 host(hub.m_ipHubS, hub.m_dnsHubS)}));
    }

    void manager::replaceHub(size_t pos, const db::hub::record &hub)
    {
        m_hubConfig[pos] = hub;
        m_hubNodes[pos] = node(hub.m_idHub,
                               {host(hub.m_ipHubP, hub.m_dnsHubP),
                                host(hub.m_ipHubS, hub.m_dnsHubS)});
    }

    void manager::swapHubs(size_t first, size_t second)
    {
        std::swap(m_hubConfig[first], m_hubConfig[second]);
        std::swap(m_hubNodes[first], m_hubNodes[second]);
    }

    void manager::removeHub(size_t pos)
    {
        m_hubConfig.erase(m_hubConfig.begin() + pos);
        m_hubNodes.erase(m_hubNodes.begin() + pos);
    }

} // namespace hubs
