#pragma once

// AFAP includes.
#include "components/db/src/client.hpp"
#include "components/filesystem/src/filesystem.hpp"
#include "components/hasher/src/hasher.hpp"
#include "components/http/src/interfaces.hpp"
#include "pm/provider/src/pm_node.hpp"
#include "vfgw/src/CryptoRsa.hpp"
#include "vfgw/src/hub_manager.hpp"

namespace vfgw
{
    const int defaultInstance = 0;

    class node : public pm::node,
                 public db::interfaces::tables::client,
                 public db::interfaces::tables::uvrs::client,
                 public db::interfaces::tables::hubs::client,
                 public db::interfaces::tables::Operators::client,
                 public http::interfaces::messages::handler::server
    {
    public:
        explicit node(const TypeInst &ownInst,
                      boost::asio::io_context &ioc,
                      const configuration::manager &config);

        ~node() override;

    protected:
        void stop(void) override;
        void prepareCounters() override;

        void handleSpecificPrimitive(const primitive::GenericPtr &prim) override;

        void handleVerifyCallRequest(const primitive::GenericPtr &prim);

        /// @returns An empty string if the @a requestData was successfully encrypted.
        ///          Otherwise, returns the error text.
        std::string encryptRequestData(const Uvr::UvrId_t uvrId,
                                       std::string &requestData);

        // Implementation of http::interfaces::messages::handler interface
        void handleRequest(http::connections::id connId,
                           const std::string_view &target,
                           const std::string &payloadData) override;
        void handleResponse(http::messages::result result,
                            const http::connection<http::role::CLIENT> *connPtr,
                            const std::string &payloadData) override;

    private:
        void genVerifyRequestHttpBody(boost::json::object &httpBody,
                                      Uvr::UvrId_t toVfnId,
                                      const std::string &sessionId,
                                      const std::string &reqData) const;

        Uvr::Gt_t getHubIdFromFilename(const std::string &filename) const;
        Uvr::UvrId_t getNodeIdFromFilename(const std::string &filename) const;

        bool checkHubSignature(Uvr::Gt_t hubId,
                               const std::string &signature);
        void clearPubKeys(void);

        std::shared_ptr<primitive::VerifyCallResponse>
        checkPreVerifyPolicy(const Uvr::UvrId_t uvrId,
                             const Uvr::SessionId_t sessionId);

        std::shared_ptr<primitive::VerifyCallResponse>
        checkPostVerifyPolicy(const Uvr::UvrId_t uvrId,
                              const Uvr::VerificationVerdict_t verificationVerdict,
                              const Uvr::SessionId_t sessionId);

    private:
        hubs::manager m_hubMngr;                               ///< HUB manager
        fs::local m_fsLocal;                                   ///< Local filesystem
        std::shared_ptr<::db::client> m_dbClient;              ///< DB client
        hasher_t m_hasher;                                     ///< Hash entity
        std::string m_salt;                                    ///< Hash salt
        std::unordered_map<Uvr::Gt_t, RSA *> m_hubPubKeys;     ///< Storage of HUBs' public keys
        mutable std::shared_mutex m_hubKeysGuard;              ///< Protector of shared HUBs' public keys storage in multiple threads (multiple-reader/single-writer mutex)
        std::unordered_map<Uvr::UvrId_t, RSA *> m_nodePubKeys; ///< Storage of NODEs' public keys
        mutable std::shared_mutex m_nodeKeysMutex;             ///< Protector of shared NODEs' public keys storage in multiple threads (multiple-reader/single-writer mutex)
        bool m_httpDataEncrypt{false};                         ///< Indicator to encrypt meaningful data in outgoing HTTP requests
        Crypto::Rsa m_rsa;                                     ///< Object to encrypt/decrypt meaningful data in HTTP messages
    };

    using api = workers::api<node, workers::type::VFGW, workers::max_num_of_insts>;

} // namespace vfgw
