#include "vfs/src/registry.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "TRecQ763Number.hpp"

#include <ostream>

namespace vfs
{
    void registry::setCallTTL(const ::timers::time_point &callTTL)
    {
        m_callTTL = callTTL;
    }

    void registry::setSearchTimeout(const ::timers::time_point &searchTimeout)
    {
        m_searchTimeout = searchTimeout;
    }

    void registry::handleOutgoingCall(const primitive::OutgoingCallIndication &ind)
    {
        std::unique_lock<std::shared_mutex> guard(m_callGuard);
        m_calls.emplace_back(ind.callingNumber, ind.calledNumber);
        LOGD(" registry fill with: ",
             "numA=", TRecQ763Number::toStdString(ind.callingNumber), " ",
             "numB=", TRecQ763Number::toStdString(ind.calledNumber));
    }

    registry::CheckResult
    registry::handleVerifyCall(const primitive::VerifyCallIndication &ind)
    {
        return wasRegistered(ind.callingNumber,
                             ind.calledNumber,
                             ind.originalCalledNumber);
    }

    registry::CheckResult
    registry::handleVerifyCall(const primitive::VerifyCallRequest &req)
    {
        return wasRegistered(req.verifyCallQuery->callingNumber,
                             req.verifyCallQuery->calledNumber,
                             req.verifyCallQuery->originalCalledNumber);
    }

    registry::CheckResult
    registry::wasRegistered(const Uvr::OctetStringPtr &callingNumber,
                            const Uvr::OctetStringPtr &calledNumber,
                            const Uvr::OctetStringPtr &originalCalledNumber)
    {
        std::shared_lock<std::shared_mutex> guard(m_callGuard);

        LOGD(" registry check with: ",
             "numA=", TRecQ763Number::toStdString(callingNumber), " ",
             "numB=", TRecQ763Number::toStdString(calledNumber), " ",
             "numC=", TRecQ763Number::toStdString(originalCalledNumber));

        timers::time_point searchStart = timers::now();

        CheckResult checkResult = wasRegistered(callingNumber,
                                                originalCalledNumber,
                                                searchStart);

        if (checkResult == CheckResult::fail)
        {
            checkResult = wasRegistered(callingNumber,
                                        calledNumber,
                                        searchStart);
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
            if (checkResult == CheckResult::fail)
            {
                LOGD(" Call was not registered: ",
                     TRecQ763Number::toStdString(callingNumber), " ",
                     TRecQ763Number::toStdString(calledNumber));
            }
#endif
        }
        LOGD(" registry check result: ", checkResult);
        return checkResult;
    }

    void registry::do_sanity()
    {
        std::unique_lock<std::shared_mutex> guard(m_callGuard);

        timers::time_point deadline = timers::now() - m_callTTL;

        while (!m_calls.empty())
        {
            const auto &call = m_calls.front();
            if (deadline > call.m_start)
            {
                LOGD(" Removed timed-out call: ",
                     TRecQ763Number::toStdString(call.m_callingNumber), " ",
                     TRecQ763Number::toStdString(call.m_calledNumber));
                m_calls.pop_front();
            }
            else
            {
                LOGT(" No more timed-out calls.");
                break;
            }
        }
    }

    registry::CheckResult
    registry::wasRegistered(const Uvr::OctetStringPtr &callingNumber,
                            const Uvr::OctetStringPtr &calledNumber,
                            const timers::time_point &searchStart)
    {
        if (calledNumber == nullptr || calledNumber->empty() || m_calls.empty())
        {
            LOGD("wasRegistered return false");
            return CheckResult::fail;
        }

        auto iter = m_calls.end();
        do
        {
            if ((timers::now() - searchStart) > m_searchTimeout)
            {
                LOGD(" Call registry search timeout!");
                return CheckResult::timeout;
            }

            --iter;
            if (!TRecQ763Number::areEqual(iter->m_callingNumber, callingNumber))
            {
                LOGD(" Skip check ",
                     TRecQ763Number::toStdString(iter->m_callingNumber),
                     " with ",
                     TRecQ763Number::toStdString(callingNumber));
                continue;
            }

            if (TRecQ763Number::areEqual(iter->m_calledNumber, calledNumber))
            {
                return CheckResult::pass;
            }
        }
        while (iter != m_calls.begin());
        LOGD("wasRegistered return false");
        return CheckResult::fail;
    }

    std::ostream &operator<<(std::ostream &os, const registry::CheckResult &r)
    {
        std::stringbuf trace;
        std::ostream logger(&trace);
        switch (r)
        {
            case registry::CheckResult::pass:
            {
                logger << "pass";
                break;
            }
            case registry::CheckResult::fail:
            {
                logger << "fail";
                break;
            }
            case registry::CheckResult::timeout:
            {
                logger << "timeout";
                break;
            }
        }
        os << trace.str();
        return os;
    }

} // namespace vfs
